#include "lcaio.h"
#include "lcaioIoctl.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/pci.h>
#include <sys/osinfo.h>
#include <sys/seginfo.h>
#include <sys/mman.h>
#include <sys/select.h>

#define   PCI_VENDOR_ID                                 ((unsigned)0x10b5)
#define   PCI_SUBSYS_ID                                 ((unsigned)0x2402)

extern int GLverbose;

static volatile unsigned long  *GLregisters;
static volatile unsigned long  *GLconfRegs;

struct dqIrProxy DQIrProxyList[1];

int lcaioIoctl(pid_t p, file_t *f, int request, long lenOfData, void *data, long maxBytesRead)
{
    struct _io_qioctl_reply         reply;
    struct pmc_lcaio_device         *lcaio_device;
    struct lcaio_pci_config_regs    pci_config_regs;
    struct lcaio_local_config_regs  local_config_regs;
    struct lcaio_registers          registers;
    struct lcaio_int_notify_params  int_notify_params;
    struct lcaio_outputmode         outputmode;
    struct lcaio_inputscan          inputscan;
    struct lcaio_outputscan         outputscan;
    struct lcaio_digital_io_dir     digital_io_dir;
    unsigned int 	                inbuf_threshold, outbuf_threshold;
    int                             inputmode, inputrange, data_format, brdcontrol_val = 0, aux_input = 0, aux_output = 0, dma_enable = 0;
    int                             gen_rate, gen_A_enable, gen_B_enable,gen_B_source, scan_reg, lcaio_digital_io, timeout;
    struct timeval                  tv;


    lcaio_device = f->inode->ino_data;

    if (GLverbose)
        printf("Process Id: %d is requesting a qioctl [0x%08x] of %d bytes, data = [%p], maxBytes = [%d]\n", p, request, lenOfData, data, maxBytesRead);

    switch (request)
    {
    case IOCTL_LCAIO_NO_COMMAND:
        if (GLverbose)
            printf (" : IOCTL_LCAIO_NO_COMMAND entry ... success\n");
        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_READ_REGISTER:
        if (GLverbose)
            printf (" : IOCTL_LCAIO_READ_REGISTER entry ...\n");

        memcpy( &registers, data, sizeof(struct lcaio_registers));

        if ( (signed long) registers.lcaio_register < BOARD_CTRL_REG_OFFSET ||  registers.lcaio_register > DIGITAL_IO_PORT_REG_OFFSET )
        {
            /* Register number that is to be read is invalid */
            reply.status = EINVAL;
            reply.ret_val = -1;
        }
        else
        {	
            registers.register_value = GLregisters[registers.lcaio_register];
            Writemsg(p,  sizeof(reply)-sizeof(reply.data), &registers, sizeof(struct lcaio_registers));
            reply.status = EOK;
            reply.ret_val = 0;
        }

        break;

    case IOCTL_LCAIO_WRITE_REGISTER:
        if (GLverbose)
            printf (" : IOCTL_LCAIO_WRITE_REGISTER entry ...\n");

        memcpy( &registers, data, sizeof(struct lcaio_registers));
                
        if ( (signed long) registers.lcaio_register < BOARD_CTRL_REG_OFFSET ||  registers.lcaio_register > DIGITAL_IO_PORT_REG_OFFSET )
        {
            /* Register number that is to be read is invalid */
            reply.status  = EINVAL;
            reply.ret_val = -1;   
        }
        else
        {   
            GLregisters[registers.lcaio_register] = registers.register_value;
            reply.status = EOK;
            reply.ret_val = 0; 
        }

        break;

    case IOCTL_LCAIO_REQ_INT_NOTIFY:
        if (GLverbose)
            printf (" : IOCTL_LCAIO_REQ_INT_NOTIFY entry ...\n");

        memcpy(&int_notify_params, data, sizeof(struct lcaio_int_notify_params));


        if (( (long) int_notify_params.irq0_int_condition < IRQ0_NO_INT)                       ||
            (int_notify_params.irq0_int_condition > IRQ0_AUX_INPUT_HIGH_TO_LOW)                ||
            ( (long) int_notify_params.irq1_int_condition < IRQ1_NO_INT)                       ||
            (int_notify_params.irq1_int_condition >  IRQ1_INPUT_BUFFER_THRESHOLD_HIGH_TO_LOW)  ||
            ( (long) int_notify_params.irq2_int_condition < IRQ2_NO_INT)                       ||
            (int_notify_params.irq2_int_condition > IRQ2_OUTPUT_BURST_COMPLETE ))
        {
            /* Interrupt condition is invalid */
            /* Register number that is to be read is invalid */
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }    
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress (read/write is in progress) */

            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }            
        if (GLverbose)
            printf ("INT NOTIFY : Int condition : %8.8lx %8.8lx, %8.8lx\n", int_notify_params.irq0_int_condition, int_notify_params.irq1_int_condition, int_notify_params.irq2_int_condition);

        /* Write the interrupt notification mask to the INT_CTRL_REG for the interrupt condition */

        if ((int_notify_params.irq0_int_condition > IRQ0_NO_INT) && (int_notify_params.irq0_int_condition <= IRQ0_AUX_INPUT_HIGH_TO_LOW))
        {
            lcaio_device->irq0_int_req = int_notify_params.irq0_int_condition; 

            if (GLverbose)
                printf ("IRQ0 INT NOTIFY : IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

            /* Set the interrupt condition to notify on */

            GLregisters[INT_CTRL_REG_OFFSET ] = (GLregisters[INT_CTRL_REG_OFFSET] & (IC_IRQ0_MASK)) | (int_notify_params.irq0_int_condition << IC_IRQ0_SHIFT);
	
            if (GLverbose)
                printf ("IRQ0 INT NOTIFY after Set: IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

        } 
        if ((int_notify_params.irq1_int_condition > IRQ1_NO_INT) && (int_notify_params.irq1_int_condition <= IRQ1_INPUT_BUFFER_THRESHOLD_HIGH_TO_LOW)) 
        {
            if (GLverbose)
                printf ("IRQ1 INT NOTIFY : IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

            /* Set the interrupt condition to notify on */

            GLregisters[INT_CTRL_REG_OFFSET] = (GLregisters[INT_CTRL_REG_OFFSET] & (~IC_IRQ1_MASK)) | (int_notify_params.irq1_int_condition << IC_IRQ1_SHIFT); 

            if (GLverbose)
                printf ("IRQ1 INT NOTIFY after Set: IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

        }
        if ((int_notify_params.irq2_int_condition > IRQ2_NO_INT) && (int_notify_params.irq2_int_condition <= IRQ2_OUTPUT_BURST_COMPLETE))
        {
            if (GLverbose)
                printf ("IRQ2 INT NOTIFY : IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

            /* Set the interrupt condition to notify on */

            GLregisters[INT_CTRL_REG_OFFSET] = (GLregisters[INT_CTRL_REG_OFFSET] & (~IC_IRQ2_MASK)) | (int_notify_params.irq2_int_condition << IC_IRQ2_SHIFT);
            if (GLverbose)
                printf ("IRQ2 INT NOTIFY after Set: IC : %8.8x\n", GLregisters[INT_CTRL_REG_OFFSET]);

        }

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_INPUT_MODE:
        if (GLverbose)
            printf (" IOCTL_LCAIO_SET_INPUT_MODE entry ...\n");

        /* Get the argument from the user */

        memcpy (&inputmode, data, sizeof(int));

        /* Check if the argument (input mode) passed is valid */

        if ( inputmode < MODE_DIFFERENTIAL || inputmode > MODE_MON_OUT_CH3 )
        {
            /* Input mode is invalid */

            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress (read/write is in progress) */

            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;  /* Return with failure */
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("I/P MODE : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        /* Mask the register for Analog input mode bits and Update the BOARD_CTRL_REG with the input mode setting mask */

        GLregisters[BOARD_CTRL_REG_OFFSET] =  ( GLregisters[BOARD_CTRL_REG_OFFSET] & ~BC_AIM_MASK) | (inputmode << BC_AIM_SHIFT);

        if (GLverbose)
            printf ("I/P MODE after SET : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        reply.status  = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_RANGE:
        if (GLverbose)
            printf ("SET I/P RANGE : BCR\n");

        /* Get the argument from the user */

        memcpy(&inputrange, data, sizeof(int));

        /* Check if the argument (input range) passed is valid */

        if ( inputrange < RANGE_2p5V || inputrange > RANGE_10V )
        {
            /* Input range is  invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;  /* Return with failure */
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("I/P RANGE : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        /* Mask the register for Analog input range bits and Update the BOARD_CTRL_REG with the input range setting mask */

        GLregisters[BOARD_CTRL_REG_OFFSET] =  ( GLregisters[BOARD_CTRL_REG_OFFSET] & ~BC_RANGE_MASK) | (inputrange << BC_RANGE_SHIFT);

        if (GLverbose)
            printf ("I/P RANGE after SET: BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_DATA_FORMAT:
        if (GLverbose)
            printf ("SET DATA FORMAT : BCR :\n");

        /* Get the argument from the user */

        memcpy( &data_format, data, sizeof(int));

        /* Check if the argument (data_format) passed is valid */

        if (data_format < FORMAT_TWOS_COMPLEMENT || data_format > FORMAT_OFFSET_BINARY )
        {
            /* Data format is  invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }

        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;  /* Return with failure */
            reply.ret_val = -1;   
            break;

        }

        if (GLverbose)
            printf ("DATA FORMAT : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        /* Mask the register for Data format bits and Update the BOARD_CTRL_REG with the data format setting mask */

        GLregisters[BOARD_CTRL_REG_OFFSET] = (GLregisters[BOARD_CTRL_REG_OFFSET] & ~BC_DATA_FORMAT_MASK) | (data_format << BC_DATA_FORMAT_SHIFT);

        if (GLverbose)
            printf ("DATA FORMAT after SET: BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_OUTPUT_MODE:
        if (GLverbose)
            printf (" IOCTL_LCAIO_SET_OUTPUT_MODE entry ...\n");

        memcpy (&outputmode, data, sizeof(outputmode));

        if (GLverbose)
            printf ("O/P MODE seq : BCR : %8.8lx\n", outputmode.sequential);

        /* Check if the mode requested is valid (i.e, either sequential or multi channel mode */

        if ( outputmode.sequential > OUT_MODE_SIMULTANEOUS || (int) outputmode.sequential <  OUT_MODE_CHAN_SEQUENTIAL )
        {
            /* Output mode is invalid */

            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }   
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */

            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;  /* Return with failure */
            reply.ret_val = -1;   
            break;
        }             
        if (GLverbose)
            printf ("O/P MODE : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        brdcontrol_val = GLregisters[BOARD_CTRL_REG_OFFSET];

        /* Set the bits to zero that are to be set */                   

        brdcontrol_val =  brdcontrol_val & ~BC_SIMUL_OUT_MASK | (~BC_EN_OUT_BURST_MASK & BC_EN_OUT_LOOPING_MASK);

        /* Set the bits for output mode */

        brdcontrol_val = brdcontrol_val | (outputmode.sequential<<BC_SIMUL_OUT_SHIFT);

        /* If output burst enable mode is requested */

        if ( outputmode.enable_output_burst)   
            brdcontrol_val |= BC_EN_OUT_BURST_MASK;
        else
            brdcontrol_val &= ~BC_EN_OUT_BURST_MASK;

        /* If enabling output looping mode is requested */

        if ( outputmode.enable_output_looping)        
            brdcontrol_val |= BC_EN_OUT_LOOPING_MASK;  
        else
            brdcontrol_val &= ~BC_EN_OUT_LOOPING_MASK;

        /* Write the final value to the BOARD_CTRL_REG */

        GLregisters[BOARD_CTRL_REG_OFFSET] = brdcontrol_val;

        printf ("O/P MODE after Set : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_INPUT_SYNC:
        if (GLverbose)
            printf ("INPUT SYNC : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        brdcontrol_val = GLregisters[BOARD_CTRL_REG_OFFSET];

        /* Set the Input sync bit in the Board Control register */
        brdcontrol_val |= BC_INPUT_SYNC_MASK;

        if (GLverbose)
            printf ("INPUT SYNC Mask written : BCR : %8.8x\n", brdcontrol_val);

        GLregisters[BOARD_CTRL_REG_OFFSET] = brdcontrol_val;

/*                sleep_on_timeout (&lcaio_device->ioctlwq, MSECS_TO_SLEEP(500)); */

        { int waitCnt;

        for( waitCnt = 0; waitCnt <= 1000000; waitCnt++ )
        {
            if( ( GLregisters[BOARD_CTRL_REG_OFFSET] & 0x00001000 ) == 0)
                break;
        }
        }

        printf ("INPUT SYNC after Set : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_OUTPUT_SYNC:
        if (GLverbose)
            printf ("OUTPUT SYNC : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);


        brdcontrol_val =  GLregisters[BOARD_CTRL_REG_OFFSET];

        /* Set the Output sync bit in the Board Control register */

        brdcontrol_val |= BC_OUTPUT_SYNC_MASK;

        if (GLverbose)
            printf ("OUTPUT SYNC Mask written : BCR : %8.8x\n", brdcontrol_val);

        GLregisters[BOARD_CTRL_REG_OFFSET] = brdcontrol_val;

        { int waitCnt;

        for( waitCnt = 0; waitCnt <= 1000000; waitCnt++ )
        {
            if( ( GLregisters[BOARD_CTRL_REG_OFFSET] & 0x00000800 ) == 0)
                break;
        }
        }

        if (GLverbose)
            printf ("OUTPUT SYNC after Set : BCR : %8.8x\n", GLregisters[BOARD_CTRL_REG_OFFSET]);


        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_AUTO_CAL:
        if (GLverbose)
            printf (" : IOCTL_LCAIO_AUTO_CAL entry ...\n");

        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;  /* Return with failure */
            reply.ret_val = -1;   
            break;
        }

        lcaio_device->autocalibration = 1;
        brdcontrol_val =  GLregisters[BOARD_CTRL_REG_OFFSET];
  
        if (GLverbose)
            printf ("BCR : %8.8x\n",brdcontrol_val);


        /* Set the Auto calibration bit in the Board Control register */
 
        brdcontrol_val |= BC_AUTO_CAL_MASK;
  
        printf ("BCR MASK written : %8.8x\n",brdcontrol_val);

        GLregisters[BOARD_CTRL_REG_OFFSET] = brdcontrol_val;

        /* sleep_on_timeout () for this task for a time of AUTO_CAL_TIMEOUT_MS */

        // probably should try select here
//                interruptible_sleep_on_timeout(&lcaio_device->ioctlwq, MSECS_TO_SLEEP(AUTO_CAL_TIMEOUT_MS)); 

        printf ("BCR after CLR: %8.8x\n",GLregisters[BOARD_CTRL_REG_OFFSET]);

        /* If interrupt notification is requested, check the timeout */

        if ( lcaio_device->irq0_int_req == 1 )
        {
            lcaio_device->procWaitQ.pid = p;

//                   if ( lcaio_device->autocalibration == 1 )   /* Timeout occurred */
//                   {
//                        lcaio_device->error = LCAIO_IOCTL_TIMEOUT;
//
//                        return -EIO;
//                    }
            return 0;
        }
        else 
        {
            lcaio_device->autocalibration = 0;
            reply.status = EOK;
            reply.ret_val = 0;
        }  
        break;

    case IOCTL_LCAIO_INITIALIZE:
        if (GLverbose)
            printf ( "IOCTL_LCAIO_INITIALIZE entry, pid %ld ...\n", p);
		
        GLregisters[ BOARD_CTRL_REG_OFFSET ] |= 0x00008000 ;

        lcaio_device->procWaitQ.pid = p;
		
        return(0);


    case IOCTL_LCAIO_SET_INPUT_BUFFER_THRESHOLD:
        if (GLverbose)
            printf ("SET I/P THRESH : IDBC : \n");

        /* Get the argument from the user */
        memcpy(&inbuf_threshold, data, sizeof(int) );

        /* Check if the argument passed is valid */

        if ( (int) inbuf_threshold < 0 )
        {
            /* Input buffer threshold value is  invalid */

            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }

        if (GLverbose)
            printf ("I/P THRESH : IDBC : %8.8x\n", GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET]);

        /* Set the Input buffer threshold in Input buffer control register */
        GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] =  (GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET]  & ~IDBC_THRESHOLD_MASK)  | inbuf_threshold;

        if (GLverbose)
            printf ("I/P THRESH after SET: IDBC : %8.8x\n", GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_CLEAR_INPUT_BUFFER:
        if (GLverbose)
            printf("IDBC mask value = %8.8x\n", (GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] |IDBC_CLEAR_BUFFER_MASK ));

        /* Set the clear buffer flag of Input buffer Control register */

        GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] = ( GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] | IDBC_CLEAR_BUFFER_MASK );


//	        sleep_on_timeout (&lcaio_device->ioctlwq, MSECS_TO_SLEEP(300));
        tv.tv_sec  = 0;
        tv.tv_usec = 300000;
        select(0, NULL, NULL, NULL, &tv);


        lcaio_device->readbuf_samples = 0;
        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_OUTPUT_BUFFER_THRESHOLD:
        if (GLverbose)
            printf ("O/P THRESH : ODBC : %8.8x\n", GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET]);

        /* Get the argument from the user */

        memcpy ( &outbuf_threshold, data, sizeof(int));

        /* Check if the argument passed is valid */

        if ( (int) outbuf_threshold < 0 )
        {
            /* Output buffer threshold value is invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        /* Set the Output buffer threshold in Output buffer control register */
        GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET] =  (GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET]  & ~ODBC_THRESHOLD_MASK ) | (outbuf_threshold);

        if (GLverbose)
            printf ("O/P THRESH after SET: ODBC : %8.8x\n", GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;
 
    case IOCTL_LCAIO_CLEAR_OUTPUT_BUFFER:

        if (GLverbose)
            printf("ODBC mask value = %8.8x\n", (GLregisters [OUTPUT_BUFFER_CTRL_REG_OFFSET] | ODBC_CLEAR_BUFFER_MASK ));

        /* Set  the Clear buffer flag of output buffer Control register */

        GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET] = ( GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET] | ODBC_CLEAR_BUFFER_MASK );

        if (GLverbose)
            printf("ODBC after CLR = %8.8x\n", GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET]);

	//      sleep_on_timeout (&lcaio_device->ioctlwq, MSECS_TO_SLEEP(300));

        tv.tv_sec  = 0;
        tv.tv_usec = 300000;
        select(0, NULL, NULL, NULL, &tv);

        lcaio_device->writebuf_samples = 0;
        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_GEN_A_RATE:
        /* Get the argument from the user */
        memcpy ( &gen_rate, data, sizeof(int));
	 
        /* Check if the argument passed is valid */

        if ( gen_rate < 0 )
        {
            /* Rate for generator A is invalid */
            if (GLverbose)
                printf ("GEN A RATE : requested rate is %d\n", gen_rate);
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("GEN A RATE : RATE A REG : %8.8X\n", GLregisters[RATE_A_GEN_REG_OFFSET]);

        /* Write the selected Generator's rate to Rate A Generator Register */

        GLregisters[RATE_A_GEN_REG_OFFSET] =  (GLregisters[RATE_A_GEN_REG_OFFSET] & ~RG_NRATE_MASK) | gen_rate;

        if (GLverbose)
            printf ("GEN A RATE after Set : RATE A REG : %8.8X\n", GLregisters[RATE_A_GEN_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_GEN_B_RATE:
        /* Get the argument from the user */
        memcpy ( &gen_rate, data, sizeof(int));
	 
        /* Check if the argument passed is valid */

        if ( gen_rate < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("GEN B RATE : RATE B REG : %8.8X\n", GLregisters[RATE_B_GEN_REG_OFFSET]);

        /* Write the selected Generator's rate to Rate B Generator Register */

        GLregisters[RATE_B_GEN_REG_OFFSET] =  (GLregisters[RATE_B_GEN_REG_OFFSET] & ~RG_NRATE_MASK) | gen_rate ;

        if (GLverbose)
            printf ("GEN B RATE after Set : RATE B REG : %8.8X\n", GLregisters[RATE_B_GEN_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_GEN_A_ENABLE:

        /* Get the argument from the user */
        memcpy ( &gen_A_enable, data, sizeof(int));
	 
        /* Check if the argument passed is valid */

        if ( gen_A_enable < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("GEN A Enable : RATE A GEN REG : %8.8X\n", GLregisters[RATE_A_GEN_REG_OFFSET]);

        /* If generator A is requested for enabling, then enable */	

        if ( gen_A_enable == 1)                 /* Set the enable bit in the Rate A Generator register */
            GLregisters[RATE_A_GEN_REG_OFFSET] =  GLregisters[RATE_A_GEN_REG_OFFSET] & ~RG_GEN_DISABLE_MASK;
        else          /* Disable the Rate A Generator register */
            GLregisters[RATE_A_GEN_REG_OFFSET]  = GLregisters[RATE_A_GEN_REG_OFFSET] | RG_GEN_DISABLE_MASK;

        if (GLverbose)
            printf ("GEN A Enable after Set : RATE A GEN REG : %8.8X\n", GLregisters[RATE_A_GEN_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;
 

    case IOCTL_LCAIO_SET_GEN_B_ENABLE:

        /* Get the argument from the user */
        memcpy ( &gen_B_enable, data, sizeof(int));
	 
        /* Check if the argument passed is valid */

        if ( gen_B_enable < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("GEN B Enable : RATE B GEN REG : %8.8X\n", GLregisters[RATE_B_GEN_REG_OFFSET]);

        /* If generator A is requested for enabling, then enable */	

        if ( gen_B_enable == 1)                 /* Set the enable bit in the Rate B Generator register */
            GLregisters[RATE_B_GEN_REG_OFFSET] =  GLregisters[RATE_B_GEN_REG_OFFSET] & ~RG_GEN_DISABLE_MASK;
        else          /* Disable the Rate B Generator register */
            GLregisters[RATE_B_GEN_REG_OFFSET]  = GLregisters[RATE_B_GEN_REG_OFFSET] | RG_GEN_DISABLE_MASK;

        if (GLverbose)
            printf ("GEN B Enable after Set : RATE B GEN REG : %8.8X\n", GLregisters[RATE_B_GEN_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_GEN_B_SRC :

        /* Get the argument from the user */
        memcpy ( &gen_B_source, data, sizeof(int));
	 
        /* Check if the argument passed is valid */

        if ( gen_B_source < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }
        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("GEN B SRC : SCAN & SYNC : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        /* Update the Scan and Sync Control Register's Gen-B clk src field with the user requested source */

        GLregisters[SCAN_SYNC_CTRL_REG_OFFSET] = (GLregisters[SCAN_SYNC_CTRL_REG_OFFSET] & ~SC_RATE_B_CLK_MASK) | gen_B_source << SC_RATE_B_CLK_SHIFT;

        if (GLverbose)
            printf ("GEN B SRC sfter set: SCAN & SYNC : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_INPUT_SCAN:

        /* Copy the input scan structure from the user/application */
        memcpy (&inputscan, data, sizeof(inputscan));

        /* Check if the scan size, scan clock source, scan mode and scan channel arguments are valid */

        if ( ( (int) inputscan.scan_size           < SCAN_SIZE_4_CHAN    || inputscan.scan_size           > SCAN_SIZE_32_CHAN)            || 
             ( (int) inputscan.scan_clock_source   < SCAN_CLK_SRC_GEN_A  || inputscan.scan_clock_source   > SCAN_CLK_SRC_BCR_INPUT_SYNC)  || 
             ( (int) inputscan.scan_mode           < SCAN_MODE_MULTIPLE  || inputscan.scan_mode           > SCAN_MODE_SINGLE)             || 
             ( (int) inputscan.single_scan_channel < MIN_CHANNEL	  || inputscan.single_scan_channel > MAX_CHANNEL) )
        {
            /* Input scan arguments are invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }             

        if (GLverbose)
            printf ("INPUT SCAN : SCAN & SYNC REG : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        /* Read the scan register*/	

        scan_reg = GLregisters[SCAN_SYNC_CTRL_REG_OFFSET];

        /* Mask off bits to be set. */

        scan_reg = scan_reg & (~ SC_AIN_SCAN_SIZE_MASK)   
            & (~ SC_AIN_SCAN_CLK_MASK)    
            & (~ SC_AIN_SCAN_MODE_MASK)   
            & (~ SC_AIN_SINGLE_CHAN_MASK) 
            & (~ SC_AIN_TWO_CHAN_MASK);

        /* Set the bits as per new selections */

        scan_reg |= (inputscan.scan_size << SC_AIN_SCAN_SIZE_SHIFT)         | 
            (inputscan.scan_clock_source << SC_AIN_SCAN_CLK_SHIFT)  | 
            (inputscan.scan_mode  << SC_AIN_SCAN_MODE_SHIFT)        | 
            (inputscan.single_scan_channel << SC_AIN_SINGLE_CHAN_SHIFT);

        if (inputscan.two_channel_scan)
            scan_reg = scan_reg | SC_AIN_TWO_CHAN_MASK;
        
        /* Write to the Scan Control Register. */

        GLregisters[SCAN_SYNC_CTRL_REG_OFFSET] = scan_reg;

        if (GLverbose)
            printf ("INPUT SCAN after Set: SCAN & SYNC REG : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_OUTPUT_SCAN:
        /* Copy the output scan structure from the user/application */
        memcpy (&outputscan, data, sizeof(outputscan));

        /* Check if the scan clock source, scan burst sync source, scan external sync source arguments are valid */	 		
        if ( ( (int) outputscan.scan_clock_src < SCAN_CLK_SRC_GEN_A)               || 
             (outputscan.scan_clock_src > SCAN_CLK_SRC_DISABLED)            || 
             ((int) outputscan.scan_burst_syncsrc < SCAN_SYNC_SRC_GEN_A)          || 
             (outputscan.scan_burst_syncsrc > SCAN_SYNC_SRC_BCR_INPUT_SYNC) || 
             ((int) outputscan.scan_ext_syncsrc < SCAN_EXT_SYNC_SRC_AIN)          || 
             (outputscan.scan_ext_syncsrc > SCAN_EXT_SYNC_SRC_DISABLED))	
        {
            /* Output scan arguments are invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("OUTPUT SCAN : SCAN & SYNC REG : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        /* Read the scan register */
        scan_reg = GLregisters[SCAN_SYNC_CTRL_REG_OFFSET];

        /* Mask off bits to be set */

        scan_reg = scan_reg & (~ SC_AOUT_CLK_MASK) & (~ SC_AOUT_SYNC_MASK) & (~ SC_AOUT_EXT_SYNC_MASK);	

        /* Set the bits as per new selections */

        scan_reg |=  (outputscan.scan_clock_src << SC_AOUT_CLK_SHIFT)      | 
            (outputscan.scan_burst_syncsrc << SC_AOUT_SYNC_SHIFT) | 
            (outputscan.scan_ext_syncsrc << SC_AOUT_EXT_SYNC_SHIFT);

        /* Write to the Scan Control Register. */
        GLregisters[SCAN_SYNC_CTRL_REG_OFFSET] = scan_reg;

        if (GLverbose)
            printf ("OUTPUT SCAN after Set: SCAN & SYNC REG : %8.8x\n", GLregisters[SCAN_SYNC_CTRL_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;
					          
    case IOCTL_LCAIO_SET_DIGITAL_IO_DIR:

        /* Copy the digital_io_dir structure from the user/application */

        memcpy (&digital_io_dir, data, sizeof(digital_io_dir));

        if (GLverbose)
            printf ("DIGITAL IO DIR : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);

        /* Check if the lowbyte_dir and highbyte_dir arguments are valid */

        if ((( (int) digital_io_dir.low_byte_dir   < DIGITAL_IO_DIR_INPUT)   || 
             (digital_io_dir.low_byte_dir   > DIGITAL_IO_DIR_OUTPUT)) || 
            (( (int) digital_io_dir.high_byte_dir < DIGITAL_IO_DIR_INPUT)   || 
             (digital_io_dir.high_byte_dir  > DIGITAL_IO_DIR_OUTPUT)))
        {
            /* Set digital IO arguments are invalid */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        /* Set the low byte and high byte direction of the digital port in the DIGITAL_IO_PORT_REG */

        GLregisters[DIGITAL_IO_PORT_REG_OFFSET] = (GLregisters[DIGITAL_IO_PORT_REG_OFFSET]  & 
                                                   (~DIO_LOW_BYTE_DIR_MASK)     & 
                                                   (~DIO_HIGH_BYTE_DIR_MASK))   | 
            (digital_io_dir.low_byte_dir  << DIO_LOW_BYTE_DIR_SHIFT) | 
            (digital_io_dir.high_byte_dir << DIO_HIGH_BYTE_DIR_SHIFT);

        if (GLverbose)
            printf ("DIGITAL IO DIR after Set : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_READ_DIGITAL_IO:

        if (GLverbose)
            printf ("READ DIGITAL IO : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);

        /* Read the Digital IO Port register contents */
        lcaio_digital_io = GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & DIO_BOTH_BYTES_MASK;

        /* Copy to the register contents to the user */
        Writemsg(p,  sizeof(reply)-sizeof(reply.data), &lcaio_digital_io, sizeof(lcaio_digital_io));
//		Writemsg(p,  sizeof(reply), &lcaio_digital_io, sizeof(lcaio_digital_io));

        if (GLverbose)
            printf ("DIGITAL IO after mask state: %8.8x\n", lcaio_digital_io);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_DIGITAL_IO:

        /* Get the argument from the user */
        memcpy ( &lcaio_digital_io, data, sizeof(lcaio_digital_io));

        /* Check if the argument passed is valid */

        if ( lcaio_digital_io < 0 )
        { 
            /* Invalid argument */
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }		   

        if (GLverbose)
            printf ("DIGITAL IO : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);

        /* Write the Input range to the Digital IO Port Register */ 
        {
            int value ;
            value = (GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & ~DIO_BOTH_BYTES_MASK ) | lcaio_digital_io;
            printf ("Set: %8.8x\n", value);
            GLregisters[DIGITAL_IO_PORT_REG_OFFSET] = value;
        }
//                GLregisters[DIGITAL_IO_PORT_REG_OFFSET] = (GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & ~DIO_BOTH_BYTES_MASK ) | lcaio_digital_io;

        if (GLverbose)
            printf ("DIGITAL IO after Set: %8.8x, %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET], lcaio_digital_io);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_READ_AUX_INPUT:

        if (GLverbose)
        {
            printf (  " IOCTL_LCAIO_READ_AUX_INPUT entry ...\n");
            printf ("AUX I/P : DIGITAL IO : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);
            printf ("AUX I/P state : DIGITAL IO : %8.8x\n",  GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & DIO_AUX_INPUT_MASK);
        }

        /* Read the auxiliary input state from the Digital IO Port register  as active or inactive */

        if (GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & DIO_AUX_INPUT_MASK)            /* Auxiliary input is in active mode */
            aux_input = AUX_INPUT_ACTIVE;
        else   /* Auxiliary input is NOT in active mode */
            aux_input = AUX_INPUT_INACTIVE;

        Writemsg(p,  sizeof(reply), &aux_input, sizeof(aux_input));

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_SET_AUX_OUTPUT:

        /* Get the argument from the user */

        memcpy ( &aux_output, data, sizeof(aux_output));

        /* Check if the aux output mode to be set is either ACTIVE or INACTIVE */

        if ((aux_output < AUX_OUTPUT_INACTIVE) || (aux_output > AUX_OUTPUT_ACTIVE))
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        if (GLverbose)
            printf ("AUX O/P: DIGITAL IO : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);
             
        /* Update the Digital IO Port Register with DIO_AUX_OUTPUT */

        GLregisters[DIGITAL_IO_PORT_REG_OFFSET] = (GLregisters[DIGITAL_IO_PORT_REG_OFFSET] & ~DIO_AUX_OUTPUT_MASK) | (aux_output << DIO_AUX_OUTPUT_SHIFT);

        if (GLverbose)
            printf ("AUX O/P after Set : DIGITAL IO : %8.8x\n", GLregisters[DIGITAL_IO_PORT_REG_OFFSET]);

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    case IOCTL_LCAIO_GET_DEVICE_ERROR:

        if (GLverbose)
            printf ( " IOCTL_LCAIO_GET_DEVICE_ERROR entry ...\n");

        /* Copy the error to the user that is updated in the device structure */
        Writemsg(p,  sizeof(reply)-sizeof(reply.data), &lcaio_device->error, sizeof(lcaio_device->error));

        reply.status = EOK;
        reply.ret_val = 0;
        break;

 
    case IOCTL_LCAIO_READ_PCI_CONFIG:
        if (GLverbose)
            printf ( " IOCTL_LCAIO_READ_PCI_CONFIG entry ...\n");
        pci_config_regs = read_pci_config_regs (lcaio_device);		
        if (GLverbose)
            printf ("subSys [%x], Vendor [%x]\n", pci_config_regs.device_vendor_ID, pci_config_regs.subsystem_vendor_ID);

        Writemsg(p,  sizeof(reply)-sizeof(reply.data), &pci_config_regs, sizeof(pci_config_regs));
        reply.zero[0] = 0;
        reply.zero[1] = 0;
        reply.status = EOK;
        reply.ret_val = 0;
        break;

 
    case IOCTL_LCAIO_READ_LOCAL_CONFIG:
        if (GLverbose)
            printf (" IOCTL_LCAIO_READ_LOCAL_CONFIG entry ...\n");

        local_config_regs = read_local_config_regs (lcaio_device);

        /* Copy the structure holding the register contents to the user */ 
        Writemsg(p,  sizeof(reply)-sizeof(reply.data), &local_config_regs, sizeof(local_config_regs));

        reply.status = EOK;
        reply.ret_val = 0;
        break;

 
    case IOCTL_LCAIO_WRITE_PCI_CONFIG_REG:
        if (GLverbose)
            printf (" IOCTL_LCAIO_WRITE_PCI_CONFIG entry ...\n");

        memcpy(&registers, data, sizeof(registers));
        if ((registers.lcaio_register < STATUS_COMMAND_OFFSET) || (registers.lcaio_register > LAT_GNT_INT_PIN_LINE_OFFSET))
        {
            /* Pci config register cannot be written */
            if (GLverbose)
                printf (" : Error :: Pci config register cannot be written");
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;

        }

        if ( (registers.lcaio_register > STATUS_COMMAND_OFFSET)  && (registers.lcaio_register < BIST_HDR_TYPE_LAT_CACHE_SIZE_OFFSET) )
        {
            /* Class revision id register cannot be written -- Invalid */
            if (GLverbose)
                printf(" : Error :: Class revision id register cannot be written -- Invalid\n");
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }	
        if ((registers.lcaio_register > PCI_BASE_ADDR_1_OFFSET) && (registers.lcaio_register < PCI_BASE_ADDR_LOC_ROM_OFFSET))
        {
            /* Subsystem id register cannot be written -- invalid */
            if (GLverbose)
                printf (" : Subsystem id register cannot be written -- invalid\n"); 
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }
        if ((registers.lcaio_register > PCI_BASE_ADDR_LOC_ROM_OFFSET) && (registers.lcaio_register < LAT_GNT_INT_PIN_LINE_OFFSET))
        {
            /* Register 13 and 14 cannot be written */
            if (GLverbose)
                printf (": Register 13 and 14 cannot be written\n");
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }
        /* Write the value to the requested register offset */

        _CA_PCI_Write_Config_DWord ( lcaio_device->busNum, lcaio_device->devFuncNum, registers.lcaio_register << 2,  1, (char *) registers.register_value);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_WRITE_LOCAL_CONFIG_REG:
        if (GLverbose)
            printf (" IOCTL_LCAIO_WRITE_LOCAL_CONFIG entry ...\n");
        memcpy( &registers, data, sizeof(registers));

        if (((int)registers.lcaio_register < PCI_TO_LOC_ADDR_0_RNG_OFFSET) || (registers.lcaio_register > BUS_REG_DESC_1_FOR_PCI_LOC_OFFSET)) 
        {
            /* Local config register to be written is Invalid */
            if (GLverbose)
                printf(" : Local config register to be written is Invalid\n");

            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }


        if ( lcaio_device->sampling_inprogress ==  1 )
        {
            /* Data sampling is in progress */
            if (GLverbose)
                printf("Data sampling is in progress\n");
	
            lcaio_device->error = LCAIO_OPERATION_CANCELLED;
            reply.status  = EPERM;
            reply.ret_val = -1;   
            break;
        }

        GLconfRegs[registers.lcaio_register] = registers.register_value;

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_TIMEOUT:
        if (GLverbose)
            printf ("TIMEOUT : %ld\n", lcaio_device->timeout_value);

        memcpy(&timeout, data, sizeof(timeout));

        if ( timeout < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }

        lcaio_device->timeout_value = timeout;

        if (GLverbose)
            printf ("TIMEOUT after set : %ld\n", lcaio_device->timeout_value);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_SET_DMA_ENABLE:
        if (GLverbose)
            printf ("DMA Enable : %d\n", lcaio_device->dma_enabled);
        /* Get the argument from the user */

        memcpy (&dma_enable, data, sizeof(dma_enable));

        /* Check if the argument passed is valid */

        if ( dma_enable < 0 )
        {
            lcaio_device->error = LCAIO_INVALID_PARAMETER;
            reply.status  = EINVAL;
            reply.ret_val = -1;   
            break;
        }		

        /* Set the DMA enable flag in the device structure */
        lcaio_device->dma_enabled = dma_enable;

        if (GLverbose)
            printf ("DMA Enable after Set: %d\n", lcaio_device->dma_enabled);

        reply.status = EOK;
        reply.ret_val = 0;
        break;


    case IOCTL_LCAIO_READ_DEBUG:
        if (GLverbose)
            printf ( " : IOCTL_LCAIO_READ_DEBUG ...entry\n");
        /* Copy the debug messages from the device structure's buffer to the user */

        Writemsg(p,  sizeof(reply)-sizeof(reply.data), lcaio_device->debug_buffer, sizeof(lcaio_device->debug_buffer));

        reply.status = EOK;
        reply.ret_val = 0;
        break;

    default :       /* Invalid ioctl command */ 
        reply.status = EINVAL;
        reply.ret_val = 0;
        break;

    } 

    Reply(p, &reply, sizeof(reply));
    return 0;
}

int lcaioWrite(pid_t p, file_t *f, unsigned n, void *buf, unsigned len)
{
    struct pmc_lcaio_device *lcaio_device;
    struct _io_write_reply  w;
    int samples_requested, transfer_samples, count;

    count = len;
	
    /* Get the device structure from the private_data field */
    lcaio_device =  f->inode->ino_data;

    if (GLverbose)
        printf (" : lcaio <write> function entry ...\n");
    /* Get the requested no. of samples count and validate */

    samples_requested = count / sizeof (long);

    if ( samples_requested <= 0 || samples_requested > MAX_BUFFER_SAMPLES)
    {
        /* Invalid no. of samples to write */  
        reply_errno(p, EINVAL);
        return 0;    /* Write failed, no data is written */   
    } 

    if (GLverbose)
    {
        int sample_count; 	
        long *buff;

	buff = (long *) buf;
        printf ("Driver write function :: \n");
	
        for ( sample_count = 0; sample_count < samples_requested; sample_count++)
        {
            printf (" %8.8lx ", buff[sample_count]);
        }
        printf ("\n");
    }

    memcpy( lcaio_device->write_buffer, buf, samples_requested * sizeof(long));

    /* Write the data to the device FIFO/buffer */

    transfer_samples = write_to_buffer (p, lcaio_device, samples_requested, nonblock(f));

    if (transfer_samples > 0)
    {
        w.status = EOK;
        w.nbytes = transfer_samples * sizeof(long);
        w.zero  = 0;
        Reply(p, &w, sizeof w);
        return(1);
    }
    return(transfer_samples);
}

/*
 * File : write_to_buffer () 
 *
 * Description : This function performs the actual write operation on the
 * buffer (FIFO). This function is called from the pmc_lcaio_write () 
 * driver module.
 *
 * Input parameters : lcaio_device  - Device structure holding device 
 *				      information and buffer to hold the
 *                                    data read from the FIFO
 *                    requested_samples - No. of bytes to be read ( No. of
 *                                        samples * 4 )  
 *                    mode  - read mode ( BLOCK or NON_BLOCK )
 *                     
 * Return Value : Returns the number of bytes that have been written 
 *                OR returns -EINVAL if no. of bytes to be written is invalid.
 *
 * Comments : This is NOT a registered driver function call.  
 *
 * Revision History : 5/12/2001 - Device sampling state reset at the end of 
 *                    the function. ( lcaio_device->sampling_inprogress = 0 )
 *
 */

int
write_to_buffer ( pid_t p, struct pmc_lcaio_device *lcaio_device, int requested_samples, int mode)
{
    unsigned long rest;
    unsigned long sample_count, remaining_samples, no_of_samples;
    unsigned long start_address = 0;
    int      ret;
    struct timeval          tv;


    if (GLverbose)
        printf (" : lcaio <write to buffer> function entry ...\n");

    /* Set the device sampling state as busy */

    lcaio_device->sampling_inprogress = 1;
    remaining_samples = requested_samples;

    lcaio_device->outbuf_threshold = (GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET] & 0x7FFF);
  
    if (GLverbose)
        printf ("thresold = %8.8lx\n", lcaio_device->outbuf_threshold);

    do 
    {
        /* Check if we have sufficient space to write to the buffer */
        if (  GLregisters[OUTPUT_BUFFER_CTRL_REG_OFFSET] & ODBC_THRESHOLD_FLAG_MASK )
        {
            /* Free space in output buffer is less to do a write operation */
            /* Request for the interrupt notification for the output buffer 
               threshold high-to-low */
            /* Write OC_INP_THRES_HIGH_TO_LOW_VAL to OUTPUT_BUFFER_CTRL_REG */

	    GLregisters[INT_CTRL_REG_OFFSET] = (GLregisters[INT_CTRL_REG_OFFSET] & (~IC_IRQ2_MASK)) | (IRQ2_OUTPUT_BUFFER_THRESHOLD_HIGH_TO_LOW << IC_IRQ1_SHIFT);

            lcaio_device->procWaitQ.pid = p;
            lcaio_device->procWaitQ.nbytes = no_of_samples * sizeof(long);
            lcaio_device->procWaitQ.writeOffset = 1;
            tv.tv_sec  = lcaio_device->timeout_value;
            tv.tv_usec = 0;
            ret = select(0, NULL, NULL, NULL, &tv);
            if (GLverbose)
                printf("return from select is %d\n");
//            interruptible_sleep_on_timeout (&lcaio_device->writewq, SECS_TO_SLEEP(lcaio_device->timeout_value));
            if ( lcaio_device->timeout_value > 0 )                  
            {
                /* Timeout has occurred, reset the device sampling state  */

                lcaio_device->sampling_inprogress = 0; 

                /* Set the error */
                lcaio_device->error = LCAIO_PIO_TIMEOUT;
                reply_errno(p, EIO);
            }

        } /* End of if (buffer space check) */

        /* If requested no. of samples to write is more than the free space 
           in the device buffer then transmit only threshold number of samples 
           as the first block */ 

        if (remaining_samples > lcaio_device->outbuf_threshold ) 
            no_of_samples = lcaio_device->outbuf_threshold; 
        else 
            no_of_samples = remaining_samples;


        if (GLverbose)
            printf ("samples = %ld, usr = %d thresh = %lx\n", no_of_samples, requested_samples, lcaio_device->outbuf_threshold);

        /* Check for DMA enabled */

        if ( lcaio_device->dma_enabled == 1 )
        {
            if (GLverbose)
	    {
                for ( sample_count = 0; sample_count < no_of_samples; sample_count++)
                    printf (" %8.8lx ", lcaio_device->write_buffer[sample_count+start_address]);
                printf ("\n");
                printf ("\nno_of_samples = %ld ( %lx)\n", no_of_samples, no_of_samples);
            }

            /* Do the DMA write operation */

            if ( no_of_samples > 4)
                if ( (rest = no_of_samples%4) ) 
                    no_of_samples = no_of_samples - rest;

            if (GLverbose)
                printf ("\nno_of_samples = %ld ( %lx)\n", no_of_samples, no_of_samples);


            write_using_dma (lcaio_device, no_of_samples, start_address);
  
            /* Wait till the DMA operation is complete */ 

            lcaio_device->procWaitQ.pid = p;
            lcaio_device->procWaitQ.nbytes = no_of_samples * sizeof(long);
            lcaio_device->procWaitQ.writeOffset = 1;
            tv.tv_sec  = lcaio_device->timeout_value;
            tv.tv_usec = 0;
            select(0, NULL, NULL, NULL, &tv);


//           interruptible_sleep_on_timeout (&lcaio_device->dma_writewq, SECS_TO_SLEEP(lcaio_device->timeout_value));

            if ( lcaio_device->dma_inprogress == 1 )                  
            {
                /* Timeout has occurred, reset the device sampling state  */

                lcaio_device->sampling_inprogress = 0; 
                lcaio_device->dma_inprogress = 0; 

                /* Set the error */

                lcaio_device->error = LCAIO_PIO_TIMEOUT;

                reply_errno(p, EIO);
            }
            else 
            {
                /*  Reset the device sampling state */

                lcaio_device->sampling_inprogress = 0; 
                lcaio_device->dma_inprogress = 0; 

            }
        }
        else
        {
            if (GLverbose)
                printf ("PIO :: samples = %ld, usr = %d thresh = %ld\n", no_of_samples, requested_samples, lcaio_device->outbuf_threshold);

            /* PIO operation */

            for ( sample_count = 0; sample_count < no_of_samples; sample_count++)
            {
                /* Write the samples to the output data buffer register */  

                GLregisters[OUTPUT_DATA_BUFFER_REG_OFFSET] =  lcaio_device->write_buffer[sample_count + start_address];

                if (GLverbose)
                    printf (" %8.8lx ", lcaio_device->write_buffer[sample_count]);

            }
        }  /* End of PIO */

        /*
         * If the output no. of samples > FIFO size then, 
         * only FIFO size is written
         */  
        remaining_samples = remaining_samples - no_of_samples;
        start_address =  start_address + no_of_samples;

        if (GLverbose)
            printf ("Write Remain samples : %ld\n",remaining_samples);


        /* If non blocked write is requested write only threshold size 
           and discard remaining data */

        if ( mode == 1 ) // non-block
        {
            /* Reset the device sampling state */

            lcaio_device->sampling_inprogress = 0;
            lcaio_device->error = LCAIO_SUCCESS;   

            return (no_of_samples * sizeof(long));  
        } 

    } while ( remaining_samples > 0 );

    /*  Reset the device sampling state */

    lcaio_device->sampling_inprogress = 0; 
    return (requested_samples);
}






void write_using_dma ( struct pmc_lcaio_device *lcaio_device, int nsamples, unsigned long start_address )
{  

    if (GLverbose)
        printf (" : lcaio <write using dma> function entry ...\n");


    /* Set the DMA operation status fields in the lcaio_device structure */

    lcaio_device->dma_inprogress = 1;
    lcaio_device->dma_func_type = DMA_WRITE;

    /* Set the DMA mode */

    GLconfRegs[DMA_CH_0_MODE_OFFSET] = NON_DEMAND_DMA_MODE; 

    /* Write the destination address, to where the data has to be copied 
       to the DMA PCI address register */

    if (GLverbose)
        printf(" buf offset = %8.8lx, start_address = %8.8lx\n", lcaio_device->writebuf_offset, start_address);


    GLconfRegs[DMA_CH_0_PCI_ADDR_OFFSET] =  lcaio_device->writebuf_offset + start_address; 

    /* Write the source address, from where the data has to be read to the DMA local address register */

    GLconfRegs[DMA_CH_0_LOCAL_ADDR_OFFSET] =  (unsigned long)(&GLregisters[OUTPUT_DATA_BUFFER_REG_OFFSET]); 

    /* Set the sample count */

    GLconfRegs[DMA_CH_0_TRANS_BYTE_CNT_OFFSET] =  nsamples;

    /* Set the DMA Direction as from Local bus to PCI bus */

    GLconfRegs[ DMA_CH_0_DESC_PTR_OFFSET] =  DMA_PCI_TO_LOCAL_BUS;

    /* Set the DMA Mode arb reg to 0 */

    GLconfRegs[DMA_MODE_ARB_REG_OFFSET] =   0;

    /* Set the DMA threshold reg to 0 */

    GLconfRegs[DMA_THRESHOLD_REG_OFFSET] =  0; 

    /* Clear the DMA Command status register */

    GLconfRegs[ DMA_CMD_STATUS_OFFSET] =  GLconfRegs[DMA_CMD_STATUS_OFFSET] & STOP_DMA_CMD_0_MASK;

    /* Start DMA operation by enabling DMA operation in DMA Command Register */

    GLconfRegs[ DMA_CMD_STATUS_OFFSET] =  GLconfRegs[DMA_CMD_STATUS_OFFSET] | START_DMA_CMD_0;

}


void read_using_dma (struct pmc_lcaio_device *lcaio_device,int nsamples)
{


    if (GLverbose)  printf (" : read_using_dma function entry ...\n");

    /* Set the mode of DMA */

    GLconfRegs[DMA_CH_0_MODE_OFFSET] =  NON_DEMAND_DMA_MODE;

    /* Write the destination address, to where the data has to be copied to 
       the DMA PCI address register */

    GLconfRegs[DMA_CH_0_PCI_ADDR_OFFSET] =  lcaio_device->readbuf_offset;

    /* Write the source address, from where the data has to be read to the 
       DMA local address register */

    GLconfRegs[DMA_CH_0_LOCAL_ADDR_OFFSET] =   (unsigned long)(&GLregisters[INPUT_DATA_BUFFER_REG_OFFSET]);

    /* Set the sample counts */

    GLconfRegs[DMA_CH_0_TRANS_BYTE_CNT_OFFSET] =  nsamples; 

    /* Set the DMA Direction as from Local bus to PCI bus */

    GLconfRegs[DMA_CH_0_DESC_PTR_OFFSET] =   DMA_LOCAL_TO_PCI_BUS;

    GLconfRegs[DMA_MODE_ARB_REG_OFFSET] =  0; 
    GLconfRegs[ DMA_THRESHOLD_REG_OFFSET] =  0;

    /* Clear the DMA Command status register */

    GLconfRegs[ DMA_CMD_STATUS_OFFSET] =  GLconfRegs[DMA_CMD_STATUS_OFFSET] & STOP_DMA_CMD_0_MASK;

    /* Start DMA operation by enabling DMA operation in DMA Command Register */

    GLconfRegs[ DMA_CMD_STATUS_OFFSET] =  GLconfRegs[DMA_CMD_STATUS_OFFSET] | START_DMA_CMD_0;

}

// mode = 1, non-block, 0 = block
int read_from_buffer (pid_t p,  struct pmc_lcaio_device *lcaio_device, int no_of_samples, int mode)
{
    int                     sample_count;
    struct _io_read_reply   reply;
    struct timeval          tv;


    /* Set the device sampling state as busy */
    lcaio_device->sampling_inprogress = 1;

    if (GLverbose)
        printf (" : read_from_buffer function entry ...\n");
    /* Check if we have sufficient data to read from the buffer i.e see if the threshold flag is set in the Input Buffer Control register */

    if ( ! GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] & IDBC_THRESHOLD_FLAG_MASK )
    {
        /* No. of samples in the input buffer is less to do a read operation */
        /* If mode of read operation is non block mode then return with error */

        if ( mode == 1 )
        {
            /* Reset the device sampling state  */
            lcaio_device->sampling_inprogress = 0; 
            reply_errno(p, EAGAIN);
            return 0;  /* Insufficient data for non block read operation */
        }

        /* Input buffer does not have sufficient data,therefore in block mode wait till there are sufficient no. of samples to read */
        /* Set the interrupt notification threshold for the input buffer to low-to-high */

        GLregisters[INT_CTRL_REG_OFFSET] =  (GLregisters[INT_CTRL_REG_OFFSET] & (~IC_IRQ1_MASK)) | (IRQ1_INPUT_BUFFER_THRESHOLD_LOW_TO_HIGH << IC_IRQ1_SHIFT);

        /* Loop to check if the process is woke up due to sufficient no. of samples or timeout condition */

        for (;;)
        {
            /* If timeout value is set  */

            if ( lcaio_device->timeout_value > 0 )
            {
                /* Sleep on the current task for a timeout value that is set 
                   or the buffer full condition that occurs first - default 
                   time is 10 seconds */

                lcaio_device->procWaitQ.pid = p;
		lcaio_device->procWaitQ.nbytes = no_of_samples * sizeof(long);
		lcaio_device->procWaitQ.writeOffset = 0;
		tv.tv_sec  = lcaio_device->timeout_value;
		tv.tv_usec = 0;
		select(0, NULL, NULL, NULL, &tv);


                //  interruptible_sleep_on_timeout (&lcaio_device->readwq, SECS_TO_SLEEP(lcaio_device->timeout_value));
            }
            else 
            {
                /* Sleep on the current task (read) till it is interrupted for buffer full */
                lcaio_device->procWaitQ.pid = p;
		lcaio_device->procWaitQ.nbytes = no_of_samples * sizeof(long);
		return(0);
            }
            /* If the threshold flag is set, break from the for (;;) loop */ 

            if ( GLregisters[INPUT_BUFFER_CTRL_REG_OFFSET] & IDBC_THRESHOLD_FLAG_MASK )
            {
                /* We have got enough number of samples to read */
                break;
            }

            /* If timeout has occurred and no sufficient data in the input 
               buffer return error */

            if ( lcaio_device->timeout_value > 0 ) 
            {
                /* Reset the device sampling state  */

                lcaio_device->sampling_inprogress = 0; 

                /* Set the error */

                lcaio_device->error = LCAIO_PIO_TIMEOUT; 

                reply_errno(p, EIO);
                return 0;
            }

        } /* for end */
    }  /* if end */

    /* Read data samples from the device FIFO */

    if (lcaio_device->dma_enabled == 1)  /* Check if DMA option is enabled */
    {
        /* DMA Read option is enabled, so start DMA read operation */

        /* Set the DMA operation status fields in the lcaio_device structure 
           as DMA is in progress */

        lcaio_device->dma_inprogress = 1;
        lcaio_device->dma_func_type = DMA_READ;

        read_using_dma (lcaio_device, MAX_BUFFER_SAMPLES);

        /* Wait till the DMA operation is complete */ 

        lcaio_device->procWaitQ.pid = p;
        lcaio_device->procWaitQ.nbytes = no_of_samples * sizeof(long);
        lcaio_device->procWaitQ.writeOffset = 0;
        tv.tv_sec  = lcaio_device->timeout_value;
        tv.tv_usec = 0;
        select(0, NULL, NULL, NULL, &tv);

//          interruptible_sleep_on_timeout (&lcaio_device->dma_readwq, SECS_TO_SLEEP(lcaio_device->timeout_value));

        if ( lcaio_device->dma_inprogress == 1 )                  
        {
            /* Timeout has occurred, reset the device sampling state  */

            if (GLverbose)
             	printf ("Timeout has occurred, reset the device sampling state\n");

            lcaio_device->sampling_inprogress = 0; 
            lcaio_device->dma_inprogress = 0; 

            /* Set the error */ 
            lcaio_device->error = LCAIO_PIO_TIMEOUT;
            reply_errno(p, EIO);
            return 0;
        }
        else 
        {
            /* DMA read operation is complete */
            lcaio_device->dma_inprogress = 0; 
            lcaio_device->sampling_inprogress = 0; 
        }
    }
    else if (lcaio_device->dma_enabled == 0)
    {
        for ( sample_count=0; sample_count< MAX_BUFFER_SAMPLES; sample_count++)
        {
            /* Read the samples from the Input data buffer register */  

            lcaio_device->read_buffer[sample_count] = GLregisters[INPUT_DATA_BUFFER_REG_OFFSET];
        }
    }

    if ( no_of_samples < MAX_BUFFER_SAMPLES )
    {

        /* Copy the user requested no. of samples to user buffer, 
           and update the buffer data start pointer accordingly and return */

        Writemsg(p, sizeof(reply)-sizeof(reply.data), lcaio_device->read_buffer, no_of_samples * sizeof (long) );

        /* Return the no. of bytes read from the buffer */

        lcaio_device->readbuf_start = no_of_samples;
        lcaio_device->readbuf_samples = MAX_BUFFER_SAMPLES - no_of_samples; 

        /* Return the no. of bytes read from the buffer */

        if (GLverbose)
            printf ("Remain samples in buffer : %lx\n",lcaio_device->readbuf_samples);

	reply.status = EOK;
	reply.zero = 0;
	reply.nbytes = no_of_samples * sizeof(long); 
	/* 
	 * This buf will contain the first part of the message 
	 * Reply to the blocked process.
	 */
	Reply(p, &reply, sizeof(reply)-sizeof(reply.data));
	return 0;

    } 
    else
    {
        /* Copy the entire data to the user buffer */

        Writemsg(p, sizeof(reply)-sizeof(reply.data), lcaio_device->read_buffer, MAX_BUFFER_SAMPLES * sizeof(long) );
        lcaio_device->readbuf_start = 0;
        lcaio_device->readbuf_samples = 0; 

        if (GLverbose)
            printf ("Remain samples in buffer : %lx\n",lcaio_device->readbuf_samples);

	reply.status = EOK;
	reply.zero = 0;
	reply.nbytes = MAX_BUFFER_SAMPLES * sizeof(long);
	/* 
	 * This buf will contain the first part of the message 
	 * Reply to the blocked process.
	 */
	Reply(p, &reply, sizeof(reply)-sizeof(reply.data));
	return 0;
    }

    /* Reset the device sampling state  */

//      lcaio_device->sampling_inprogress = 0; 

}


int lcaioRead(pid_t p, file_t *f, unsigned n)
{
    struct pmc_lcaio_device *lcaio_device;
    struct _io_read_reply   reply;
    unsigned long           samples_requested;

    if (GLverbose)
        printf (" : lcaio <read> function entry ...\n");

    lcaio_device =  f->inode->ino_data;

    samples_requested = n / sizeof (long);

    /* Check if the number of samples requested to be read is valid */

    if ( (long) samples_requested <= 0 )
    { 
        /* Invalid no. of samples to be read */  
	reply_errno(p, EINVAL);
        return 0;    
    } 

    /* If any samples are left in the buffer, copy that samples to user buffer and return    */

    if ( lcaio_device->readbuf_samples > 0 )
    {

        /* If requested no. of samples is more than that are in the buffer,  
         * read the samples that are in the buffer
         */

        if ( samples_requested > lcaio_device->readbuf_samples )
            samples_requested = lcaio_device->readbuf_samples;

	/* 
	 * Add the contents from buf->buf at tail to the waiting 
	 * processes (blocked) receive message buffer.
	 */
	Writemsg(p, sizeof(reply)-sizeof(reply.data), &lcaio_device->read_buffer[lcaio_device->readbuf_start], samples_requested*sizeof(long));

        /* Update the no. of samples in the buffer and the start pointer */

        lcaio_device->readbuf_start   += samples_requested;
        lcaio_device->readbuf_samples -= samples_requested;

        if (GLverbose)
            printf ("Remain samples in buffer : %lx\n",lcaio_device->readbuf_samples);

	reply.status = EOK;
	reply.zero = 0;
	reply.nbytes = samples_requested * sizeof (long);
	/* 
	 * This buf will contain the first part of the message 
	 * Reply to the blocked process.
	 */
	Reply(p, &reply, sizeof(reply)-sizeof(reply.data));
	return 0;
    }      

    /* If no .of samples are more than MAX_BUFFER_SAMPLES, read only
     * MAX_BUFFER_SAMPLES
     */ 

    if (samples_requested > MAX_BUFFER_SAMPLES )
    {
        /* Read one full buffer at a time */
        samples_requested = MAX_BUFFER_SAMPLES;
    }

    /* Return the count of bytes read -  read_from_buffer( ) does the actual operation of reading from the input FIFO and filling the buffer */
    /* the reply will be sent from read_from_buffer in this case */
    return ( read_from_buffer (p, lcaio_device, samples_requested, (nonblock(f)))); 

}


int create_dev(inode_t **ip)
{
    inode_t  *io;
    static   iofuns pclass;
    struct pmc_lcaio_device *lcaioDevice;
    int    ret;

    fprintf(stderr, "Call to Create/Open!\n");

    lcaioDevice = (struct pmc_lcaio_device *) malloc (sizeof(struct pmc_lcaio_device));
    if (lcaioDevice == NULL)
        return ENOMEM;


    lcaioDevice->procWaitQ.pid = 0;

    ret = verifyCard(lcaioDevice);
    if ( ret == -1)
        return ENOMEM;

    ret = qnxMapCard(lcaioDevice);
    if ( ret == -1)
        return ENOMEM;

    ret = qnxAllocBuf(lcaioDevice);
    if ( ret == -1)
        return ENOMEM;

    if (!pclass.write) {
        pclass.qioctl = lcaioIoctl;
        pclass.read   = lcaioRead;
        pclass.write   = lcaioWrite;
    }

    if ((io=inode_alloc(&pclass)) == 0) 
        return ENOMEM;


    DQIrProxyList[0].proxy = qnx_proxy_attach(0, NULL, 0, -1);
    if (GLverbose) printf("Proxy Id is %d\n", DQIrProxyList[0].proxy);
	

    if (GLverbose) printf("Installing hardware interrupt\n");

    /* Install hardware Interrupt. Share data from my_ds() */
    DQIrProxyList[0].hintid = qnx_hint_attach(0, irHandler, my_ds());

    /* Setup the proxy  callback. */
    if (server_addpid(DQIrProxyList[0].proxy, DQHintProxy) != 0)
    {
        if (GLverbose) printf("Failed adding server pid\n");
        return ENOMEM;
    }

    pclass.minor++;
    io->ino_ftype |= S_IFCHR;
    io->ino_data = lcaioDevice;
    DQIrProxyList[0].dev = lcaioDevice;
    *ip = io;

    // enable interrupts
    GLregisters[INT_CTRL_STATUS_OFFSET] = GLregisters[INT_CTRL_STATUS_OFFSET] | PCI_INT_ENABLE;

    return EOK;
}

int qnxAllocBuf(struct pmc_lcaio_device *lcaioDevice)
{
    int fd1, sel;
    void * bufp;
    struct _seginfo _seginfo;

//	lcaioDevice->read_buffer  = (unsigned long *)malloc(32767);
//	lcaioDevice->write_buffer = (unsigned long *)malloc(32767);
//	if (lcaioDevice->read_buffer == NULL || lcaioDevice->write_buffer == NULL)
//		return(-1);
//	return(0);


    if((fd1 = shm_open("Physical", O_RDWR, 0)) == -1)
        return(-1);

    if ((sel = qnx_segment_alloc_flags(0x7FFF, _PMF_DMA_SAFE)) == 0 || sel == -1)
        return(-1);

    qnx_segment_info(PROC_PID, getpid(), sel, &_seginfo);

    bufp = (caddr_t)mmap(0, 0x7FFF, PROT_READ | PROT_WRITE, MAP_SHARED, fd1, _seginfo.addr);
    if (bufp == (void *) -1)
        return(-1);

    lcaioDevice->read_buffer  = bufp;
    lcaioDevice->readbuf_offset  = _seginfo.addr;     // Physical

    if ((sel = qnx_segment_alloc_flags(0x7FFF, _PMF_DMA_SAFE)) == 0 || sel == -1)
        return(-1);

    qnx_segment_info(PROC_PID, getpid(), sel, &_seginfo);

    bufp = (caddr_t)mmap(0, 0x7FFF, PROT_READ | PROT_WRITE, MAP_SHARED, fd1, _seginfo.addr);
    if (bufp == (void *)-1)
        return(-1);

    lcaioDevice->write_buffer  = bufp;
    lcaioDevice->writebuf_offset  = _seginfo.addr;     // Physical

    close(fd1);

    return(0);      

}




int qnxMapCard(struct pmc_lcaio_device *lcaioDevice)
{
    int retCode;
    static unsigned long    physAddr;               // Physical memory address of registers.
    static unsigned long    pcrValue;               // 32 bit PCI control register values
    static unsigned long    pageOffset;             // Offset within page to registers.
    static void             *pMappedPage;   // Pointer to mapped page(s) containing registers. 
    static off_t            pageAddr;               // Address of page to be mapped.
    static size_t           memSize;                // Size of mapped memory region.
    static int              fileDesc;               // File descriptor for physical shared memory.


    pcrValue = 0;   
    retCode = _CA_PCI_Read_Config_DWord( lcaioDevice->busNum, lcaioDevice->devFuncNum, 0x18, 1, (char *)&pcrValue);
    if ( retCode != PCI_SUCCESS )
    {
        fprintf(stderr, "Error reading base I/O address from PCI control register 0x18 via _CA_PCI_Read_Config_DWord().\n");
        return ( -1 ) ;
    }

    physAddr = pcrValue & 0xfffffffc;

    // open physical memory
    fileDesc = shm_open ( "Physical", O_RDWR, 0777);
    if ( fileDesc == -1)
    {
        fprintf(stderr, "Error opening physical memory to map PCI-16LCAIO configuration registers");
        return ( -1 ) ;
    }

    // Map the register set of each card into this process space.
    // Since mmap() maps a minimum of a page when mapping physical
    // memory, first it must be determined if the 128 byte register
    // space happens to cross a page boundry.  If the space crosses
    // a page boundry then two pages must be mapped.  In this
    // case this is done by adding 128 to the offset within the
    // page and checking to see if the result is greater than
    // 0x1000 (page size).  If the result is greater, map in two
    // pages, otherwise only one page is required to map in the
    // register set.
    pageAddr   = physAddr & 0xfffff000;
    pageOffset = physAddr & 0xfff;

    memSize = 0x1000;
    if ((pageOffset + 128) >= 0x1000)
    	memSize = 0x2000;

    pMappedPage = mmap (0, memSize, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_SHARED, fileDesc, pageAddr);
    if (pMappedPage == (void *) -1)
    {
        fprintf(stderr, "Error mapping PCI-16LCAIO registers via mmap()\n");
        return ( -1 ) ;
    }
    pMappedPage = pMappedPage;

    GLregisters = (unsigned long *)((char *)pMappedPage + pageOffset);


    pcrValue = 0;   
    retCode = _CA_PCI_Read_Config_DWord( lcaioDevice->busNum, lcaioDevice->devFuncNum, 0x10, 1, (char *)&pcrValue);
    if ( retCode != PCI_SUCCESS )
    {
        fprintf(stderr, "Error reading base mem address from PCI control register 0x10 via _CA_PCI_Read_Config_DWord().\n");
        return ( -1 ) ;
    }

    physAddr = pcrValue & 0xfffffffc;

    // Map the register set of each card into this process space.
    // Since mmap() maps a minimum of a page when mapping physical
    // memory, first it must be determined if the 128 byte register
    // space happens to cross a page boundry.  If the space crosses
    // a page boundry then two pages must be mapped.  In this
    // case this is done by adding 128 to the offset within the
    // page and checking to see if the result is greater than
    // 0x1000 (page size).  If the result is greater, map in two
    // pages, otherwise only one page is required to map in the
    // register set.
    pageAddr   = physAddr & 0xfffff000;
    pageOffset = physAddr & 0xfff;

    memSize = 0x1000;
    if ((pageOffset + 128) >= 0x1000)
    	memSize = 0x2000;

    pMappedPage = mmap (0, memSize, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_SHARED, fileDesc, pageAddr);
    if (pMappedPage == (void *) -1)
    {
        fprintf(stderr, "Error mapping PCI-16LCAIO registers via mmap()\n");
        return ( -1 ) ;
    }
    pMappedPage = pMappedPage;

    GLconfRegs = (unsigned long *)((char *)pMappedPage + pageOffset);

    return(0);
}


int verifyCard(struct pmc_lcaio_device *lcaioDevice)
{
    struct _osinfo                  info;
    static unsigned                 busNum;                 // PCI bus number for found device
    static unsigned                 pciIdx;                 // PCI bus index (used to find multiple cards of the same type)
    static unsigned                 devFuncNum;             // PCI device or function ID for found
    static unsigned                 irq;  	
    static unsigned                 devId;                  // Brand specific PCI bus device ID code
    static unsigned                 subSysId;               // Brand specific PCI bus subsystem ID code
    static unsigned                 vendorId;               // Brand specific PCI bus vendor ID code
    int 	                        retCode;


    if( qnx_osinfo( 0, &info ) == -1 )   
        return(-1);

                    
    // we have a PCI BIOS
    if( !( info.sflags & _PSF_PCI_BIOS ) )
    {               
        printf("No PCI BIOS - can't access a/d card...\n");
        return( -1 );
    }

    vendorId = PCI_VENDOR_ID;
    devId    =  LCAIO_DEVICE_ID;


    // Look for a PCI-16LCAIO card on the bus.
    pciIdx = 0;
    while( 1 )
    {
        // Try to find the next device.
        retCode = _CA_PCI_Find_Device(devId, vendorId, pciIdx, &busNum, &devFuncNum);
        if ( retCode == PCI_DEVICE_NOT_FOUND )
            return( -1 );
        if ( retCode != PCI_SUCCESS )
            return( -1 );

        // Verify that the Subsystem ID matches that of a PCI-16LCAIO card.
        subSysId = 0;
        retCode = _CA_PCI_Read_Config_Word( busNum,  devFuncNum, 0x2e, 1, (char *)&subSysId);
        if ( retCode != PCI_SUCCESS )
            return( -1 );
        if (subSysId == PCI_SUBSYS_ID)
            break;

        pciIdx++;
    }

    _CA_PCI_Read_Config_DWord ( busNum, devFuncNum, LAT_GNT_INT_PIN_LINE_OFFSET << 2,   1, (char *) &irq);

    lcaioDevice->busNum              = busNum;
    lcaioDevice->devFuncNum          = devFuncNum;
    lcaioDevice->device_id           = devId;
    lcaioDevice->vendor_id           = vendorId;
    lcaioDevice->subsys_id           = subSysId;
    lcaioDevice->irq                 = irq & 0xF;
    lcaioDevice->timeout_value       = 10;
    lcaioDevice->sampling_inprogress =  0;
    lcaioDevice->readbuf_samples     =  0;

    return(0);
}


struct lcaio_pci_config_regs 
read_pci_config_regs (struct pmc_lcaio_device *lcaioDevice)
{
    struct lcaio_pci_config_regs pci_config_regs;

    /* Read all the pci configuration registers using the pcibios calls */

    pci_config_regs.device_vendor_ID = 0x00;
	
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, DEVICE_VENDOR_ID_OFFSET,             1, (char *) &pci_config_regs.device_vendor_ID);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, STATUS_COMMAND_OFFSET << 2,               1, (char *) &pci_config_regs.status_command);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, CLASS_CODE_REVISION_ID_OFFSET << 2,       1, (char *) &pci_config_regs.classcode_revision_ID);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, BIST_HDR_TYPE_LAT_CACHE_SIZE_OFFSET << 2, 1, (char *) &pci_config_regs.BIST_hdrtype_lattimer_cacheline_size);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, PCI_MEM_BASE_ADDR_OFFSET << 2,       1, (char *) &pci_config_regs.runtime_register_addr);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, PCI_IO_BASE_ADDR_OFFSET << 2,             1, (char *) &pci_config_regs.config_register_addr);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, PCI_BASE_ADDR_0_OFFSET << 2,              1, (char *) &pci_config_regs.PCI_base_addr2);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, PCI_BASE_ADDR_1_OFFSET << 2,              1, (char *) &pci_config_regs.PCI_base_addr3);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, CARDBUS_CIS_PTR_OFFSET << 2,              1, (char *) &pci_config_regs.cardbus_CIS_ptr);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, SUBSYS_ID_VENDOR_ID_OFFSET << 2,     1, (char *) &pci_config_regs.subsystem_vendor_ID);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, PCI_BASE_ADDR_LOC_ROM_OFFSET << 2,        1, (char *) &pci_config_regs.PCI_ROM_address);
    _CA_PCI_Read_Config_DWord ( lcaioDevice->busNum, lcaioDevice->devFuncNum, LAT_GNT_INT_PIN_LINE_OFFSET << 2,         1, (char *) &pci_config_regs.maxlat_mingnt_intpin_intline);

    return (pci_config_regs);
}

struct lcaio_local_config_regs 
read_local_config_regs (struct pmc_lcaio_device *lcaio_device)
{
    int reg;
    struct lcaio_local_config_regs local_config_regs;

    /* Read all the local configuration registers */

    local_config_regs.pci_local_range0 = GLconfRegs[PCI_TO_LOC_ADDR_0_RNG_OFFSET]; 
    local_config_regs.pci_local_remap0= GLconfRegs[LOC_BASE_ADDR_REMAP_0_OFFSET];
    local_config_regs.mode_arbitration = GLconfRegs[MODE_ARBITRATION_OFFSET];
    local_config_regs.endian_setting = GLconfRegs[BIG_LITTLE_ENDIAN_DESC_OFFSET];
    local_config_regs.pci_local_exp_ROM_range = GLconfRegs[PCI_TO_LOC_ROM_RNG_OFFSET];
    local_config_regs.pci_local_exp_ROM_remap = GLconfRegs[LOC_BASE_ADDR_REMAP_EXP_ROM_OFFSET];
    local_config_regs.pci_local_bus_region0 =   GLconfRegs[BUS_REG_DESC_0_FOR_PCI_LOC_OFFSET];
    local_config_regs.local_pci_range = GLconfRegs[DIR_MASTER_TO_PCI_RNG_OFFSET];
    local_config_regs.local_pci_membase_addr =  GLconfRegs[LOC_ADDR_FOR_DIR_MASTER_MEM_OFFSET];
    local_config_regs.local_pci_IObase_addr =   GLconfRegs[LOC_ADDR_FOR_DIR_MASTER_IO_OFFSET];
    local_config_regs.local_pci_remap = GLconfRegs[PCI_ADDR_REMAP_DIR_MASTER_OFFSET];
    local_config_regs.local_pci_config = GLconfRegs[PCI_CFG_ADDR_DIR_MASTER_IO_OFFSET];

    /* Get the shared runtime register contents */

    for ( reg = 0; reg < 8; reg++)
    {
        local_config_regs.mailbox[reg] = GLconfRegs[MAILBOX_REGISTER_0_OFFSET + reg];
    }

    local_config_regs.pci_local_doorbell = GLconfRegs[PCI_TO_LOC_DOORBELL_OFFSET];
    local_config_regs.local_pci_doorbell = GLconfRegs[LOC_TO_PCI_DOORBELL_OFFSET];
    local_config_regs.interrupt_control_status = GLconfRegs[INT_CTRL_STATUS_OFFSET];
    local_config_regs.runtime_control = GLconfRegs[PROM_CTRL_CMD_CODES_CTRL_OFFSET];
    local_config_regs.device_vendor_ID = GLconfRegs[DEVICE_ID_VENDOR_ID_OFFSET];
    local_config_regs.revision_ID = GLconfRegs[REVISION_ID_OFFSET];
    local_config_regs.mailbox_reg0 = GLconfRegs[MAILBOX_REG_0_OFFSET];
    local_config_regs.mailbox_reg1 = GLconfRegs[MAILBOX_REG_0_OFFSET];

    /* Get the DMA registers contents */

    local_config_regs.DMA_mode0 = GLconfRegs[DMA_CH_0_MODE_OFFSET];
    local_config_regs.DMA_PCI_address0 = GLconfRegs[DMA_CH_0_PCI_ADDR_OFFSET];
    local_config_regs.DMA_local_address0 = GLconfRegs[DMA_CH_0_LOCAL_ADDR_OFFSET];
    local_config_regs.DMA_byte_count0 = GLconfRegs[DMA_CH_0_TRANS_BYTE_CNT_OFFSET];
    local_config_regs.DMA_descriptor_ptr0 = GLconfRegs[DMA_CH_0_DESC_PTR_OFFSET];
    local_config_regs.DMA_mode1 = GLconfRegs[DMA_CH_1_MODE_OFFSET];
    local_config_regs.DMA_PCI_address1 = GLconfRegs[DMA_CH_1_PCI_ADDR_OFFSET];
    local_config_regs.DMA_local_address1 = GLconfRegs[DMA_CH_1_LOCAL_ADDR_OFFSET];
    local_config_regs.DMA_byte_count1 = GLconfRegs[DMA_CH_1_TRANS_BYTE_CNT_OFFSET];
    local_config_regs.DMA_descriptor_ptr1 = GLconfRegs[DMA_CH_1_DESC_PTR_OFFSET];
    local_config_regs.DMA_command_status = GLconfRegs[DMA_CMD_STATUS_OFFSET];
    local_config_regs.DMA_arbitration = GLconfRegs[DMA_MODE_ARB_REG_OFFSET];
    local_config_regs.DMA_threshold = GLconfRegs[DMA_THRESHOLD_REG_OFFSET];

    /* Get the messaging queue register contents */

    local_config_regs.msg_unit_config = GLconfRegs[MSG_UNIT_CONFIG_OFFSET];
    local_config_regs.queue_baseaddr = GLconfRegs[Q_BASE_ADDR_OFFSET];
    local_config_regs.inbound_free_headptr = GLconfRegs[IN_FREE_HEAD_PTR_OFFSET];
    local_config_regs.inbound_free_tailptr = GLconfRegs[IN_FREE_TAIL_PTR_OFFSET];
    local_config_regs.inbound_post_headptr = GLconfRegs[IN_POST_HEAD_PTR_OFFSET];
    local_config_regs.inbound_post_tailptr = GLconfRegs[IN_POST_TAIL_PTR_OFFSET];
    local_config_regs.outbound_free_headptr = GLconfRegs[OUT_FREE_HEAD_PTR_OFFSET];
    local_config_regs.outbound_free_tailptr = GLconfRegs[OUT_FREE_TAIL_PTR_OFFSET];
    local_config_regs.outbound_post_headptr = GLconfRegs[OUT_POST_HEAD_PTR_OFFSET];
    local_config_regs.outbound_post_tailptr = GLconfRegs[OUT_POST_TAIL_PTR_OFFSET];
    local_config_regs.queue_status_control = GLconfRegs[Q_STATUS_CTRL_REG_OFFSET];
    local_config_regs.pci_local_range1 = GLconfRegs[PCI_TO_LOC_ADDR_1_RNG_OFFSET];
    local_config_regs.pci_local_remap1 = GLconfRegs[LOC_BASE_ADDR_REMAP_1_OFFSET];
    local_config_regs.pci_local_bus_region1 = GLconfRegs[BUS_REG_DESC_1_FOR_PCI_LOC_OFFSET];
    return (local_config_regs);		
}


// NO printf's in this code, bad things happen.. :)

pid_t far irHandler(void)
{
    int int_condition, irq0_condition, irq1_condition, irq2_condition;
    pid_t retProxy;
    struct pmc_lcaio_device         *lcaio_device;

    retProxy = (pid_t far) 0;
    int_condition = GLregisters[ INT_CTRL_REG_OFFSET] & IC_INT_OCC_MASK;

    lcaio_device = DQIrProxyList[0].dev;

    if ( lcaio_device->dma_inprogress == 1 )
    {
        /* Check if the DMA_CMD_STAT_DONE flag is set in the  DMA_CMD_STATUS Register */

        if ( GLconfRegs[DMA_CMD_STATUS_OFFSET] & DMA_CMD_STAT_DONE )
        {

            /* Clear the DMA interrupt by clearing DMA Command Status Register */
  
            GLconfRegs[DMA_CMD_STATUS_OFFSET] = GLconfRegs[DMA_CMD_STATUS_OFFSET] | DMA_CMD_STAT_INT_CLEAR;
                                                           
            lcaio_device->dma_inprogress = 0;

            /* Wake up the process that is waiting on the DMA queue depending on the DMA operation type */

            return DQIrProxyList[0].proxy;
        }
    }


    if ( GLregisters[INT_CTRL_REG_OFFSET]  & IC_IRQ0_REQUEST_MASK)
    {
        /* Clear the interrupt request by clearing in the Interrupt ControlRegister*/

        GLregisters[INT_CTRL_REG_OFFSET] = GLregisters[INT_CTRL_REG_OFFSET] & ~IC_IRQ0_REQUEST_MASK;

        irq0_condition = GLregisters[ INT_CTRL_REG_OFFSET];

        /* Identify the interrupt condition */    

        if ( irq0_condition == IC_IDLE_VAL )   
        {

	    retProxy = DQIrProxyList[0].proxy;
            lcaio_device->initialize = 0;

        }

        /* Auto calibration Interrupt condition */    

        if ( irq0_condition == IC_AUTO_CAL_COMP_VAL )   
        {
            /* Clear the auto calibration flag in the lcaio device structure */

	    retProxy = DQIrProxyList[0].proxy;
            lcaio_device->autocalibration = 0;
        }

	if ( irq0_condition == IC_AUX_LOW_TO_HIGH_VAL )   
        {
            /* Set the aux input low to high flag in the lcaio device structure */
            lcaio_device->aux_low_to_high = 0;
	    retProxy = DQIrProxyList[0].proxy;
        }

        if ( irq0_condition == IC_AUX_HIGH_TO_LOW_VAL )   
        {
            /* Set the aux input high to low flag in the lcaio device structure */
            lcaio_device->aux_high_to_low = 0;
	    retProxy = DQIrProxyList[0].proxy;
	}
    }

    if ( GLregisters[INT_CTRL_REG_OFFSET]  & IC_IRQ1_REQUEST_MASK)
    {
        /* Clear the interrupt request by clearing in the Interrupt ControlRegister*/

        GLregisters[INT_CTRL_REG_OFFSET] = GLregisters[INT_CTRL_REG_OFFSET] & ~IC_IRQ1_REQUEST_MASK;

        irq1_condition = GLregisters[ INT_CTRL_REG_OFFSET];
        if ( irq1_condition == IC_INP_THRES_LOW_TO_HIGH_VAL  )   
        {
            /* Set the input buffer low to high flag in the lcaio device structure */
            lcaio_device->inbuf_thres_lowtohigh = 0;
	    retProxy = DQIrProxyList[0].proxy;
	}
        if ( irq1_condition == IC_INP_THRES_HIGH_TO_LOW_VAL  )   
        {
            /* Set the input buffer high to low flag in the lcaio device structure */
            lcaio_device->inbuf_thres_hightolow = 0;
	    retProxy = DQIrProxyList[0].proxy;
        }    
    }

    if ( GLregisters[INT_CTRL_REG_OFFSET]  & IC_IRQ2_REQUEST_MASK)
    {
        /* Clear the interrupt request by clearing in the Interrupt ControlRegister*/

        GLregisters[INT_CTRL_REG_OFFSET] = GLregisters[INT_CTRL_REG_OFFSET] & ~IC_IRQ2_REQUEST_MASK;

        irq2_condition = GLregisters[ INT_CTRL_REG_OFFSET];

        /* Output buffer threshold low to high Interrupt condition */    

        if ( irq2_condition == IC_OUT_THRES_LOW_TO_HIGH_VAL  )   
        {
            /* Set the Output buffer low to high flag in the lcaio device structure */

            lcaio_device->outbuf_thres_lowtohigh = 0;
	    retProxy = DQIrProxyList[0].proxy;
        }
        if ( irq2_condition == IC_OUT_THRES_HIGH_TO_LOW_VAL  )   
        {
            /* Set the output buffer high to low flag in the lcaio device structure */

            lcaio_device->outbuf_thres_hightolow = 0;
	    retProxy = DQIrProxyList[0].proxy;
	}
        /* Output burst complete Interrupt condition */    
        if ( irq2_condition == IC_OUT_BURST_COMP_VAL  )   
        {
            /* Set the output buffer high to low flag in the lcaio device structure */
            lcaio_device->output_burst_complete = 0;
        }
    }
    /* returning this proxy will cause the approriate process to be continue */
    return(retProxy);	
}

int DQHintProxy(pid_t p)
{
    struct _io_qioctl_reply         reply;
    pid_t  replyProc;
    struct pmc_lcaio_device *dev;

    if (GLverbose)
        printf ("HINT PROXY CALLED\n"); 

    if (DQIrProxyList[0].proxy != p)
        return 0;

    dev = DQIrProxyList[0].dev;

    if (GLverbose)
        printf ("dev ptr is %p\n", dev); 

    if (dev->sampling_inprogress == 1)
    {
	if (dev->procWaitQ.writeOffset == 1)
            write_to_buffer(dev->procWaitQ.pid,  dev , dev->procWaitQ.nbytes, 0);
	else
            read_from_buffer(dev->procWaitQ.pid,  dev , dev->procWaitQ.nbytes, 0);
	return(1);
    }

    replyProc = dev->procWaitQ.pid;

    if (GLverbose)
        printf ("proxy reply to %ld\n", replyProc); 

    if (replyProc == 0)
	return(0);

    reply.status = EOK;
    reply.ret_val = 0;

    Reply(replyProc, &reply, sizeof(reply));

    return(1);
}

