/* File Name:
 *		LCAIOdriver.h
 *
 * Contents:
 *		Constants, structures, and function
 *		declarations specific to this driver.
 *
 * Revision History : LINUX_2_4 switch is added to be compatible with 2.4 
 *                    linux  kernel.
 *                    waitqueues added for 2.4 linux kernel compatibility.
 */

#ifndef __LCAIO_DRIVER__
#define __LCAIO_DRIVER__

/*
 * Header files
 */

#include <sys/io_msg.h>
#include <sys/kernel.h>
#include <sys/uio.h>
#include <sys/inline.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include "iomgr.h"



#define DEBUG_SIZE 	     16
#define MAX_BUFFER_SAMPLES   0x1FFF

#define N0_BLOCK		  0
#define DMA_READ		  0
#define DMA_WRITE		  1

#define INITIALIZE_TIMEOUT_MS    2000
#define AUTO_CAL_TIMEOUT_MS 	 2000

#define MSECS_TO_SLEEP(MS)       ((MS * HZ)/1000) 
#define SECS_TO_SLEEP(S)         (S * HZ) 


pid_t far irHandler( void );
int DQHintProxy(pid_t p);


struct dqQEntry
{
        void      *data;
        struct dqQEntry *next;
        struct dqQEntry *prev;
};

struct dqQ 
{
        struct dqQEntry *first;
        struct dqQEntry *last;
        int       count;
};

struct dqProcWait 
{
        pid_t   pid;
        long    nbytes;
        long    writeOffset;
};




/*
 * Description:
 *		Driver-defined structure used to hold 
 *		miscellaneous device information.
 */
struct pmc_lcaio_device
{
//      struct pci_dev 	 *pdev;		/* structure for the pcidevice info */

      unsigned int        busNum;
      unsigned int        devFuncNum;

      unsigned int	 device_id; 	/* Device id 			    */
      unsigned int	 vendor_id; 	/* Vendor id 			    */
      unsigned int       subsys_id;	/* Subsystem id of the card found   */
      unsigned long      board_id;  	/* Board id : lcaio-16 or lcaio-12  */
      unsigned long      major;         /* Major number for the driver      */
      unsigned long 	 minor; 	/* Minor number for the device      */
      unsigned int       irq;		/* Irq no. allocated for the device */
      unsigned long	 busy;		/* Device state 	 	    */

      unsigned long 	*read_buffer;   /* Buffer for dmasamples read       */
      unsigned long 	*write_buffer;  /* Buffer for dmasamples to write   */
      unsigned long	readbuf_offset; /* Read Dmabuffer Offset address    */
      unsigned long	writebuf_offset;/* Write Dmabuffer Offset address   */
      unsigned char     dma_enabled;	/* Dma option - enabled or disabled */
      unsigned char	dma_inprogress;	/* Dma operation status		    */
      unsigned char 	dma_func_type;	/* Dma either read or write	    */

      unsigned long     readbuf_samples;  /* Read buffer samples 	    */
      unsigned long     writebuf_samples; /* Write buffer samples           */
      unsigned long     readbuf_start;    /* start pointer for read buffer  */
      unsigned long     inbuf_threshold;  /* threshold value of read buffer */
      unsigned long     outbuf_threshold; /* threshold value of write buffer*/
     
      unsigned char 	sampling_inprogress; /* device sampling state        */

#ifdef XXXX
      wait_queue_head_t readwq;
      wait_queue_head_t writewq;
      wait_queue_head_t dma_readwq;
      wait_queue_head_t dma_writewq;
      wait_queue_head_t ioctlwq;

      struct wait_queue *readwq;	/* Wait queue for PIO read operation  */
      struct wait_queue *writewq;	/* Wait queue for PIO write operation */
      struct wait_queue *dma_readwq;	/* Wait queue for dma read operation  */
      struct wait_queue *dma_writewq;	/* Wait queue for dma write operation */
      struct wait_queue *ioctlwq;	/* Wait queue for PIO ioctl operation */
#endif

//      dpQ  *ioctlwq;	/* Wait queue for PIO ioctl operation */
//      struct  dqQ       inputQ;

	struct dqProcWait  procWaitQ;


      long              timeout_value;	/* Timeout value		    */
      unsigned long	error;		/* Recent error type occured	    */
      unsigned long     debug_buffer[DEBUG_SIZE]; /* Debug messages buffer  */

      unsigned long	initialize;	 /* Initialize status 		    */
      unsigned long	autocalibration; /* Autocailbration status 	    */
      unsigned long	aux_low_to_high; /* Aux input low to high setting   */
      unsigned long	aux_high_to_low; /* Aux input high to low setting   */
      unsigned long	inbuf_thres_lowtohigh; /* Input buffer threshold 
                                                  low to high 	            */
      unsigned long	inbuf_thres_hightolow; /* Input buffer threshold 
                                                  high to low               */
      unsigned long	outbuf_thres_lowtohigh;/* output buffer threshold 
           				          low to high               */
      unsigned long	outbuf_thres_hightolow;/* output buffer threshold 
                                                  high to low               */
      unsigned long	output_burst_complete; /* output burst complete     */ 

      unsigned long     irq0_int_req;    /* Interrupt notification for irq0 */  
      unsigned long     irq1_int_req;    /* Interrupt notification for irq1 */  
      unsigned long     irq2_int_req;    /* Interrupt notification for irq2 */  
   
//      spinlock_t        lock;            /* lock for critical section in ISR*/

      unsigned long	*config_reg_address; /* configuration register 
                                               space base address           */ 	
      unsigned long 	*local_reg_address;  /* Card registers base address  */ 
       
      struct pmc_lcaio_device *next;        /* linked list pointer          */

};


struct dqIrProxy
{
        pid_t   proxy;
        struct  pmc_lcaio_device   *dev;
        int     hintid;
        
};



/* 
 * Following registers are accessed from the "local_reg_address" offset.
 */

struct lcaio_card_registers
{
  unsigned long board_control_reg;
  unsigned long int_control_reg;
  unsigned long input_data_buffer_reg;
  unsigned long input_buffer_control_reg;
  unsigned long rate_A_gen_reg;
  unsigned long rate_B_gen_reg;
  unsigned long output_data_buffer_reg;
  unsigned long output_buffer_ctrl_reg;
  unsigned long scan_sync_ctrl_reg;
  unsigned long digital_io_port_reg;
};


/*
 * Prototypes for globally defined functions...
 */

int 
create_dev (inode_t **ip);

int 
lcaio_release (inode_t *,struct file *);

ssize_t 
lcaio_read (struct file *,unsigned long *,size_t, unsigned long *);

ssize_t 
lcaio_write (struct file *,const unsigned long *,size_t, unsigned long *);

int 
lcaio_ioctl (inode_t *,struct file *, unsigned int, unsigned long);


static void
lcaio_irq_handler ( int, void *, struct pt_regs *);

int 
read_from_buffer (pid_t, struct pmc_lcaio_device *, int, int);

void 
read_using_dma ( struct pmc_lcaio_device *, int);

int 
write_to_buffer (pid_t, struct pmc_lcaio_device *, int, int);

void 
write_using_dma (struct pmc_lcaio_device *, int, unsigned long);

struct lcaio_pci_config_regs 
read_pci_config_regs (struct pmc_lcaio_device *);

struct lcaio_local_config_regs 
read_local_config_regs (struct pmc_lcaio_device *);

#endif	// __LCAIO_DRIVER__
