/* File Name:
 *		LCAIOIoctl.h
 *
 * Contents:
 *		Definitions of IOCTL codes and data.
 */
#include <stdio.h>
#include <math.h>
#include <ioctl.h>

#ifndef __LCAIO_IOCTL__
#define __LCAIO_IOCTL__

#define LCAIO_DEVICE_ID   0x9080
#define LCAIO_VENDOR_ID   0x10B5


//#define LCAIO12_SUBSYS_ID 0x00010000
//#define LCAIO16_SUBSYS_ID 0x00020000

#define LCAIO12_SUBSYS_ID 0x240110b5    // Modified
#define LCAIO16_SUBSYS_ID 0x240210b5    // Modified

#define LCAIO12		  0
#define LCAIO16		  1

#define PCI_INT_ENABLE	0x00050900

/* ******************   IOCTL Commands ********************* */       

#define IOCTL_LCAIO	's'	

#define IOCTL_LCAIO_NO_COMMAND                1
#define IOCTL_LCAIO_READ_REGISTER             2
#define IOCTL_LCAIO_WRITE_REGISTER            3
#define IOCTL_LCAIO_REQ_INT_NOTIFY            4
#define IOCTL_LCAIO_SET_INPUT_MODE            5
#define IOCTL_LCAIO_SET_RANGE		      6
#define IOCTL_LCAIO_SET_DATA_FORMAT           7
#define IOCTL_LCAIO_SET_OUTPUT_MODE	      8
#define IOCTL_LCAIO_SET_INPUT_SYNC	      9
#define IOCTL_LCAIO_SET_OUTPUT_SYNC	      10
#define IOCTL_LCAIO_AUTO_CAL                  11
#define IOCTL_LCAIO_INITIALIZE                12
#define IOCTL_LCAIO_SET_INPUT_BUFFER_THRESHOLD 13
#define IOCTL_LCAIO_CLEAR_INPUT_BUFFER         14
#define IOCTL_LCAIO_SET_OUTPUT_BUFFER_THRESHOLD 15
#define IOCTL_LCAIO_CLEAR_OUTPUT_BUFFER       16
#define IOCTL_LCAIO_SET_GEN_A_RATE            17
#define IOCTL_LCAIO_SET_GEN_B_RATE            18
#define IOCTL_LCAIO_SET_GEN_A_ENABLE	      19
#define IOCTL_LCAIO_SET_GEN_B_ENABLE	      20
#define IOCTL_LCAIO_SET_GEN_B_SRC	      21
#define IOCTL_LCAIO_SET_INPUT_SCAN	      22
#define IOCTL_LCAIO_SET_OUTPUT_SCAN	      23
#define IOCTL_LCAIO_SET_DIGITAL_IO_DIR        24
#define IOCTL_LCAIO_READ_DIGITAL_IO	      25
#define IOCTL_LCAIO_SET_DIGITAL_IO	      26
#define IOCTL_LCAIO_READ_AUX_INPUT            27
#define IOCTL_LCAIO_SET_AUX_OUTPUT            28
#define IOCTL_LCAIO_GET_DEVICE_ERROR          29
#define IOCTL_LCAIO_READ_PCI_CONFIG           30
#define IOCTL_LCAIO_READ_LOCAL_CONFIG         31
#define IOCTL_LCAIO_WRITE_PCI_CONFIG_REG      32
#define IOCTL_LCAIO_WRITE_LOCAL_CONFIG_REG    33
#define IOCTL_LCAIO_SET_TIMEOUT               34
#define IOCTL_LCAIO_SET_DMA_ENABLE            35
#define IOCTL_LCAIO_READ_DEBUG		      36

/*
 * Structures used by the IOCTL commands 
 */

/*
 * IOCTL_LCAIO_NO_COMMAND
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_READ_REGISTER          &
 * IOCTL_LCAIO_WRITE_REGISTER         &
 * IOCTL_LCAIO_WRITE_PCI_CONFIG_REG   &
 * IOCTL_LCAIO_WRITE_LOCAL_CONFIG_REG
 *
 * Parameter = lcaio_registers *registers;
 */

/*** LCAIO Registers ***/

#define BOARD_CTRL_REG_OFFSET                0
#define INT_CTRL_REG_OFFSET                  1
#define INPUT_DATA_BUFFER_REG_OFFSET         2
#define INPUT_BUFFER_CTRL_REG_OFFSET         3
#define RATE_A_GEN_REG_OFFSET                4
#define RATE_B_GEN_REG_OFFSET                5
#define OUTPUT_DATA_BUFFER_REG_OFFSET        6
#define OUTPUT_BUFFER_CTRL_REG_OFFSET        7
#define SCAN_SYNC_CTRL_REG_OFFSET            8
#define DIGITAL_IO_PORT_REG_OFFSET           9


#define BOARD_CTRL_REG              (lcaio_device->local_reg_address + BOARD_CTRL_REG_OFFSET)  
#define INT_CTRL_REG                (lcaio_device->local_reg_address + INT_CTRL_REG_OFFSET)   
#define INPUT_DATA_BUFFER_REG       (lcaio_device->local_reg_address + INPUT_DATA_BUFFER_REG_OFFSET)   
#define INPUT_BUFFER_CTRL_REG       (lcaio_device->local_reg_address + INPUT_BUFFER_CTRL_REG_OFFSET)   
#define RATE_A_GEN_REG              (lcaio_device->local_reg_address + RATE_A_GEN_REG_OFFSET)    
#define RATE_B_GEN_REG              (lcaio_device->local_reg_address + RATE_B_GEN_REG_OFFSET)    
#define OUTPUT_DATA_BUFFER_REG      (lcaio_device->local_reg_address + OUTPUT_DATA_BUFFER_REG_OFFSET)  
#define OUTPUT_BUFFER_CTRL_REG      (lcaio_device->local_reg_address + OUTPUT_BUFFER_CTRL_REG_OFFSET)  
#define SCAN_SYNC_CTRL_REG          (lcaio_device->local_reg_address + SCAN_SYNC_CTRL_REG_OFFSET)   
#define DIGITAL_IO_PORT_REG         (lcaio_device->local_reg_address + DIGITAL_IO_PORT_REG_OFFSET)    

/*** DMA Registers ***/

#define DMA_CH_0_MODE_OFFSET                32
#define DMA_CH_0_PCI_ADDR_OFFSET            33
#define DMA_CH_0_LOCAL_ADDR_OFFSET          34
#define DMA_CH_0_TRANS_BYTE_CNT_OFFSET      35
#define DMA_CH_0_DESC_PTR_OFFSET            36
#define DMA_CH_1_MODE_OFFSET                37
#define DMA_CH_1_PCI_ADDR_OFFSET            38
#define DMA_CH_1_LOCAL_ADDR_OFFSET          39
#define DMA_CH_1_TRANS_BYTE_CNT_OFFSET      40
#define DMA_CH_1_DESC_PTR_OFFSET            41
#define DMA_CMD_STATUS_OFFSET               42
#define DMA_MODE_ARB_REG_OFFSET             43
#define DMA_THRESHOLD_REG_OFFSET            44

#define DMA_CH_0_MODE 		     (lcaio_device->config_reg_address + DMA_CH_0_MODE_OFFSET)  
#define DMA_CH_0_PCI_ADDR            (lcaio_device->config_reg_address + DMA_CH_0_PCI_ADDR_OFFSET)
#define DMA_CH_0_LOCAL_ADDR          (lcaio_device->config_reg_address + DMA_CH_0_LOCAL_ADDR_OFFSET)
#define DMA_CH_0_TRANS_BYTE_CNT      (lcaio_device->config_reg_address + DMA_CH_0_TRANS_BYTE_CNT_OFFSET)
#define DMA_CH_0_DESC_PTR            (lcaio_device->config_reg_address + DMA_CH_0_DESC_PTR_OFFSET)
#define DMA_CH_1_MODE                (lcaio_device->config_reg_address + DMA_CH_1_MODE_OFFSET)
#define DMA_CH_1_PCI_ADDR            (lcaio_device->config_reg_address + DMA_CH_1_PCI_ADDR_OFFSET)
#define DMA_CH_1_LOCAL_ADDR          (lcaio_device->config_reg_address + DMA_CH_1_LOCAL_ADDR_OFFSET)
#define DMA_CH_1_TRANS_BYTE_CNT      (lcaio_device->config_reg_address + DMA_CH_1_TRANS_BYTE_CNT_OFFSET)
#define DMA_CH_1_DESC_PTR            (lcaio_device->config_reg_address + DMA_CH_1_DESC_PTR_OFFSET)
#define DMA_CMD_STATUS               (lcaio_device->config_reg_address + DMA_CMD_STATUS_OFFSET)
#define DMA_MODE_ARB_REG             (lcaio_device->config_reg_address + DMA_MODE_ARB_REG_OFFSET)
#define DMA_THRESHOLD_REG            (lcaio_device->config_reg_address + DMA_THRESHOLD_REG_OFFSET )

/*** PCI Configuration Registers ***/

#define DEVICE_VENDOR_ID_OFFSET              0
#define STATUS_COMMAND_OFFSET                1
#define CLASS_CODE_REVISION_ID_OFFSET        2
#define BIST_HDR_TYPE_LAT_CACHE_SIZE_OFFSET  3
#define PCI_MEM_BASE_ADDR_OFFSET             4
#define PCI_IO_BASE_ADDR_OFFSET              5
#define PCI_BASE_ADDR_0_OFFSET               6
#define PCI_BASE_ADDR_1_OFFSET               7
#define CARDBUS_CIS_PTR_OFFSET              10
#define SUBSYS_ID_VENDOR_ID_OFFSET          11
#define PCI_BASE_ADDR_LOC_ROM_OFFSET        12
#define LAT_GNT_INT_PIN_LINE_OFFSET         15

#define DEVICE_VENDOR_ID              (lcaio_device->config_reg_address + DEVICE_VENDOR_ID_OFFSET)
#define STATUS_COMMAND                (lcaio_device->config_reg_address + STATUS_COMMAND_OFFSET)
#define CLASS_CODE_REVISION_ID        (lcaio_device->config_reg_address + CLASS_CODE_REVISION_ID_OFFSET)
#define BIST_HDR_TYPE_LAT_CACHE_SIZE  (lcaio_device->config_reg_address + BIST_HDR_TYPE_LAT_CACHE_SIZE_OFFSET)
#define PCI_MEM_BASE_ADDR             (lcaio_device->config_reg_address + PCI_IO_BASE_ADDR_OFFSET)
#define PCI_IO_BASE_ADDR              (lcaio_device->config_reg_address + PCI_BASE_ADDR_0_OFFSET )
#define PCI_BASE_ADDR_0               (lcaio_device->config_reg_address + PCI_BASE_ADDR_1_OFFSET)
#define PCI_BASE_ADDR_1               (lcaio_device->config_reg_address + PCI_BASE_ADDR_1_OFFSET)
#define CARDBUS_CIS_PTR               (lcaio_device->config_reg_address + CARDBUS_CIS_PTR_OFFSET)
#define SUBSYS_ID_VENDOR_ID           (lcaio_device->config_reg_address + SUBSYS_ID_VENDOR_ID_OFFSET)
#define PCI_BASE_ADDR_LOC_ROM         (lcaio_device->config_reg_address + PCI_BASE_ADDR_LOC_ROM_OFFSET)
#define LAT_GNT_INT_PIN_LINE          (lcaio_device->config_reg_address + LAT_GNT_INT_PIN_LINE_OFFSET )

/*** Local Configuration Registers. ***/

#define PCI_TO_LOC_ADDR_0_RNG_OFFSET         0
#define LOC_BASE_ADDR_REMAP_0_OFFSET         1
#define MODE_ARBITRATION_OFFSET              2
#define BIG_LITTLE_ENDIAN_DESC_OFFSET        3
#define PCI_TO_LOC_ROM_RNG_OFFSET            4
#define LOC_BASE_ADDR_REMAP_EXP_ROM_OFFSET   5
#define BUS_REG_DESC_0_FOR_PCI_LOC_OFFSET    6
#define DIR_MASTER_TO_PCI_RNG_OFFSET         7
#define LOC_ADDR_FOR_DIR_MASTER_MEM_OFFSET   8
#define LOC_ADDR_FOR_DIR_MASTER_IO_OFFSET    9
#define PCI_ADDR_REMAP_DIR_MASTER_OFFSET    10
#define PCI_CFG_ADDR_DIR_MASTER_IO_OFFSET   11
#define PCI_TO_LOC_ADDR_1_RNG_OFFSET        92
#define LOC_BASE_ADDR_REMAP_1_OFFSET        93
#define BUS_REG_DESC_1_FOR_PCI_LOC_OFFSET   94

#define PCI_TO_LOC_ADDR_0_RNG         (lcaio_device->config_reg_address + PCI_TO_LOC_ADDR_0_RNG_OFFSET )
#define LOC_BASE_ADDR_REMAP_0         (lcaio_device->config_reg_address + LOC_BASE_ADDR_REMAP_0_OFFSET)
#define MODE_ARBITRATION              (lcaio_device->config_reg_address + MODE_ARBITRATION_OFFSET)
#define BIG_LITTLE_ENDIAN_DESC        (lcaio_device->config_reg_address + BIG_LITTLE_ENDIAN_DESC_OFFSET)
#define PCI_TO_LOC_ROM_RNG            (lcaio_device->config_reg_address + PCI_TO_LOC_ROM_RNG_OFFSET)
#define LOC_BASE_ADDR_REMAP_EXP_ROM   (lcaio_device->config_reg_address + LOC_BASE_ADDR_REMAP_EXP_ROM_OFFSET)
#define BUS_REG_DESC_0_FOR_PCI_LOC    (lcaio_device->config_reg_address + BUS_REG_DESC_0_FOR_PCI_LOC_OFFSET)
#define DIR_MASTER_TO_PCI_RNG         (lcaio_device->config_reg_address + DIR_MASTER_TO_PCI_RNG_OFFSET)
#define LOC_ADDR_FOR_DIR_MASTER_MEM   (lcaio_device->config_reg_address + LOC_ADDR_FOR_DIR_MASTER_MEM_OFFSET)
#define LOC_ADDR_FOR_DIR_MASTER_IO    (lcaio_device->config_reg_address + LOC_ADDR_FOR_DIR_MASTER_IO_OFFSET)
#define PCI_ADDR_REMAP_DIR_MASTER     (lcaio_device->config_reg_address + PCI_ADDR_REMAP_DIR_MASTER_OFFSET)
#define PCI_CFG_ADDR_DIR_MASTER_IO    (lcaio_device->config_reg_address + PCI_CFG_ADDR_DIR_MASTER_IO_OFFSET) 
#define PCI_TO_LOC_ADDR_1_RNG         (lcaio_device->config_reg_address + PCI_TO_LOC_ADDR_1_RNG_OFFSET)
#define LOC_BASE_ADDR_REMAP_1         (lcaio_device->config_reg_address + LOC_BASE_ADDR_REMAP_1_OFFSET) 
#define BUS_REG_DESC_1_FOR_PCI_LOC    (lcaio_device->config_reg_address + BUS_REG_DESC_1_FOR_PCI_LOC_OFFSET) 

/*** Run Time Registers ***/

#define MAILBOX_REGISTER_0_OFFSET           16
#define MAILBOX_REGISTER_1_OFFSET           17
#define MAILBOX_REGISTER_2_OFFSET           18
#define MAILBOX_REGISTER_3_OFFSET           19
#define MAILBOX_REGISTER_4_OFFSET           20
#define MAILBOX_REGISTER_5_OFFSET           21
#define MAILBOX_REGISTER_6_OFFSET           22
#define MAILBOX_REGISTER_7_OFFSET           23
#define PCI_TO_LOC_DOORBELL_OFFSET          24
#define LOC_TO_PCI_DOORBELL_OFFSET          25
#define INT_CTRL_STATUS_OFFSET              26
#define PROM_CTRL_CMD_CODES_CTRL_OFFSET     27
#define DEVICE_ID_VENDOR_ID_OFFSET          28
#define REVISION_ID_OFFSET                  29

#define MAILBOX_REG_0_OFFSET                30
#define MAILBOX_REG_1_OFFSET                31

#define MAILBOX_REGISTER_0           (lcaio_device->config_reg_address + MAILBOX_REGISTER_0_OFFSET)
#define MAILBOX_REGISTER_1           (lcaio_device->config_reg_address + MAILBOX_REGISTER_1_OFFSET)
#define MAILBOX_REGISTER_2           (lcaio_device->config_reg_address + MAILBOX_REGISTER_2_OFFSET)
#define MAILBOX_REGISTER_3           (lcaio_device->config_reg_address + MAILBOX_REGISTER_3_OFFSET)
#define MAILBOX_REGISTER_4           (lcaio_device->config_reg_address + MAILBOX_REGISTER_4_OFFSET)
#define MAILBOX_REGISTER_5           (lcaio_device->config_reg_address + MAILBOX_REGISTER_5_OFFSET)
#define MAILBOX_REGISTER_6           (lcaio_device->config_reg_address + MAILBOX_REGISTER_6_OFFSET)
#define MAILBOX_REGISTER_7           (lcaio_device->config_reg_address + MAILBOX_REGISTER_7_OFFSET)
#define PCI_TO_LOC_DOORBELL          (lcaio_device->config_reg_address + PCI_TO_LOC_DOORBELL_OFFSET)
#define LOC_TO_PCI_DOORBELL          (lcaio_device->config_reg_address + LOC_TO_PCI_DOORBELL_OFFSET)
#define INT_CTRL_STATUS              (lcaio_device->config_reg_address + INT_CTRL_STATUS_OFFSET)
#define PROM_CTRL_CMD_CODES_CTRL     (lcaio_device->config_reg_address + PROM_CTRL_CMD_CODES_CTRL_OFFSET)
#define DEVICE_ID_VENDOR_ID          (lcaio_device->config_reg_address + DEVICE_ID_VENDOR_ID_OFFSET)
#define REVISION_ID                  (lcaio_device->config_reg_address + REVISION_ID_OFFSET)
#define MAILBOX_REG_0                (lcaio_device->config_reg_address + MAILBOX_REG_0_OFFSET)
#define MAILBOX_REG_1                (lcaio_device->config_reg_address + MAILBOX_REG_1_OFFSET )

/*** Messaging Queue Registers ***/

#define OUT_POST_Q_INT_STATUS_OFFSET        12
#define OUT_POST_Q_INT_MASK_OFFSET          13
#define IN_Q_PORT_OFFSET                    16
#define OUT_Q_PORT_OFFSET                   17
#define MSG_UNIT_CONFIG_OFFSET              48
#define Q_BASE_ADDR_OFFSET                  49
#define IN_FREE_HEAD_PTR_OFFSET             50
#define IN_FREE_TAIL_PTR_OFFSET             51
#define IN_POST_HEAD_PTR_OFFSET             52
#define IN_POST_TAIL_PTR_OFFSET             53
#define OUT_FREE_HEAD_PTR_OFFSET            54
#define OUT_FREE_TAIL_PTR_OFFSET            55
#define OUT_POST_HEAD_PTR_OFFSET            56
#define OUT_POST_TAIL_PTR_OFFSET            57
#define Q_STATUS_CTRL_REG_OFFSET            58

#define OUT_POST_Q_INT_STATUS        (lcaio_device->config_reg_address + OUT_POST_Q_INT_STATUS_OFFSET)
#define OUT_POST_Q_INT_MASK          (lcaio_device->config_reg_address + OUT_POST_Q_INT_MASK_OFFSET )
#define IN_Q_PORT                    (lcaio_device->config_reg_address + IN_Q_PORT_OFFSET)
#define OUT_Q_PORT                   (lcaio_device->config_reg_address + OUT_Q_PORT_OFFSET)
#define MSG_UNIT_CONFIG              (lcaio_device->config_reg_address + MSG_UNIT_CONFIG_OFFSET)
#define Q_BASE_ADDR                  (lcaio_device->config_reg_address + Q_BASE_ADDR_OFFSET)
#define IN_FREE_HEAD_PTR             (lcaio_device->config_reg_address + IN_FREE_HEAD_PTR_OFFSET)
#define IN_FREE_TAIL_PTR             (lcaio_device->config_reg_address + IN_FREE_TAIL_PTR_OFFSET)
#define IN_POST_HEAD_PTR             (lcaio_device->config_reg_address + IN_POST_HEAD_PTR_OFFSET)
#define IN_POST_TAIL_PTR             (lcaio_device->config_reg_address + IN_POST_TAIL_PTR_OFFSET)
#define OUT_FREE_HEAD_PTR            (lcaio_device->config_reg_address + OUT_FREE_HEAD_PTR_OFFSET)
#define OUT_FREE_TAIL_PTR            (lcaio_device->config_reg_address + OUT_FREE_TAIL_PTR_OFFSET)
#define OUT_POST_HEAD_PTR            (lcaio_device->config_reg_address + OUT_POST_HEAD_PTR_OFFSET)
#define OUT_POST_TAIL_PTR            (lcaio_device->config_reg_address + OUT_POST_TAIL_PTR_OFFSET)
#define Q_STATUS_CTRL_REG            (lcaio_device->config_reg_address + Q_STATUS_CTRL_REG_OFFSET )

struct lcaio_registers
{
    unsigned long lcaio_register;	/* RANGE: 0-94 */
    unsigned long register_value;	/* RANGE: 0x0-0xFFFFFFFF */
};
struct lcaio_registers registers;

/*
 * IOCTL_LCAIO_REQ_INT_NOTIFY
 *
 * Parameter = lcaio_int_notify_params *int_notify_params;
 */

/* eIRQ0IntCondition */
#define IRQ0_NO_INT                              0x00000000
#define IRQ0_INIT_COMPLETE                       0x00000000
#define IRQ0_AUTOCAL_COMPLETE                    0x00000001
#define IRQ0_AUX_INPUT_LOW_TO_HIGH               0x00000002
#define IRQ0_AUX_INPUT_HIGH_TO_LOW               0x00000003

/* eIRQ1IntCondition */
#define IRQ1_NO_INT                              0x00000000
#define IRQ1_INPUT_BUFFER_THRESHOLD_LOW_TO_HIGH  0x00000001
#define IRQ1_INPUT_BUFFER_THRESHOLD_HIGH_TO_LOW  0x00000002

/* eIRQ2IntCondition */
#define IRQ2_NO_INT                              0x00000000
#define IRQ2_OUTPUT_BUFFER_THRESHOLD_LOW_TO_HIGH 0x00000001
#define IRQ2_OUTPUT_BUFFER_THRESHOLD_HIGH_TO_LOW 0x00000002
#define IRQ2_OUTPUT_BURST_COMPLETE               0x00000003

struct lcaio_int_notify_params
{
	unsigned long  irq0_int_condition;	// RANGE: 0-3
	unsigned long  irq1_int_condition;	// RANGE: 1-2
	unsigned long  irq2_int_condition;	// RANGE: 1-3
};


/*
 * IOCTL_LCAIO_SET_INPUT_MODE
 *
 * Parameter = unsigned long *inputmode;
 *     RANGE: 0-7
 */

#define MODE_DIFFERENTIAL 0
#define MODE_SINGLE_ENDED 1
#define MODE_ZERO_TEST    2
#define MODE_VREF_TEST    3
#define MODE_MON_OUT_CH0  4
#define MODE_MON_OUT_CH1  5
#define MODE_MON_OUT_CH2  6
#define MODE_MON_OUT_CH3  7

/*
 * IOCTL_LCAIO_SET_RANGE
 *
 * Parameter = unsigned long *inputrange;
 *     RANGE: 0-2
 */

#define RANGE_2p5V  0
#define RANGE_5V    1
#define RANGE_10V   2

/*
 * IOCTL_LCAIO_SET_DATA_FORMAT
 *
 * Parameter = unsigned long *data_format;
 *     RANGE: 0-1
 */

#define FORMAT_TWOS_COMPLEMENT 0
#define FORMAT_OFFSET_BINARY   1

/*
 * IOCTL_LCAIO_SET_OUTPUT_MODE
 *
 * Parameter = lcaio_output_mode *output_mode;
 */

/* Options for eSequential */
#define OUT_MODE_CHAN_SEQUENTIAL 0
#define OUT_MODE_SIMULTANEOUS    1

struct lcaio_outputmode
{
	unsigned long   sequential;			// RANGE: 0-1
	unsigned char   enable_output_burst;		// RANGE: FALSE-TRUE
	unsigned char   enable_output_looping;	      	// RANGE: FALSE-TRUE
};

/*
 * IOCTL_LCAIO_SET_OUTPUT_SYNC
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_SET_INPUT_SYNC
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_AUTO_CAL
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_INITIALIZE
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_SET_INPUT_BUFFER_THRESHOLD
 *
 * Parameter = unsigned long *inbuf_threshold;
 *     RANGE: 0x0 - 0x7FFF
 */

/*
 * IOCTL_LCAIO_CLEAR_INPUT_BUFFER
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_SET_OUTPUT_BUFFER_THRESHOLD
 *
 * Parameter = unsigned long *outbuf_threshold;
 *     RANGE: 0x0 - 0x7FFF
 */

/*
 * IOCTL_LCAIO_CLEAR_OUTPUT_BUFFER
 *
 * Parameter = NONE
 */

/*
 * IOCTL_LCAIO_SET_GEN_A_RATE
 * IOCTL_LCAIO_SET_GEN_B_RATE
 *
 * Parameter = unsigned long *gen_rate;		// Range: 0-0xFFFF
 */

#define MIN_FGEN 0.366217
#define MAX_FGEN 600.000

#define MIN_NRATE 0x0028
#define MAX_NRATE 0xFFFF

#define GEN_MULT   0.063875
#define GEN_OFFSET 511.0

#define ROUND(fValue) (((fValue - floor(fValue)) >= 0.5) ? ceil(fValue) : floor(fValue))

#define ROUND_TO_ULONG(fValue) ((unsigned long) ((long) ROUND(fValue)))

/* Converts a clock frequency in KHz to a Nrate value */
#define Fgen_To_Nrate(Fgen) ((Fgen < MIN_FGEN) ? MAX_NRATE : ((Fgen > MAX_FGEN) ? MIN_NRATE : ROUND_TO_ULONG(24000/Fgen)))

/*
 * IOCTL_LCAIO_SET_GEN_A_ENABLE
 *
 * Parameter = unsigned char *gen_A_enable;	// Range: FALSE-TRUE
 */

/*
 * IOCTL_LCAIO_SET_GEN_B_ENABLE
 *
 * Parameter = unsigned char *gen_B_enable;	// Range: FALSE-TRUE
 */

/*
 * IOCTL_LCAIO_SET_GEN_B_SRC
 *
 * Parameter = unsigned long *gen_B_source;	// Range: 0-1
 */

#define GEN_B_TIED_TO_MASTER_CLOCK 0
#define GEN_B_TIED_TO_GEN_A        1

/*
 * IOCTL_LCAIO_SET_INPUT_SCAN
 *
 * Parameter = lcaio_inputscan *inputscan;
 */

/* scan size options */
#define SCAN_SIZE_4_CHAN  0
#define SCAN_SIZE_8_CHAN  1
#define SCAN_SIZE_16_CHAN 2
#define SCAN_SIZE_32_CHAN 3

/* scan clock source options */
#define SCAN_CLK_SRC_GEN_A           0
#define SCAN_CLK_SRC_GEN_B           1
#define SCAN_CLK_SRC_EXT_SYNC        2
#define SCAN_CLK_SRC_BCR_INPUT_SYNC  3

/* scan mode options */
#define SCAN_MODE_MULTIPLE 0
#define SCAN_MODE_SINGLE   1

/* single scan channel limits */
#define MIN_CHANNEL  0
#define MAX_CHANNEL 31

struct lcaio_inputscan
{
	unsigned long   scan_size;			// Range: 0-3
	unsigned long   scan_clock_source;		// Range: 0-3
	unsigned long   scan_mode;			// Range: 0-1
	unsigned long   single_scan_channel;		// Range: 0-31
	unsigned int    two_channel_scan;		// Range: FALSE-TRUE
};

/*
 * IOCTL_LCAIO_SET_OUTPUT_SCAN
 *
 * Parameter = lcaio_outputscan *outputscan;
 */

/* scan clock source options */
#define SCAN_CLK_SRC_GEN_A    0
#define SCAN_CLK_SRC_GEN_B    1
#define SCAN_CLK_SRC_EXT_SYNC 2
#define SCAN_CLK_SRC_DISABLED 3

/* scan burst sync source */
#define SCAN_SYNC_SRC_GEN_A          0
#define SCAN_SYNC_SRC_GEN_B          1
#define SCAN_SYNC_SRC_EXT_SYNC       2
#define SCAN_SYNC_SRC_BCR_INPUT_SYNC 3

/* scan External sync source */
#define SCAN_EXT_SYNC_SRC_AIN       0
#define SCAN_EXT_SYNC_SRC_AOUT_SYNC 1
#define SCAN_EXT_SYNC_SRC_EXT_SYNC  2
#define SCAN_EXT_SYNC_SRC_DISABLED  3

struct lcaio_outputscan 
{
	unsigned long scan_clock_src;		// Range: 0-3
	unsigned long scan_burst_syncsrc;	// Range: 0-3
	unsigned long scan_ext_syncsrc;		// Range: 0-3
};

/*
 * IOCTL_LCAIO_SET_DIGITAL_IO_DIR
 *
 * Parameter = lcaio_digital_io_dir *digital_io_dir;
 */

#define DIGITAL_IO_DIR_INPUT  0
#define DIGITAL_IO_DIR_OUTPUT 1

struct lcaio_digital_io_dir
{
	unsigned long low_byte_dir;		// RANGE: 0-1
	unsigned long high_byte_dir;		// RANGE: 0-1
};

/*
 * IOCTL_LCAIO_READ_DIGITAL_IO
 *
 * Parameter = unsigned long *lcaio_digital_io;
 *     RANGE: 0-0xFFFF
 */
/*
 * IOCTL_LCAIO_SET_DIGITAL_IO
 *
 * Parameter = unsigned long *lcaio_digital_io;
 *     RANGE: 0-0xFFFF
 */

/* 
 * IOCTL_LCAIO_READ_AUX_INPUT
 *
 * Parameter = unsigned long *aux_input
 *     RANGE: 0-1
 */

#define AUX_INPUT_INACTIVE 0
#define AUX_INPUT_ACTIVE   1

/*
 * IOCTL_LCAIO_SET_AUX_OUTPUT
 *
 * Parameter = unsigned long *aux_output;
 *     RANGE: 0-1
 */

#define AUX_OUTPUT_INACTIVE 0
#define AUX_OUTPUT_ACTIVE   1

/*
 * IOCTL_LCAIO_GET_DEVICE_ERROR
 *
 * Parameter = unsigned long *lcaio_device_error;
 *     RANGE: 0x20000000-0x20000008
 */

#define LCAIO_SUCCESS                    0x20000000
#define LCAIO_INVALID_PARAMETER          0x20000001
#define LCAIO_INVALID_BUFFER_SIZE        0x20000002
#define LCAIO_PIO_TIMEOUT                0x20000003
#define LCAIO_DMA_TIMEOUT                0x20000004
#define LCAIO_IOCTL_TIMEOUT              0x20000005
#define LCAIO_OPERATION_CANCELLED        0x20000006
#define LCAIO_RESOURCE_ALLOCATION_ERROR  0x20000007
#define LCAIO_INVALID_REQUEST            0x20000008

/*
 * IOCTL_LCAIO_READ_PCI_CONFIG
 *
 * Parameter = lcaio_pci_config_regs *pci_config_regs;
 */

struct lcaio_pci_config_regs 
{
	unsigned int device_vendor_ID;
	unsigned int status_command;
	unsigned int classcode_revision_ID;
	unsigned int BIST_hdrtype_lattimer_cacheline_size;
	unsigned int runtime_register_addr;
	unsigned int config_register_addr;
	unsigned int PCI_base_addr2;
	unsigned int PCI_base_addr3;
	unsigned int unused_base_addr1;
	unsigned int unused_base_addr2;
	unsigned int cardbus_CIS_ptr;
	unsigned int subsystem_vendor_ID;
	unsigned int PCI_ROM_address;
	unsigned int reserved1;
	unsigned int reserved2;
	unsigned int maxlat_mingnt_intpin_intline;
};

/*
 * IOCTL_LCAIO_READ_LOCAL_CONFIG
 *
 * Parameter = lcaio_local_config_regs *local_config_regs;
 */

struct lcaio_local_config_regs
{

   /*********************************/
   /* Local Configuration Registers */
   /*********************************/

    unsigned long pci_local_range0;      /* 000  rw  range for pci to local 0 */
    unsigned long pci_local_remap0;      /* 004  rw  remap for pci to local 0 */
    unsigned long mode_arbitration;      /* 008  ro  mode arbitration         */
    unsigned long endian_setting;        /* 00C  ro  big/little endian descr. */
    unsigned long pci_local_exp_ROM_range; /* 010  rw  range for pci to local
                                              expansion rom     	      */
    unsigned long pci_local_exp_ROM_remap; /* 014  rw  remap for pci to local
                                              expansion rom                   */
    unsigned long pci_local_bus_region0;   /* 018  rw  bus region descriptions 
                                              for pci to local (space 0)      */
    unsigned long local_pci_range;         /* 01C  rw  range for local to pci */
    unsigned long local_pci_membase_addr;  /* 020  rw  base addr for local to 
                                              pci memory                      */
    unsigned long local_pci_IObase_addr;   /* 024  rw  base addr for local to 
                                              pci IO/Config    		      */
    unsigned long local_pci_remap;         /* 028  rw  remap for local to pci */
    unsigned long local_pci_config;        /* 02C  rw  pci config address reg 
                                              for local to pci IO/Config      */
    unsigned long outpost_Qint_status; 	   /* 030  ro  outboard post queue
                                              interrupt status                */
    unsigned long outpost_Qint_mask;   	   /* 034  rw  outboard post queue
                                              interrupt mask   		      */
    unsigned char reserved1[8];     	   /* 038      reserved               */

    /*****************************/
    /* Shared Run Time Registers */
    /*****************************/

    unsigned long mailbox[8];        	   /* 040  rw  8 mailbox registers   */
    unsigned long pci_local_doorbell;      /* 060  rw  pci to local doorbell */
    unsigned long local_pci_doorbell;      /* 064  rw  local to pci doorbell */
    unsigned long interrupt_control_status; /* 068  rw interrupt controlstatus*/
    unsigned long runtime_control;         /* 06C  rw  eeprom control, pci 
                                              command codes,user I/O,init ctrl*/
    unsigned long device_vendor_ID;    	   /* 070  ro  device id              */
    unsigned long revision_ID;        	   /* 074  ro  revision id            */
    unsigned long mailbox_reg0;       	   /* 078  rw  mailbox register 0     */
    unsigned long mailbox_reg1;       	   /* 07C  rw  mailbox register 1     */

    /***********************/
    /* Local DMA Registers */
    /***********************/

    unsigned long DMA_mode0;          	/* 080  rw  dma channel 0 mode        */
    unsigned long DMA_PCI_address0;    	/* 084  rw  dma channel 0 pci address */
    unsigned long DMA_local_address0;  	/* 088  rw  dma channel 0 localaddress*/
    unsigned long DMA_byte_count0;     	/* 08C  rw  dma channel 0 transfer 
                                           byte count                  	      */
    unsigned long DMA_descriptor_ptr0; 	/* 090  rw  dma channel 0 descriptor                                               pointer                            */
    unsigned long DMA_mode1;          	/* 094  rw  dma channel 1 mode        */
    unsigned long DMA_PCI_address1;    	/* 098  rw  dma channel 1 pci address */
    unsigned long DMA_local_address1;  	/* 09C  rw  dma channel 1 localaddress*/
    unsigned long DMA_byte_count1;     	/* 0A0  rw  dma channel 1 transfer 
                                           byte count                 	      */
    unsigned long DMA_descriptor_ptr1; 	/* 0A4  rw  dma channel 1 descriptor
                                           pointer                     	      */
    unsigned long DMA_command_status;   /* 0A8  rw  dma command/status 
                                           registers                          */
    unsigned long DMA_arbitration;    	/* 0AC  rw  dma arbitration register  */
    unsigned long DMA_threshold;      	/* 0B0  rw  dma threshold register    */
    unsigned char reserved3[12];    	/* 0B4  reserved                      */

    /*****************************/
    /* Messaging Queue Registers */
    /*****************************/

    unsigned long msg_unit_config;     /* 0C0  rw  messaging unit config      */
    unsigned long queue_baseaddr;      /* 0C4  rw  queue base address register*/
    unsigned long inbound_free_headptr;/* 0C8  rw  inbound free head pointer  */
    unsigned long inbound_free_tailptr;/* 0CC  rw  inbound free tail pointer  */
    unsigned long inbound_post_headptr;/* 0D0  rw  inbound post head pointer  */
    unsigned long inbound_post_tailptr;/* 0D4  rw  inbound post tail pointer  */
    unsigned long outbound_free_headptr;/* 0D8  rw  outbound free head pointer*/
    unsigned long outbound_free_tailptr;/* 0DC  rw  outbound free tail pointer*/
    unsigned long outbound_post_headptr;/* 0E0  rw  outbound post head pointer*/
    unsigned long outbound_post_tailptr;/* 0E4  rw  outbound post tail pointer*/
    unsigned long queue_status_control; /* 0E8  rw  queue status/control      */
    unsigned char reserved4[4];     	/* 0EC      reserved                  */

    unsigned long pci_local_range1;     /* 0F0  rw  range for pci to local 1  */
    unsigned long pci_local_remap1;     /* 0F4  rw  remap for pci to local 1  */
    unsigned long pci_local_bus_region1;/* 0F8  rw  bus region descriptions for
                                           pci to local (space 1)             */
};

/*
 * IOCTL_LCAIO_SET_TIMEOUT
 *
 * Parameter = unsigned long *timeout;
 *     RANGE: 0x0-0xFFFFFFFF, 0xFFFFFFFF=No Timeout
 */

/*
 * IOCTL_LCAIO_SET_DMA_ENABLE
 *
 * Parameter = unsigned char *DMA_enable;
 *     RANGE: 0 = Disable, 1 = Enable
 */

/*
 * IOCTL_LCAIO_READ_DEBUG
 *
 * Parameter = unsigned long *debug[16];
 */

/*
 * IOCTL_LCAIO_GET_WRITE_FIFO_AVAILABLE_SIZE 
 *
 *  Parameter = unsigned long write_fifo_size
 *       RANGE : 0 - FiFo available, 1 - FiFo not available
 */ 

#define WRITE_BUFFER_AVAILABLE 1
#define WRITE_BUFFER_NOT_AVAILABLE 0

/************** End Of Ioctl Commands ************************/


/* These are the masks for the LCAIO Board Control Register. */
#define BC_AIM0_MASK            0x0001
#define BC_AIM1_MASK            0x0002
#define BC_AIM2_MASK            0x0004
#define BC_AIM_MASK             0x0007
#define BC_AIM_SHIFT            0
#define BC_RANGE0_MASK          0x0010
#define BC_RANGE1_MASK          0x0020
#define BC_RANGE_MASK           0x0030
#define BC_RANGE_SHIFT          4
#define BC_DATA_FORMAT_MASK     0x0040
#define BC_DATA_FORMAT_SHIFT    6
#define BC_SIMUL_OUT_MASK       0x0100
#define BC_SIMUL_OUT_SHIFT      8
#define BC_EN_OUT_BURST_MASK    0x0200
#define BC_EN_OUT_BURST_SHIFT   9
#define BC_EN_OUT_LOOPING_MASK  0x0400
#define BC_EN_OUT_LOOPING_SHIFT 10
#define BC_OUTPUT_SYNC_MASK     0x0800
#define BC_OUTPUT_SYNC_SHIFT    11
#define BC_INPUT_SYNC_MASK      0x1000
#define BC_INPUT_SYNC_SHIFT     12
#define BC_AUTO_CAL_MASK        0x2000
#define BC_AUTO_CAL_SHIFT       13
#define BC_AUTOCAL_PASS_MASK    0x4000
#define BC_AUTOCAL_PASS_SHIFT   14
#define BC_INITIALIZE_MASK      0x8000
#define BC_INITIALIZE_SHIFT     15

/* These are the masks for the LCAIO Interrupt Control Registers. */
#define IC_INT_EVENT_MASK            0x0777
#define IC_INT_OCC_MASK              0x0888
#define IC_DISABLE_ALL_INTS_VAL      0x0000

#define IC_IRQ0_MASK                 0x0008
#define IC_IRQ0_SHIFT                0
#define IC_IDLE_VAL                  0x0000
#define IC_AUTO_CAL_COMP_VAL         0x0001
#define IC_AUX_LOW_TO_HIGH_VAL       0x0002
#define IC_AUX_HIGH_TO_LOW_VAL       0x0003
#define IC_IRQ0_REQUEST_MASK         0x0008

#define IC_IRQ1_MASK                 0x0070
#define IC_IRQ1_SHIFT                4
#define IC_IDLE_VAL                  0x0000
#define IC_INP_THRES_LOW_TO_HIGH_VAL 0x0010
#define IC_INP_THRES_HIGH_TO_LOW_VAL 0x0020
#define IC_IRQ1_REQUEST_MASK         0x0080

#define IC_IRQ2_MASK                 0x0700
#define IC_IRQ2_SHIFT                8
#define IC_IDLE_VAL                  0x0000
#define IC_OUT_THRES_LOW_TO_HIGH_VAL 0x0100
#define IC_OUT_THRES_HIGH_TO_LOW_VAL 0x0200
#define IC_OUT_BURST_COMP_VAL        0x0300
#define IC_IRQ2_REQUEST_MASK         0x0800

/* These are the masks for the LCAIO Input Data Buffer Register. */
#define IDB_DATA_IN_MASK     0x0000FFFF
#define IDB_CHAN_00_TAG_MASK 0x00010000

/* These are the masks for the LCAIO Analog Input Data. */

/*** OFFSET BINARY FORMAT ***/
#define OB_16BIT_MAX_POSITIVE             0xFFFF
#define OB_16BIT_PLUS_ONE                 0x8001
#define OB_16BIT_ZERO                     0x8000
#define OB_16BIT_MINUS_ONE                0x7FFF
#define OB_16BIT_MAX_NEGATIVE             0x0000

#define OB_12BIT_MAX_POSITIVE             0x0FFF
#define OB_12BIT_PLUS_ONE                 0x0801
#define OB_12BIT_ZERO                     0x0800
#define OB_12BIT_MINUS_ONE                0x07FF
#define OB_12BIT_MAX_NEGATIVE             0x0000

/*** TWO'S COMPLEMENT FORMAT ***/
#define TC_16BIT_MAX_POSITIVE             0x7FFF
#define TC_16BIT_PLUS_ONE                 0x0001
#define TC_16BIT_ZERO                     0x0000
#define TC_16BIT_MINUS_ONE                0xFFFF
#define TC_16BIT_MAX_NEGATIVE             0x8000

#define TC_12BIT_MAX_POSITIVE             0x07FF
#define TC_12BIT_PLUS_ONE                 0x0001
#define TC_12BIT_ZERO                     0x0000
#define TC_12BIT_MINUS_ONE                0xFFFF
#define TC_12BIT_MAX_NEGATIVE             0xF800

/* These are the masks for the LCAIO Input Data Buffer Control Register. */
#define IDBC_THRESHOLD_MASK      0x00007FFF
#define IDBC_CLEAR_BUFFER_MASK   0x00008000
#define IDBC_THRESHOLD_FLAG_MASK 0x00010000

/* These are the masks for the LCAIO Rate Generator Registers. */
#define RG_NRATE_MASK       0x0000FFFF
#define RG_GEN_DISABLE_MASK 0x00010000

/* These are the masks for the LCAIO Output Data Buffer Register. */
#define ODB_DATA_IN_MASK        0x0000FFFF
#define ODB_CHAN_TAG_MASK       0x00030000
#define ODB_CHAN_TAG_SHIFT      16
#define ODB_GROUP_END_FLAG_MASK 0x00040000
#define ODB_BURST_END_FLAG_MASK 0x00080000

/* These are the masks for the LCAIO Output Data Buffer Control Register. */
#define ODBC_THRESHOLD_MASK      0x00007FFF
#define ODBC_CLEAR_BUFFER_MASK   0x00008000
#define ODBC_THRESHOLD_FLAG_MASK 0x00010000

/* These are the masks for the LCAIO Scan and Sync Control Register. */
#define SC_AIN_SCAN_SIZE_MASK    0x00000003
#define SC_AIN_SCAN_SIZE_SHIFT   0
#define SC_AIN_SCAN_CLK_MASK     0x0000000C
#define SC_AIN_SCAN_CLK_SHIFT    2
#define SC_AOUT_CLK_MASK         0x00000030
#define SC_AOUT_CLK_SHIFT        4
#define SC_AOUT_SYNC_MASK        0x000000C0
#define SC_AOUT_SYNC_SHIFT       6
#define SC_AOUT_EXT_SYNC_MASK    0x00000300
#define SC_AOUT_EXT_SYNC_SHIFT   8
#define SC_RATE_B_CLK_MASK       0x00000400
#define SC_RATE_B_CLK_SHIFT      10
#define SC_AIN_SCAN_MODE_MASK    0x00000800
#define SC_AIN_SCAN_MODE_SHIFT   11
#define SC_AIN_SINGLE_CHAN_MASK  0x0001F000
#define SC_AIN_SINGLE_CHAN_SHIFT 12
#define SC_AIN_TWO_CHAN_MASK     0x00020000
#define SC_AIN_TWO_CHAN_SHIFT    17

/* These are the masks for the LCAIO Digital I/O Port Register. */
#define DIO_BOTH_BYTES_MASK     0x0000FFFF
#define DIO_LOW_BYTE_MASK       0x000000FF
#define DIO_HIGH_BYTE_MASK      0x0000FF00
#define DIO_AUX_INPUT_MASK      0x00010000
#define DIO_AUX_INPUT_SHIFT     16
#define DIO_AUX_OUTPUT_MASK     0x00020000
#define DIO_AUX_OUTPUT_SHIFT    17
#define DIO_LOW_BYTE_DIR_MASK   0x00040000
#define DIO_LOW_BYTE_DIR_SHIFT  18
#define DIO_HIGH_BYTE_DIR_MASK  0x00080000
#define DIO_HIGH_BYTE_DIR_SHIFT 19

/* These are the masks for the PCI DMA Control Registers. */
#define DMA_CMD_STAT_DONE      0x10
#define DMA_CMD_STAT_CMD_MASK  0x07
#define DMA_CMD_STAT_INT_CLEAR 0x08
#define DMA_CMD_STAT_START     0x0003
#define DMA_CMD_STAT_ABORT     0x0004
#define DMA_MODE_DONE_ENABLE   0x00000400

#define START_DMA_CMD_0        0x0B
#define START_DMA_CMD_1        0xB00

#define STOP_DMA_CMD_0_MASK    0xFF01
#define STOP_DMA_CMD_1_MASK    0x00FF

#define NON_DEMAND_DMA_MODE    0x00020D43
#define DEMAND_DMA_MODE        0x00021D43

#define DMA_LOCAL_TO_PCI_BUS   0xA
#define DMA_PCI_TO_LOCAL_BUS   0x2  

/***************** End Of Masks for registers ******************/

#endif // __LCAIO_IOCTL__

