
/*
 * Headers
 */
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include<string.h>
#include <curses.h>
#include <math.h>

#include "lcaioIoctl.h"
#include "LCAIO.h"

#define DEBUG_SIZE         16
#define ERROR_SUCCESS      0
#define INVALID_INT_VALUE -1

#define READ_BUFFER_SIZE 0x100000
#define WRITE_BUFFER_SIZE 0x100000

unsigned long ReadBuffer[READ_BUFFER_SIZE];
unsigned long WriteBuffer[WRITE_BUFFER_SIZE];

/* Pattern Types */

#define FIXED_PATTERN        0
#define	INCREMENTING_PATTERN 1
#define INVERTING_PATTERN    2
#define WALKING_ONES         3
#define WALKING_ZEROS        4
#define SPECIFIC_PATTERN     5

/* Scan Types */

#define SET_INPUT_SCAN  0
#define SET_OUTPUT_SCAN 1

/* Buffer Types */

#define INPUT_BUFFER  0
#define OUTPUT_BUFFER 1

/* Mode Types */

#define INPUT_MODE  0
#define OUTPUT_MODE 1

/* Sync Types */

#define INPUT_SYNC  0
#define OUTPUT_SYNC 1

/* Generator Types */

#define SET_GEN_A_RATE   0
#define SET_GEN_B_RATE   1
#define SET_GEN_A_ENABLE 2
#define SET_GEN_B_ENABLE 3

/* Register Types */

#define PCI_CONFIG_REGS   0
#define LOCAL_CONFIG_REGS 1
#define LCAIO_REGS        2

/* Local Register Types */

#define DMA_REGISTERS          0
#define LOCAL_CONFIG_REGISTERS 1
#define RUN_TIME_REGISTERS     2
#define MESSAGE_REGISTERS      3

/* Menu Entries */

#define LCAIO_NO_COMMAND                   0
#define LCAIO_AUTO_CAL                     1
#define LCAIO_INITIALIZE                   2
#define LCAIO_SET_MODE                     3
#define LCAIO_SET_FORMAT                   4
#define LCAIO_SET_SYNC                     5
#define LCAIO_SET_BUFFER_THRESHOLD         6
#define LCAIO_CLEAR_BUFFER                 7
#define LCAIO_SET_GEN                      8
#define LCAIO_SET_GEN_B_SRC                9
#define LCAIO_SET_SCAN                    10
#define LCAIO_SET_DIGITAL_IO_DIR          11
#define LCAIO_READ_DIGITAL_IO             12
#define LCAIO_SET_DIGITAL_IO              13
#define LCAIO_READ_AUX_INPUT              14
#define LCAIO_SET_AUX_OUTPUT              15
#define LCAIO_WRITE_FIFO                  16
#define LCAIO_READ_FIFO                   17
#define LCAIO_READ_REGISTERS              18
#define LCAIO_WRITE_REGISTER              19
#define LCAIO_REQ_INT_NOTIFY              20
#define LCAIO_GET_DEVICE_ERROR            21
#define LCAIO_SET_TIMEOUT                 22
#define LCAIO_SET_DMA_ENABLE              23
#define LCAIO_READ_DEBUG                  24
#define LCAIO_EXIT                        25

int GetLastError();

// struct sigaction io_act;

void signal_initialize();
//void sigio_handler (int );

/*
 * Global data
 */

int  fd;
unsigned long ulDataMask = 0xFFFF;
unsigned char bExitIntThread;
unsigned char bLCAIO16;

/*
 * Constants
 */

const char *pszRegisterNames[DIGITAL_IO_PORT_REG_OFFSET+1] = {
    "Board Control Register                  ",
    "Interrupt Control Register              ",
    "Input Data Buffer Register              ",
    "Input Buffer Control Register           ",
    "Rate A Generator Register               ",
    "Rate B Generator Register               ",
    "Output Data Buffer Register             ",
    "Output Buffer Control Register          ",
    "Scan and Sync Control Register          ",
    "Digital I/O Port Register               "
};

const char *pszPCIConfigRegisterNames[LAT_GNT_INT_PIN_LINE_OFFSET+1] = {
    "Device/Vendor ID Register               ",
    "Status/Command Register                 ",
    "Class Code/Revision ID Register         ",
    "BIST/Header Type/Lat Cache Size Register",
    "PCI Memory Base Address Register        ",
    "PCI I/O Base Address Register           ",
    "PCI Base Address 0 Register             ",
    "PCI Base Address 1 Register             ",
    (char *) NULL,
    (char *) NULL,
    "Cardbus CIS Pointer Register            ",
    "Subsystem ID/Vendor ID Register         ",
    "PCI Base Address Local ROM Register     ",
    (char *) NULL,
    (char *) NULL,
    "Lat Gnt/Int Pin Line Register           "
};

typedef struct RegStruct
{
    const char *pszRegisterName;
    int         iRegOffset;
} REG_STRUCT, *PREG_STRUCT;

const REG_STRUCT DMA_Regs[] = {
    {"DMA_CH_0_MODE",           32},
    {"DMA_CH_0_PCI_ADDR",       33},
    {"DMA_CH_0_LOCAL_ADDR",     34},
    {"DMA_CH_0_TRANS_BYTE_CNT", 35},
    {"DMA_CH_0_DESC_PTR",       36},
    {"DMA_CH_1_MODE",           37},
    {"DMA_CH_1_PCI_ADDR",       38},
    {"DMA_CH_1_LOCAL_ADDR",     39},
    {"DMA_CH_1_TRANS_BYTE_CNT", 40},
    {"DMA_CH_1_DESC_PTR",       41},
    {"DMA_CMD_STATUS",          42},
    {"DMA_MODE_ARB_REG",        43},
    {"DMA_THRESHOLD_REG",       44},
    {(char *) NULL,              0}
};

const REG_STRUCT Local_Config_Regs[] = {
    {"PCI_TO_LOC_ADDR_0_RNG",        0},
    {"LOC_BASE_ADDR_REMAP_0",        1},
    {"MODE_ARBITRATION",             2},
    {"BIG_LITTLE_ENDIAN_DESC",       3},
    {"PCI_TO_LOC_ROM_RNG",           4},
    {"LOC_BASE_ADDR_REMAP_EXP_ROM",  5},
    {"BUS_REG_DESC_0_FOR_PCI_LOC",   6},
    {"DIR_MASTER_TO_PCI_RNG",        7},
    {"LOC_ADDR_FOR_DIR_MASTER_MEM",  8},
    {"LOC_ADDR_FOR_DIR_MASTER_IO",   9},
    {"PCI_ADDR_REMAP_DIR_MASTER",   10},
    {"PCI_CFG_ADDR_DIR_MASTER_IO",  11},
    {"PCI_TO_LOC_ADDR_1_RNG",       92},
    {"LOC_BASE_ADDR_REMAP_1",       93},
    {"BUS_REG_DESC_1_FOR_PCI_LOC",  94},
    {(char *) NULL,                  0}
};

const REG_STRUCT Run_Time_Regs[] = {
    {"MAILBOX_REGISTER_0",       16},
    {"MAILBOX_REGISTER_1",       17},
    {"MAILBOX_REGISTER_2",       18},
    {"MAILBOX_REGISTER_3",       19},
    {"MAILBOX_REGISTER_4",       20},
    {"MAILBOX_REGISTER_5",       21},
    {"MAILBOX_REGISTER_6",       22},
    {"MAILBOX_REGISTER_7",       23},
    {"PCI_TO_LOC_DOORBELL",      24},
    {"LOC_TO_PCI_DOORBELL",      25},
    {"INT_CTRL_STATUS",          26},
    {"PROM_CTRL_CMD_CODES_CTRL", 27},
    {"DEVICE_ID_VENDOR_ID",      28},
    {"REVISION_ID",              29},
    {"MAILBOX_REG_0",            30},
    {"MAILBOX_REG_1",            31},
    {(char *) NULL,               0}
};

const REG_STRUCT Message_Regs[] = {
    {"OUT_POST_Q_INT_STATUS", 12},
    {"OUT_POST_Q_INT_MASK",   13},
    {"IN_Q_PORT",             16},
    {"OUT_Q_PORT",            17},
    {"MSG_UNIT_CONFIG",       48},
    {"Q_BASE_ADDR",           49},
    {"IN_FREE_HEAD_PTR",      50},
    {"IN_FREE_TAIL_PTR",      51},
    {"IN_POST_HEAD_PTR",      52},
    {"IN_POST_TAIL_PTR",      53},
    {"OUT_FREE_HEAD_PTR",     54},
    {"OUT_FREE_TAIL_PTR",     55},
    {"OUT_POST_HEAD_PTR",     56},
    {"OUT_POST_TAIL_PTR",     57},
    {"Q_STATUS_CTRL_REG",     58},
    {(char *) NULL,            0}
};

const REG_STRUCT * const LocalRegs[4] = {
    DMA_Regs,
    Local_Config_Regs,
    Run_Time_Regs,
    Message_Regs
};

const char *pszDeviceError[9] = {
    "LCAIO Success",
    "LCAIO Invalid Parameter",
    "LCAIO Invalid Buffer Size",
    "LCAIO PIO Timeout",
    "LCAIO DMA Timeout",
    "LCAIO Ioctl Timeout",
    "LCAIO Operation Cancelled",
    "LCAIO Resource Allocation Error",
    "LCAIO Invalid Request"
};

/*
 * Main entry point...
 */
void main(int argc, char *argv[])
{
    char				    iCh, ch;
    int             	            status;
    int				    iSelection;
    unsigned int                        uiMode;
    unsigned int                        uiSync;
    unsigned int                        uiBuffer;
    unsigned int                        uiGeneratorFunc;
    unsigned int                        uiScanFunc;
    unsigned int                        uiDeviceNum;
    float                      	    Fgen;
    unsigned int			    dwErrorCode;
    unsigned int			    dwTransferSize;
    unsigned long			    eLCAIORegister;
    unsigned int			    i;
    unsigned long                       ulAux;
    unsigned long                       ulRegType;
    unsigned long                       ulGenBSrc;
    unsigned long                       ulDebug[DEBUG_SIZE];
    unsigned long                       ulTimeout;
    unsigned long                       ulDeviceError;
    unsigned long                       ulRange;
    unsigned long                       eLowByteDir;
    unsigned long                       eHighByteDir;
    unsigned long                       ulInputMode;
    unsigned long                       ulDataFormat;
    unsigned long                       ulThreshold;
    unsigned long                       ulDigitalIO;
    unsigned long                       ulRegisterValue;
    unsigned long                       eScanSize;
    unsigned long                       eScanClockSource;
    unsigned long                       eScanMode;
    unsigned long			     eScanBurstSyncSource;
    unsigned long			     eScanExtSyncSource;
    unsigned long                       ulSingleScanChannel;
    unsigned long			     eIRQ0IntCondition;
    unsigned long			     eIRQ1IntCondition;
    unsigned long			     eIRQ2IntCondition;
    unsigned long			     eSequential;
    unsigned long                     bEnable;
    unsigned long			  GenEnable;
    unsigned char                     bTwoChannelScan;
    unsigned char			  bEnableOutputBurst;
    unsigned char			  bEnableOutputLooping;

    struct lcaio_local_config_regs    LocalConfigRegs;
    struct lcaio_pci_config_regs      ConfigRegs;

    if (argc < 2)
    {
        printf("\nUsage: LCAIOTEST <device_num>\n");
        printf("\nPress Enter to Continue: ");
        getchar();
        printf("\n");
        exit(ERROR_SUCCESS);
    }

    sscanf(argv[1], "%d", &uiDeviceNum);

    fd = LCAIO_open(uiDeviceNum);
    if (fd == INVALID_INT_VALUE)
    {
        printf ("LCAIO_open : Open failed, fd = %d\n", fd);
        exit(ERROR_SUCCESS);
    }
    signal_initialize();
    dwErrorCode = LCAIO_ioctl_read_pci_config_regs(fd,&ConfigRegs);
    if (dwErrorCode != LCAIO_SUCCESS)
    {
        ErrorMessage("LCAIOReadPCIConfigRegisters", dwErrorCode);
    }
    else
    {
        if (ConfigRegs.subsystem_vendor_ID == LCAIO16_SUBSYS_ID)
        {
            ulDataMask = 0xFFFF;
            bLCAIO16   = 1;
        }
        else
        {
            if (ConfigRegs.subsystem_vendor_ID == LCAIO12_SUBSYS_ID)
            {
                ulDataMask = 0x0FFF;
                bLCAIO16   = 0;
            }
            else
            {
                printf("Invalid Subsystem ID Detected\n");
                printf("\n");
                printf("Press Enter to Continue: ");
                getchar();
                printf("\n");

                LCAIO_close(fd);
                exit(ERROR_SUCCESS);
            }
        }
    }

    bExitIntThread = 0;

    do
    {
	printf("\n");
	printf("    PMC-%dLCAIO NT Driver Menu\n",
               bLCAIO16 ? 16 : 12);
	printf("\n");

	printf("A = No Command                N = Set Digital Output\n");
	printf("B = Auto Calibration          O = Read Aux Input\n");
	printf("C = Initialize                P = Set Aux Output\n");
	printf("D = Set Mode                  Q = Write FIFO\n");
	printf("E = Set Format                R = Read FIFO\n");
	printf("F = Set Sync                  S = Read Register\n");
	printf("G = Set Buffer Threshold      T = Write Register\n");
	printf("H = Clear Buffer              U = Request Int Notify\n");
	printf("I = Set Generator             V = Get Device Error\n");
	printf("J = Set Generator B Source    W = Set Timeout\n");
	printf("K = Set Scan                  X = Set DMA Enable\n");
	printf("L = Set Digital I/O Direction Y = Read Debug\n");
	printf("M = Read Digital Input        Z = Exit\n");
	printf("\n");

	printf("Enter Selection: ");
       
      	iCh = getchar();
        ch = getchar();

	if (islower(iCh))
	{
            iSelection = (int)(iCh - ((int) 'a'));
	}
	else
	{
            iSelection = (int)(iCh - ((int) 'A'));
	}
	switch (iSelection)
	{
        case LCAIO_NO_COMMAND:
            dwErrorCode = LCAIO_ioctl_no_function(fd);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIONoCommand", dwErrorCode);
            }
            break;

        case LCAIO_AUTO_CAL:
            printf ("\nBoard Autocalibration in Progress ...\n");
            dwErrorCode = LCAIO_ioctl_autocalibration(fd);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOAutoCalibration", dwErrorCode);
            }
            else printf ("\nBoard Autocalibration ... Done\n");
            break;

        case LCAIO_INITIALIZE:
            printf ("\nBoard Initialization in Progress ...\n");
            dwErrorCode = LCAIO_ioctl_initialize(fd);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOInitialize", dwErrorCode);
            }
            else printf ("\nBoard Initialization ... Done\n");
            break;

        case LCAIO_SET_MODE:
            uiMode = GetInt("\n\
0 = Set Input Mode\n\
1 = Set Output Mode\n\n\
Enter Mode to Set (0-1): ", INPUT_MODE, OUTPUT_MODE);

            if (uiMode == INPUT_MODE)
            {
                ulInputMode = GetInt("\n\
0 = Differential\n\
1 = Single Ended\n\
2 = ZERO Test\n\
3 = VREF Test\n\
4 = Monitor Output Channel 0\n\
5 = Monitor Output Channel 1\n\
6 = Monitor Output Channel 2\n\
7 = Monitor Output Channel 3\n\n\
Enter Input Mode (0-7): ", MODE_DIFFERENTIAL, MODE_MON_OUT_CH3);

                dwErrorCode = LCAIO_ioctl_set_input_mode(fd, ulInputMode);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetInputMode", dwErrorCode);
                }
            }
            else
            {
                eSequential = GetInt("\n\
0 = Channel Sequential\n\
1 = Simultaneous\n\n\
Enter Output Mode (0-1): ", OUT_MODE_CHAN_SEQUENTIAL, OUT_MODE_SIMULTANEOUS);

                bEnableOutputBurst   = 
                    YorN("\nEnter Output Burst Mode (Y/N): ");
                bEnableOutputLooping =
                    YorN("Enter Output Looping Mode (Y/N): ");

                dwErrorCode = LCAIO_ioctl_set_output_mode(fd, eSequential,
                                                          bEnableOutputBurst, bEnableOutputLooping);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetOutputMode", dwErrorCode);
                }
            }
            break;

        case LCAIO_SET_FORMAT:
            ulRange = GetInt("\n\
0 = +/- 2.5  Volts\n\
1 = +/- 5    Volts\n\
2 = +/- 10   Volts\n\n\
Enter Input Range (0-2): ", RANGE_2p5V, RANGE_10V);

            dwErrorCode = LCAIO_ioctl_set_input_range(fd, ulRange);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetRange", dwErrorCode);
            }

            ulDataFormat = GetInt("\n\
0 = Two's Complement\n\
1 = Offset Binary\n\n\
Enter Data Format (0-1): ", FORMAT_TWOS_COMPLEMENT, FORMAT_OFFSET_BINARY);

            dwErrorCode = LCAIO_ioctl_set_data_format(fd, ulDataFormat);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetDataFormat", dwErrorCode);
            }
            break;

        case LCAIO_SET_SYNC:
            uiSync = GetInt("\n\
0 = Set Input Sync\n\
1 = Set Output Sync\n\n\
Enter Sync to Set (0-1): ", INPUT_SYNC, OUTPUT_SYNC);

            if (uiSync == INPUT_SYNC)
            {
                printf ("\nInput Sync in Progress ...\n");
                dwErrorCode = LCAIO_ioctl_set_input_sync(fd);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetInputSync", dwErrorCode);
                }
                else printf ("\nInput Sync Operation ... Done\n");
            }
            else
            {
                printf ("\nOutput Sync in Progress ...\n");
                dwErrorCode = LCAIO_ioctl_set_output_sync(fd);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetOutput", dwErrorCode);
                }
                else printf ("\nOutput Sync Operation ... Done\n");
            }
            break;

        case LCAIO_SET_BUFFER_THRESHOLD:
            ulThreshold = GetHexLong(
                "Enter Input Buffer Threshold (0-0x7FFF): ");
            dwErrorCode = LCAIO_ioctl_set_inputbuffer_threshold (fd, 
                                                                 ulThreshold);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetInputBufferThreshold", dwErrorCode);
            }

            ulThreshold = GetHexLong(
                "Enter Output Buffer Threshold (0-0x7FFF): ");
            dwErrorCode = LCAIO_ioctl_set_outputbuffer_threshold ( fd, 
                                                                   ulThreshold);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetOutputBufferThreshold", dwErrorCode);
            }
            break;

        case LCAIO_CLEAR_BUFFER:
            uiBuffer = GetInt("\n\
0 = Clear Input Buffer\n\
1 = Clear Output Buffer\n\n\
Enter Sync to Set (0-1): ", INPUT_BUFFER, OUTPUT_BUFFER);

            if (uiBuffer == INPUT_BUFFER)
            {
                printf ("\nClearing Board Input Buffer in Progress ...\n");
                dwErrorCode = LCAIO_ioctl_clear_input_buffer(fd);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOClearInputBuffer", dwErrorCode);
                }
                else printf ("\nClearing Input Buffer ... Done\n");
            }
            else
            {
                printf ("\nClearing Board Output Buffer in Progress ...\n");
                dwErrorCode = LCAIO_ioctl_clear_output_buffer(fd);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOClearOutputBuffer", dwErrorCode);
                }
                else printf ("\nClearing Board Output Buffer ... Done\n");
            }
            break;

        case LCAIO_SET_GEN:
            uiGeneratorFunc = GetInt("\n\
0 = Set Generator A Rate\n\
1 = Set Generator B Rate\n\
2 = Set Generator A Enable\n\
3 = Set Generator B Enable\n\n\
Enter Generator Function (0-3): ", SET_GEN_A_RATE, SET_GEN_B_ENABLE);

            switch (uiGeneratorFunc)
            {
            default:

            case SET_GEN_A_RATE:
                Fgen = GetFloat(
                    "Enter Generator A Frequency (0.366217-600.000 kHz): ",
                    (float) MIN_FGEN, (float) MAX_FGEN);

                dwErrorCode = LCAIO_ioctl_set_generator_A_rate(fd,Fgen);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetGeneratorARate", dwErrorCode);
                }
                break;

            case SET_GEN_B_RATE:
                Fgen = GetFloat(
                    "Enter Generator B Frequency (0.366217-600.000 kHz): ",
                    (float) MIN_FGEN, (float) MAX_FGEN);
                dwErrorCode = LCAIO_ioctl_set_generator_B_rate(fd,Fgen);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetGeneratorBRate", dwErrorCode);
                }
                break;

            case SET_GEN_A_ENABLE:
                GenEnable = YorN("Enable Generator A (Y/N)?: ");
                dwErrorCode = LCAIO_ioctl_set_generator_A_enable(fd, 
                                                                 GenEnable);
                if (dwErrorCode != LCAIO_SUCCESS)
                { 
                    ErrorMessage("LCAIOSetGeneratorAEnable", dwErrorCode);
                }
                break;

            case SET_GEN_B_ENABLE:
                GenEnable = YorN("Enable Generator B (Y/N)?: ");
                printf ("GenEnable = %d\n", GenEnable);
                dwErrorCode = LCAIO_ioctl_set_generator_B_enable(fd, 
                                                                 GenEnable);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetGeneratorBEnable", dwErrorCode);
                }
                break;
            }
            break;

        case LCAIO_SET_GEN_B_SRC:
            ulGenBSrc = GetInt("\n\
0 = Tied to Master Clock\n\
1 = Tied to Generator A\n\n\
Enter Input Range (0-1): ", GEN_B_TIED_TO_MASTER_CLOCK, GEN_B_TIED_TO_GEN_A);

            dwErrorCode = LCAIO_ioctl_set_gen_b_clock_src(fd, ulGenBSrc);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetGeneratorBSource", dwErrorCode);
            }
            break;

        case LCAIO_SET_SCAN:
            uiScanFunc = GetInt("\n\
0 = Set Input Scan\n\
1 = Set Output Scan\n\n\
Enter Generator Function (0-1): ", SET_INPUT_SCAN, SET_OUTPUT_SCAN);

            if (uiScanFunc == SET_INPUT_SCAN)
            {
                eScanSize = GetInt("\n\
0 = 4  Channel Scan\n\
1 = 8  Channel Scan\n\
2 = 16 Channel Scan\n\
3 = 32 Channel Scan\n\n\
Enter Number of Channels to Input Scan (0-3): ", 
                                   SCAN_SIZE_4_CHAN, SCAN_SIZE_32_CHAN);

                eScanClockSource = GetInt("\n\
0 = Generator A\n\
1 = Generator B\n\
2 = External Sync\n\
3 = BCR Input Sync\n\n\
Enter Input Scan Clock Source (0-3): ", 
                                          SCAN_CLK_SRC_GEN_A, SCAN_CLK_SRC_BCR_INPUT_SYNC);

                eScanMode = GetInt("\n\
0 = Multiple Scan\n\
1 = Single Scan\n\n\
Enter Scan Mode (0-1): ", SCAN_MODE_MULTIPLE, SCAN_MODE_SINGLE);

                ulSingleScanChannel = GetInt("\n\
Enter Single Channel to Input Scan (0-31): ", MIN_CHANNEL, MAX_CHANNEL);

                bTwoChannelScan = 
                    YorN("Enable Two Channel Input Scan (Y/N)?: ");

                dwErrorCode = LCAIO_ioctl_set_input_scan(fd, eScanSize,						                          eScanClockSource,
                                                         eScanMode,									  ulSingleScanChannel,
                                                         bTwoChannelScan);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetupInputScan", dwErrorCode);
                }
            }
            else
            {
                eScanClockSource = GetInt("\n\
0 = Generator A\n\
1 = Generator B\n\
2 = External Sync\n\
3 = Disabled\n\n\
Enter Output Scan Clock Source (0-3): ", SCAN_CLK_SRC_GEN_A,
                                          SCAN_CLK_SRC_DISABLED);

                eScanBurstSyncSource = GetInt("\n\
0 = Generator A\n\
1 = Generator B\n\
2 = External Sync\n\
3 = BCR Input Sync\n\n\
Enter Input Scan Burst Sync Source (0-3): ", SCAN_SYNC_SRC_GEN_A,
                                              SCAN_SYNC_SRC_BCR_INPUT_SYNC);

                eScanExtSyncSource = GetInt("\n\
0 = Analog Input\n\
1 = Analog Output Sync\n\
2 = External Sync\n\
3 = Disabled\n\n\
Enter Input Scan External Sync Source (0-3): ", SCAN_EXT_SYNC_SRC_AIN,
                                            SCAN_EXT_SYNC_SRC_DISABLED);

                dwErrorCode = LCAIO_ioctl_set_output_scan(fd,					                               eScanClockSource,
                                                          eScanBurstSyncSource,							       eScanExtSyncSource);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOSetupOutputScan", dwErrorCode);
                }
            }
            break;

        case LCAIO_SET_DIGITAL_IO_DIR:
            eLowByteDir = GetInt("\n\
0 = Input\n\
1 = Output\n\n\
Enter Low Byte Digital I/O Direction (0-1): ", DIGITAL_IO_DIR_INPUT, 
                                 DIGITAL_IO_DIR_OUTPUT);

            eHighByteDir = GetInt("\n\
0 = Input\n\
1 = Output\n\n\
Enter High Byte Digital I/O Direction (0-1): ", DIGITAL_IO_DIR_INPUT,
                                  DIGITAL_IO_DIR_OUTPUT);

            dwErrorCode = LCAIO_ioctl_set_digital_io_dir(fd,
                                                         eLowByteDir, eHighByteDir);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetDigitalIODirection", dwErrorCode);
            }
            break;

        case LCAIO_READ_DIGITAL_IO:
            dwErrorCode = LCAIO_ioctl_read_digital_io(fd, &ulDigitalIO);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOReadDigitalIO", dwErrorCode);
            }
            else
            {
                printf("Digital I/O = %08lx\n", ulDigitalIO);
                printf("\n");
            }
            break;

        case LCAIO_SET_DIGITAL_IO:
            ulDigitalIO = ( unsigned long) GetHexShort(
                "Enter Digital I/O Value (0-0xFFFF): ");
            dwErrorCode = LCAIO_ioctl_set_digital_io(fd, ulDigitalIO);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetDigitalIO", dwErrorCode);
            }
            break;

        case LCAIO_READ_AUX_INPUT:
            dwErrorCode = LCAIO_ioctl_read_aux_input(fd, ulAux);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOReadAuxilliaryInput", dwErrorCode);
            }
            else
            {
                if (ulAux == AUX_INPUT_INACTIVE)
                {
                    printf("Aux Input INACTIVE\n");
                }
                else
                {
                    printf("Aux Input ACTIVE\n");
                }

                printf("\n");
                printf("Press Enter to Continue: ");
                getchar();
                printf("\n");
            }
            break;

        case LCAIO_SET_AUX_OUTPUT:
            ulAux = GetInt("\n\
0 = INACTIVE\n\
1 = ACTIVE\n\n\
Enter Aux Output State (0-1): ", AUX_OUTPUT_INACTIVE, AUX_OUTPUT_ACTIVE);

            dwErrorCode = LCAIO_ioctl_set_aux_output(fd, ulAux);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetAuxilliaryOutput", dwErrorCode);
            }
            break;

        case LCAIO_READ_REGISTERS:
            ulRegType = GetInt("\n\
0 = PCI Configuration Register\n\
1 = Local Configuration Register\n\
2 = LCAIO Register\n\n\
Enter Register Type to Read (0-2): ", PCI_CONFIG_REGS, LCAIO_REGS);

            switch (ulRegType)
            {
            default:
            case LCAIO_REGS:
                printf("\n");
                for (eLCAIORegister=BOARD_CTRL_REG_OFFSET;
                     eLCAIORegister<=DIGITAL_IO_PORT_REG_OFFSET;
                     eLCAIORegister++)
                {
                    iCh = 0;
                    if (eLCAIORegister == INPUT_DATA_BUFFER_REG_OFFSET)
                    {
                        printf("Read Data Buffer (Y/N): ");
                        iCh = getchar();
                        printf("\n");
                    }
                    if ((eLCAIORegister != INPUT_DATA_BUFFER_REG_OFFSET) ||
                        (iCh == 'Y') || (iCh == 'y'))
                    {
                        dwErrorCode = LCAIO_ioctl_read_register(fd,
                                                                eLCAIORegister,								&ulRegisterValue);
                        if (dwErrorCode != LCAIO_SUCCESS)
                        {
                            ErrorMessage("LCAIOReadRegister", dwErrorCode);
                        }	
                        else
                        {
                            printf("%s = %08lx\n", 
                                   pszRegisterNames[eLCAIORegister],
                                   ulRegisterValue);
                        }
                    }
                }
                break;

            case PCI_CONFIG_REGS:
                dwErrorCode = LCAIO_ioctl_read_pci_config_regs(fd, 
                                                               &ConfigRegs);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOReadPCIConfigRegisters", 
                                 dwErrorCode);
                }
                else
                {
                    printf("\n");
                    printf("Device ID/Vendor ID Reg        = %08lx\n",
                           ConfigRegs.device_vendor_ID);
                    printf("Status/Command Reg             = %08lx\n",
                           ConfigRegs.status_command);
                    printf("Class Code/Revision ID Reg     = %08lx\n",
                           ConfigRegs.classcode_revision_ID);

                    printf("BIST/HdrType/LatTimer/Cache... = %08lx\n", 			      ConfigRegs.BIST_hdrtype_lattimer_cacheline_size);

                    printf("Runtime Register Address Reg   = %08lx\n",
                           ConfigRegs.runtime_register_addr);
                    printf("Config Register Address Reg    = %08lx\n",
                           ConfigRegs.config_register_addr);
                    printf("PCI Base Address 2 Reg         = %08lx\n",
                           ConfigRegs.PCI_base_addr2);
                    printf("PCI Base Address 3 Reg         = %08lx\n",
                           ConfigRegs.PCI_base_addr3);
                    printf("Unused Base Address 1 Reg      = %08lx\n",
                           ConfigRegs.unused_base_addr1);
                    printf("Unused Base Address 2 Reg      = %08lx\n",
                           ConfigRegs.unused_base_addr2);
                    printf("Cardbus CIS Pointer Reg        = %08lx\n",
                           ConfigRegs.cardbus_CIS_ptr);
                    printf("Subsystem ID/Vendor ID Reg     = %08lx\n",
                           ConfigRegs.subsystem_vendor_ID);
                    printf("PCI Rom Address Reg            = %08lx\n",
                           ConfigRegs.PCI_ROM_address);
                    printf("Reserved 1 Reg                 = %08lx\n",
                           ConfigRegs.reserved1);
                    printf("Reserved 2 Reg                 = %08lx\n",
                           ConfigRegs.reserved2);
                    printf("MaxLat/MinGnt/Int Pin/Int Line = %08lx\n", 				ConfigRegs.maxlat_mingnt_intpin_intline);
                    getchar();
                }
                break;

            case LOCAL_CONFIG_REGS:
                dwErrorCode = LCAIO_ioctl_read_local_config_regs(fd,
                                                                 &LocalConfigRegs);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOReadLocalConfigRegisters", 
                                 dwErrorCode);
                }
                else
                {
                    printf("\n");
                    printf("    LOCAL CONFIGURATION REGISTERS\n\n");
                    printf("Range for PCI to Local 0 Reg    = %08lx\n",
                           LocalConfigRegs.pci_local_range0);
                    printf("Remap for PCI to Local 0 Reg    = %08lx\n",
                           LocalConfigRegs.pci_local_remap0);
                    printf("Mode Arbitration Reg            = %08lx\n",
                           LocalConfigRegs.mode_arbitration);
                    printf("Big/Little Endian Descr. Reg    = %08lx\n",
                           LocalConfigRegs.endian_setting);
                    printf("Range for PCI to Local Reg      = %08lx\n",
                           LocalConfigRegs.pci_local_exp_ROM_range);
                    printf("Remap for PCI to Local Reg      = %08lx\n",
                           LocalConfigRegs.pci_local_exp_ROM_remap);
                    printf("Bus Region Descriptions for Reg = %08lx\n",
                           LocalConfigRegs.pci_local_bus_region0);
                    printf("Range for Local to PCI Reg      = %08lx\n",
                           LocalConfigRegs.local_pci_range);
                    printf("Base Addr for Local to PCI Reg  = %08lx\n",
                           LocalConfigRegs.local_pci_membase_addr);
                    printf("Base Addr for Local to PCI Reg  = %08lx\n",
                           LocalConfigRegs.local_pci_IObase_addr);
                    printf("Remap for Local to PCI Reg      = %08lx\n",
                           LocalConfigRegs.local_pci_remap);
                    printf("PCI Config Address Reg for Reg  = %08lx\n",
                           LocalConfigRegs.local_pci_config);
                    printf("Range for PCI to Local 1 Reg    = %08lx\n",
                           LocalConfigRegs.pci_local_range1);
                    printf("Remap for PCI to Local 1 Reg    = %08lx\n",
                           LocalConfigRegs.pci_local_remap1);
                    printf("Bus Region Descriptor Reg       = %08lx\n",
                           LocalConfigRegs.pci_local_bus_region1);
                    printf("\n");
                    printf("Press Enter to Continue: ");
                    getchar();
                    getchar();
                    printf("\n");
                    printf("\n");
                    printf("    RUNTIME REGISTERS\n\n");
                    printf("Mailbox Register 0          = %08lx\n",
                           LocalConfigRegs.mailbox[0]);
                    printf("Mailbox Register 1          = %08lx\n",
                           LocalConfigRegs.mailbox[1]);
                    printf("Mailbox Register 2          = %08lx\n",
                           LocalConfigRegs.mailbox[2]);
                    printf("Mailbox Register 3          = %08lx\n",
                           LocalConfigRegs.mailbox[3]);
                    printf("Mailbox Register 4          = %08lx\n",
                           LocalConfigRegs.mailbox[4]);
                    printf("Mailbox Register 5          = %08lx\n",
                           LocalConfigRegs.mailbox[5]);
                    printf("Mailbox Register 6          = %08lx\n",
                           LocalConfigRegs.mailbox[6]);
                    printf("Mailbox Register 7          = %08lx\n",
                           LocalConfigRegs.mailbox[7]);
                    printf("PCI to Local Doorbell Reg   = %08lx\n",
                           LocalConfigRegs.pci_local_doorbell);
                    printf("Local to PCI Doorbell Reg   = %08lx\n",
                           LocalConfigRegs.local_pci_doorbell);
                    printf("Interrupt Control/Status    = %08lx\n",
                           LocalConfigRegs.interrupt_control_status);
                    printf("EEPROM Control, PCI Command = %08lx\n",
                           LocalConfigRegs.runtime_control);
                    printf("Device ID                   = %08lx\n",
                           LocalConfigRegs.device_vendor_ID);
                    printf("Revision ID                 = %08lx\n",
                           LocalConfigRegs.revision_ID);
                    printf("Mailbox Register 0          = %08lx\n",
                           LocalConfigRegs.mailbox_reg0);
                    printf("Mailbox Register 1          = %08lx\n",
                           LocalConfigRegs.mailbox_reg1);
                    printf("\n");
                    printf("Press Enter to Continue: ");
                    getchar();
                    printf("\n");
                    printf("\n");

                    printf("    DMA REGISTERS\n\n");
                    printf("dma channel 0 mode Reg           = %08lx\n",					   LocalConfigRegs.DMA_mode0);
                    printf("dma channel 0 pci address Reg    = %08lx\n",
                           LocalConfigRegs.DMA_PCI_address0);
                    printf("dma channel 0 local address Reg  = %08lx\n",
                           LocalConfigRegs.DMA_local_address0);
                    printf("dma channel 0 transfer byte Reg  = %08lx\n",
                           LocalConfigRegs.DMA_byte_count0);
                    printf("dma channel 0 descriptor Reg     = %08lx\n",
                           LocalConfigRegs.DMA_descriptor_ptr0);
                    printf("dma channel 1 mode Reg           = %08lx\n",
                           LocalConfigRegs.DMA_mode1);
                    printf("dma channel 1 pci address Reg    = %08lx\n",
                           LocalConfigRegs.DMA_PCI_address1);
                    printf("dma channel 1 local address Reg  = %08lx\n",
                           LocalConfigRegs.DMA_local_address1);
                    printf("dma channel 1 transfer byte Reg  = %08lx\n",
                           LocalConfigRegs.DMA_byte_count1);
                    printf("dma channel 1 descriptor Reg     = %08lx\n",
                           LocalConfigRegs.DMA_descriptor_ptr1);
                    printf("dma command/status registers Reg = %08lx\n",
                           LocalConfigRegs.DMA_command_status);
                    printf("dma arbitration register Reg     = %08lx\n",
                           LocalConfigRegs.DMA_arbitration);
                    printf("dma threshold register Reg       = %08lx\n",
                           LocalConfigRegs.DMA_threshold);
                    printf("\n");
                    printf("Press Enter to Continue: ");
                    getchar();
                    printf("\n");
                    printf("\n");
                    printf("	MESSAGING QUEUE REGISTERS\n\n");

                    printf("outbound post queue Int Status   = %08lx\n", 					LocalConfigRegs.outpost_Qint_status);
                    printf("outbound post queue Int Mask Reg = %08lx\n", 					LocalConfigRegs.outpost_Qint_mask);
                    printf("Mailbox Reg 0                    = %08lx\n",
                           LocalConfigRegs.mailbox[0]);
                    printf("Mailbox Reg 1                    = %08lx\n",
                           LocalConfigRegs.mailbox[1]);
                    printf("messaging unit configuration Reg = %08lx\n",
                           LocalConfigRegs.msg_unit_config);
                    printf("queue base address register Reg  = %08lx\n",
                           LocalConfigRegs.queue_baseaddr);
                    printf("inbound free head pointer Reg    = %08lx\n",
                           LocalConfigRegs.inbound_free_headptr);
                    printf("inbound free tail pointer Reg    = %08lx\n",
                           LocalConfigRegs.inbound_free_tailptr);
                    printf("inbound post head pointer Reg    = %08lx\n",
                           LocalConfigRegs.inbound_post_headptr);
                    printf("inbound post tail pointer Reg    = %08lx\n",
                           LocalConfigRegs.inbound_post_tailptr);
                    printf("outbound free head pointer Reg   = %08lx\n",
                           LocalConfigRegs.outbound_free_headptr);
                    printf("outbound free tail pointer Reg   = %08lx\n",
                           LocalConfigRegs.outbound_free_tailptr);
                    printf("outbound post head pointer Reg   = %08lx\n",
                           LocalConfigRegs.outbound_post_headptr);
                    printf("outbound post tail pointer Re    = %08lx\n",
                           LocalConfigRegs.outbound_post_tailptr);
                    printf("queue status/control Reg         = %08lx\n",
                           LocalConfigRegs.queue_status_control);
                }
                break;
            }
            printf("\n");
            break;

        case LCAIO_WRITE_REGISTER:
            ulRegType = GetInt("\n\
0 = PCI Configuration Register\n\
1 = Local Configuration Register\n\
2 = LCAIO Register\n\n\
Enter Register Type to Write (0-2): ", PCI_CONFIG_REGS, LCAIO_REGS);

            switch (ulRegType)
            {
            case LCAIO_REGS:
                eLCAIORegister  = eGetLCAIORegister();
                ulRegisterValue = ulGetRegisterValue();
                dwErrorCode = LCAIO_ioctl_write_register(fd,
                                                         eLCAIORegister, ulRegisterValue);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOWriteRegister", dwErrorCode);
                }
                break;

            case PCI_CONFIG_REGS:
                eLCAIORegister  = eGetLCAIOConfigRegister();
                ulRegisterValue = ulGetRegisterValue();
                dwErrorCode = LCAIO_ioctl_write_pci_config_regs(fd,
                                                                eLCAIORegister, ulRegisterValue);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOWritePCIConfigRegister", 
                                 dwErrorCode);
                }
                break;

            case LOCAL_CONFIG_REGS:
                eLCAIORegister  = eGetLCAIOLocalRegister();
                ulRegisterValue = ulGetRegisterValue();
                dwErrorCode = LCAIO_ioctl_write_local_config_regs(fd,
                                                                  eLCAIORegister, ulRegisterValue);
                if (dwErrorCode != LCAIO_SUCCESS)
                {
                    ErrorMessage("LCAIOWriteLocalConfigRegister", 
                                 dwErrorCode);
                }
                break;
            }
            break;

        case LCAIO_REQ_INT_NOTIFY:
            printf("Request IRQ0 Notification (Y/N)? ");
            iCh = getchar();
            getchar();
            printf("\n");

            if ((iCh == 'y') || (iCh == 'Y'))
            {
                eIRQ0IntCondition = GetInt("\n\
0 = Initialization Complete\n\
1 = Auto Calibration Complete\n\
2 = Aux Input LOW->HIGH\n\
3 = Aux Output HIGH->LOW\n\n\
Enter IRQ0 Interrupt Condition (0-3): ", IRQ0_INIT_COMPLETE, 
                                           IRQ0_AUX_INPUT_HIGH_TO_LOW);

            }
            else eIRQ0IntCondition = IRQ0_NO_INT;

            printf("Request IRQ1 Notification (Y/N)? ");
            iCh = getchar();
            getchar();
            printf("\n");

            if ((iCh == 'y') || (iCh == 'Y'))
            {
                eIRQ1IntCondition = GetInt("\n\
0 = No Interrupt\n\
1 = Input Buffer Threshold LOW->HIGH\n\
2 = Input Buffer Threshold HIGH->LOW\n\n\
Enter IRQ1 Interrupt Condition (0-2): ", IRQ1_NO_INT, 
                                           IRQ1_INPUT_BUFFER_THRESHOLD_HIGH_TO_LOW);

            }
            else
            {
                eIRQ1IntCondition = IRQ1_NO_INT;
            }

            printf("Request IRQ2 Notification (Y/N)? ");
            iCh = getchar();
            getchar();
            printf("\n");

            if ((iCh == 'y') || (iCh == 'Y'))
            {
                eIRQ2IntCondition = GetInt("\n\
0 = No Interrupt\n\
1 = Output Buffer Threshold LOW->HIGH\n\
2 = Output Buffer Threshold HIGH->LOW\n\
3 = Output Burst Complete\n\n\
Enter IRQ2 Interrupt Condition (0-3): ",IRQ2_NO_INT,IRQ2_OUTPUT_BURST_COMPLETE);

            }
            else
            {
                eIRQ2IntCondition = IRQ2_NO_INT;
            }

            dwErrorCode = LCAIO_ioctl_request_interrupt_notify(fd,
                                                               eIRQ0IntCondition,
                                                               eIRQ1IntCondition,
                                                               eIRQ2IntCondition);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIORequestInterruptNotify", dwErrorCode);
            }
            break;

        case LCAIO_SET_TIMEOUT:
            ulTimeout = GetLong("Enter Timeout (sec) (0=no timeout): ");
            dwErrorCode = LCAIO_ioctl_set_timeout(fd, 
                                                  (unsigned long)&ulTimeout);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetTimeout", dwErrorCode);
            }
            break;

        case LCAIO_GET_DEVICE_ERROR:
            dwErrorCode = LCAIO_ioctl_get_device_error(fd, 	&ulDeviceError);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOGetDeviceError", dwErrorCode);
            }
            else
            {
                printf("\nDevice Error = %s\n",
                       pszDeviceError[ulDeviceError-LCAIO_SUCCESS]);
            }
            break;

        case LCAIO_READ_FIFO:
            ReadFIFO();
            break;

        case LCAIO_WRITE_FIFO:
            WriteFIFO();
            break;

        case LCAIO_SET_DMA_ENABLE:
            bEnable = YorN("Enable DMA (Y/N)?: ");
            dwErrorCode = LCAIO_ioctl_set_dma_enable(fd, bEnable);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOSetDMAEnable", dwErrorCode);
            }
            break;

        case LCAIO_READ_DEBUG:
            dwErrorCode = LCAIO_ioctl_read_debug(fd, &ulDebug[0]);
            if (dwErrorCode != LCAIO_SUCCESS)
            {
                ErrorMessage("LCAIOReadDebug", dwErrorCode);
            }
            else
            {
                for (i=0;i<DEBUG_SIZE;i++)
                {
                    printf("Word %02d=%08lx\n", i, ulDebug[i]);
                }
                printf("\n");
            }
            break;

        case LCAIO_EXIT:
            break;

        default:
            printf("Invalid Selection!\n");
            break;
        }

        printf("\n");
        printf("Press Enter to Continue : ");
        ch = getchar();
        printf("\n");

    } while (iSelection != LCAIO_EXIT);

    bExitIntThread = 1;

    /*
     * All done with this; get rid of it
     */
    LCAIO_close(fd);
    exit(ERROR_SUCCESS);

} 

static  unsigned long eGetLCAIOConfigRegister()
{

    int iCh;
    int iRegister;

    do
    {
        printf("\n");
        for (iRegister=DEVICE_VENDOR_ID_OFFSET;
             iRegister<=LAT_GNT_INT_PIN_LINE_OFFSET;iRegister++)
        {
            if (pszPCIConfigRegisterNames[iRegister])
            {
                printf("%c = %s\n", (char) (((int) 'A') + iRegister),
                       pszPCIConfigRegisterNames[iRegister]);
            }
        }
        printf("Enter Selection: ");
        getchar();
        iCh = getchar();
        printf("\n");

        if (islower(iCh))
        {
            iRegister = iCh - ((int) 'a');
        }
        else
        {
            iRegister = iCh - ((int) 'A');
        }

        if ((iRegister < DEVICE_VENDOR_ID_OFFSET) || 
            (iRegister > LAT_GNT_INT_PIN_LINE_OFFSET) ||
            (! pszPCIConfigRegisterNames[iRegister]))
        {
            printf("Invalid Selection!\n");
        }
    } while ((iRegister < DEVICE_VENDOR_ID_OFFSET) 
             || (iRegister > LAT_GNT_INT_PIN_LINE_OFFSET) 
             || (! pszPCIConfigRegisterNames[iRegister]));

    return (( unsigned long) iRegister);

}  /* eGetLCAIOConfigRegister */


static  unsigned long eGetLCAIOLocalRegister()
{

    int               iCh;
    int               iRegister;
    int               iRegisterType;
    unsigned int              uiNumRegs;
    REG_STRUCT const *pRegs;

    do
    {
        printf("\n");
        printf("A = DMA Registers\n");
        printf("B = Local Configuration Registers\n");
        printf("C = Run Time Registers\n");
        printf("D = Message Queue Registers\n");
        printf("Enter Register Type: ");

        getchar();
        iCh = getchar();
        printf("\n");
        if (islower(iCh))
        {
            iRegisterType = iCh - ((int) 'a');
        }
        else
        {
            iRegisterType = iCh - ((int) 'A');
        }

        if ((iRegisterType < DMA_REGISTERS) ||
            (iRegisterType > MESSAGE_REGISTERS))
        {
            printf("Invalid Selection!\n");
        }
    } while ((iRegisterType < DMA_REGISTERS) ||
             (iRegisterType > MESSAGE_REGISTERS));

    pRegs = LocalRegs[iRegisterType];

    do
    {
        printf("\n");
        uiNumRegs = 0;
        for (iRegister=0;pRegs[iRegister].pszRegisterName;iRegister++)
        {
            uiNumRegs++;
            printf("%c = %s\n", (char) (((int) 'A') + iRegister),
                   pRegs[iRegister].pszRegisterName);
        }
        printf("Enter Selection: ");
        getchar();
        iCh = getchar();
        printf("\n");
        if (islower(iCh))
        {
            iRegister = iCh - ((int) 'a');
        }
        else
        {
            iRegister = iCh - ((int) 'A');
        }

        if ((iRegister < 0) || (iRegister >= ((int) uiNumRegs)))
        {
            printf("Invalid Selection!\n");
        }
    } while ((iRegister < 0) || (iRegister > ((int) uiNumRegs)));

    return (( unsigned long) pRegs[iRegister].iRegOffset);

}  /* eGetLCAIOLocalRegister */


static  unsigned long eGetLCAIORegister()
{

    int iCh;
    int iRegister;

    do
    {
        printf("\n");
        for (iRegister=BOARD_CTRL_REG_OFFSET;
             iRegister<=DIGITAL_IO_PORT_REG_OFFSET;iRegister++)
        {
            printf("%c = %s\n", (char) (((int) 'A') + iRegister),
                   pszRegisterNames[iRegister]);
        }
        printf("Enter Selection: ");
        getchar();
        iCh = getchar();
        printf("\n");
        if (islower(iCh))
        {
            iRegister = iCh - ((int) 'a');
        }
        else
        {
            iRegister = iCh - ((int) 'A');
        }

        if ((iRegister < BOARD_CTRL_REG_OFFSET) 
            || (iRegister > DIGITAL_IO_PORT_REG_OFFSET))
        {
            printf("Invalid Selection!\n");
        }
    } while ((iRegister < BOARD_CTRL_REG_OFFSET) 
             || (iRegister > DIGITAL_IO_PORT_REG_OFFSET));

    return (( unsigned long) iRegister);

}  /* eGetLCAIORegister */


static  unsigned long ulGetRegisterValue()
{

    int   iConv;
    unsigned long ulValue;

    do
    {
        printf("Enter Register Value (hex): ");
        iConv = scanf("%lx", &ulValue);
        if (iConv != 1)
        {
            printf("Invalid Entry!\n");
            printf("\n");
        }
    } while (iConv != 1);

    return (ulValue);

}  /* ulGetRegisterValue */


static float GetFloat(char *pszPrompt, float fMin, float fMax)
{
    float fValue;

    do
    {
        /* Prompt user for a floating point value. */
        printf(pszPrompt);

        /* Read the value. */
        scanf("%f", &fValue);

        /* Check Value */
        if ((fValue < fMin) || (fValue > fMax))
        {
            printf("Invalid Entry!\n\n");
        }
    } while ((fValue < fMin) || (fValue > fMax));

    /* Return the value. */
    return(fValue);

}  /* GetFloat */


static  unsigned long GetHexLong(char *pszPrompt)
{

    unsigned long ulValue;

    /* Prompt user for a hex long value. */
    printf(pszPrompt);

    /* Read the value. */
    scanf("%lx", &ulValue);

    /* Return the value. */
    return(ulValue);

}  /* GetHexLong */


static unsigned short GetHexShort(char *pszPrompt)
{

    unsigned short uhValue;

    /* Prompt the user for a hex short value. */
    printf(pszPrompt);

    /* Read the value from the user. */
    scanf("%hx", &uhValue);

    /* Return the value. */
    return(uhValue);

}  /* GetHexShort */


static unsigned int GetInt(char *pszPrompt, unsigned int uiLowerLimit, unsigned int uiUpperLimit)
{

    unsigned int uiValue;

    /* Loop through asking the user for an integer until the user enters a
       value within the given range. */
    do
    {
        /* Prompt for an integer. */
        printf(pszPrompt);
        /* Read the integer. */
        scanf("%d", &uiValue);
    } while((uiValue < uiLowerLimit) || (uiValue > uiUpperLimit));

    /* Return the value entered. */
    return(uiValue);

}  /* GetInt */


static long GetLong(char *pszPrompt)
{

    long lValue;

    /* Prompt for an integer. */
    printf(pszPrompt);
    /* Read the long. */
    scanf("%ld", &lValue);

    /* Return the value entered. */
    return(lValue);

}  /* GetLong */


static unsigned char YorN(char *pszPrompt)
{
    char iCh;
    char Ch1;
    unsigned char ret;

    printf(pszPrompt);

    Ch1 = getchar();
    iCh = getchar();

    printf("\n");
       
    if (( iCh == 'Y' )  || (iCh == 'y')) ret = 1;
    if (( iCh == 'N' )  || (iCh == 'n')) ret = 0;

    if ( ret != 0 || ret != 1 )  
    {
        if (( Ch1 == 'Y' )  || ( Ch1 == 'y')) ret = 1;
        if (( Ch1 == 'N' )  || ( Ch1 == 'n')) ret = 0;
    }

    return ret;          

}  /* YorN */


/*
 * This helper routine converts a system-service error
 * code into a text message and prints it on StdOutput
 */
static void ErrorMessage(char *lpOrigin, unsigned int dwMessageId)
{

    char *msgBuffer;		// string returned from system
    unsigned int  cBytes;		// length of returned string
    char   *pszMsgPtr;

    if ((dwMessageId >= LCAIO_SUCCESS) &&
        (dwMessageId <= LCAIO_INVALID_REQUEST))
    {
        pszMsgPtr = (char *) pszDeviceError[dwMessageId-LCAIO_SUCCESS];
    }
#ifdef DEBUG
    printf ("msg id = %x\n",dwMessageId);
#endif

    if (pszMsgPtr)
    {
        printf("Error: %s -- %s\n", lpOrigin, pszMsgPtr);
    }
    else
    {
        printf("FormatMessage error: %d\n", GetLastError());
    }

}  // ErrorMessage



int LCAIO_close(int fd)
{
    int	status;

    status	= close(fd);

    if (status == -1)
        printf("close() failure, errno = %d\n", errno);

    return(status);
}

int LCAIO_open(int board)
{
    int		fd;
    char	name[80];

    assert(board >= 0);

    sprintf(name, "/lcaio/%d", board);
    fd	= open(name, O_RDWR);

    if (fd == -1)
    {
        printf(	"open() failure on %s, errno = %d\n",
                name, errno);
    }

    printf("Open fd is %d\n", fd);

    return(fd);
}

int LCAIO_ioctl(int fd, int request, void *arg)
{
    int	status;

    status	= qnx_ioctl(fd, request, (unsigned long) arg, sizeof(arg), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}


int LCAIO_read(int fd, void *buf, size_t count)
{
    int	status;

    status	= read(fd, buf, count);

    if (status == -1)
        printf("read() failure, errno = %d\n", errno);
    else if (status != count)
        printf("read() %d of %d.\n", status, (int) count);

    return(status);
}

int LCAIO_write(int fd, const void *buf, size_t count)
{
    int	status;

    status	= write(fd, buf, count);

    if (status == -1)
        printf("write() failure, errno = %d\n", errno);
    else if (status != count)
        printf("write() %d of %d.\n", status, (int) count);

    return(status);
}

int LCAIO_ioctl_no_function(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_NO_COMMAND, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_register( int fd, unsigned long reg_no,
                               unsigned long *value)
{
    struct lcaio_registers parm;
    int status;

    parm.lcaio_register		= reg_no;
    parm.register_value		= 0xDEADBEEF;
    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_REGISTER, (unsigned long) &parm, sizeof(parm), (unsigned long) &parm, sizeof(parm));

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else 
    {
        value[0]	= parm.register_value;
        return(LCAIO_SUCCESS);
    }
}

int LCAIO_ioctl_write_register(	int fd,	unsigned long	reg_no,
                                unsigned long	value)
{
    struct lcaio_registers	parm;
    int status;

    parm.lcaio_register		= reg_no;
    parm.register_value		= value;
    status	= qnx_ioctl(fd,IOCTL_LCAIO_WRITE_REGISTER, &parm, sizeof(parm), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_request_interrupt_notify ( int fd, 
                                           unsigned long IRQ0IntCondition, 
                                           unsigned long IRQ1IntCondition, 
                                           unsigned long IRQ2IntCondition )
{
    int	status;
    struct lcaio_int_notify_params int_notify_params; 

    int_notify_params.irq0_int_condition = IRQ0IntCondition; 
    int_notify_params.irq1_int_condition = IRQ1IntCondition; 
    int_notify_params.irq2_int_condition = IRQ2IntCondition; 

    status	= qnx_ioctl(fd, IOCTL_LCAIO_REQ_INT_NOTIFY, (unsigned long)&int_notify_params, sizeof(int_notify_params), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_input_mode (int fd, unsigned int inputmode)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_INPUT_MODE, &inputmode, sizeof(inputmode), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_input_range (int fd, unsigned int inputrange)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_RANGE, &inputrange, sizeof(inputrange), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_data_format (int fd, unsigned int dataformat)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_DATA_FORMAT, &dataformat, sizeof(dataformat), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_output_mode (int fd, unsigned long Sequential,
                                 unsigned char EnableOutputBurst,
                                 unsigned char EnableOutputLooping)
{
    int	status;
    struct lcaio_outputmode outputmode; 

    outputmode.sequential = Sequential;
    outputmode.enable_output_burst = EnableOutputBurst; 
    outputmode.enable_output_looping = EnableOutputLooping;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_OUTPUT_MODE, (unsigned long)&outputmode, sizeof(outputmode), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_input_sync(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_INPUT_SYNC, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_output_sync(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_OUTPUT_SYNC, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_autocalibration(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_AUTO_CAL, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_initialize(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_INITIALIZE, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_inputbuffer_threshold (int fd, 
                                           unsigned int inbuf_threshold)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_INPUT_BUFFER_THRESHOLD,	&inbuf_threshold, sizeof(inbuf_threshold), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_clear_input_buffer(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_CLEAR_INPUT_BUFFER, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_outputbuffer_threshold (int fd, 
                                            unsigned int outbuf_threshold)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_OUTPUT_BUFFER_THRESHOLD, &outbuf_threshold, sizeof(outbuf_threshold), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_clear_output_buffer(int fd)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_CLEAR_OUTPUT_BUFFER, NULL, 0, NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_generator_A_rate(int fd, float genrate)
{
    int	status;
    unsigned long Nrate;
 
    Nrate = Fgen_To_Nrate (genrate);

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_GEN_A_RATE, &Nrate, sizeof(Nrate), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_generator_B_rate(int fd, float genrate)
{
    int	status;
    unsigned long Nrate;
 
    Nrate = Fgen_To_Nrate (genrate);

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_GEN_B_RATE, &Nrate, sizeof(Nrate), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_generator_A_enable(int fd, unsigned int genAenable)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_GEN_A_ENABLE, &genAenable, sizeof(genAenable), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_generator_B_enable(int fd, unsigned int genBenable)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_GEN_B_ENABLE, &genBenable, sizeof(genBenable), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_gen_b_clock_src(int fd, unsigned long source)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_GEN_B_SRC, &source, sizeof(source), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_input_scan ( int fd,unsigned long scansize, 
                                 unsigned long clksource, 
                                 unsigned long scanmode,
                                 unsigned long singlescanchannel,
                                 unsigned char twochannelscan )
{
    int	status;
    struct lcaio_inputscan inputscan; 

    inputscan.scan_size = scansize; 
    inputscan.scan_clock_source = clksource; 
    inputscan.scan_mode = scanmode; 
    inputscan.single_scan_channel = singlescanchannel;
    inputscan.two_channel_scan = twochannelscan;

    status	= qnx_ioctl (fd, IOCTL_LCAIO_SET_INPUT_SCAN, (unsigned long)&inputscan, sizeof(inputscan), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_output_scan ( int fd, unsigned long clksource, 
                                  unsigned long burstsyncsrc,
                                  unsigned long extsyncsrc)
{
    int	status;
    struct lcaio_outputscan outputscan; 

    outputscan.scan_clock_src = clksource; 
    outputscan.scan_burst_syncsrc = burstsyncsrc; 
    outputscan.scan_ext_syncsrc = extsyncsrc;
	
    status	= qnx_ioctl (fd, IOCTL_LCAIO_SET_OUTPUT_SCAN, (unsigned long)&outputscan, sizeof(outputscan), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_digital_io_dir ( int fd, unsigned long lowbytedir, unsigned long highbytedir)
{
    int	status;
    struct lcaio_digital_io_dir digital_io_dir; 

    digital_io_dir.low_byte_dir = lowbytedir; 
    digital_io_dir.high_byte_dir = highbytedir;
		
    status	= qnx_ioctl (fd, IOCTL_LCAIO_SET_DIGITAL_IO_DIR, (unsigned long)&digital_io_dir, sizeof(digital_io_dir), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_digital_io (int fd, unsigned long *digital_io)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_DIGITAL_IO, digital_io, sizeof(digital_io), digital_io, sizeof(digital_io));

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_digital_io (int fd, unsigned long digital_io)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_DIGITAL_IO, &digital_io, sizeof(digital_io), NULL, 0);


    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_aux_input (int fd, unsigned long auxinput)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_AUX_INPUT, &auxinput, sizeof(auxinput), &auxinput, sizeof(auxinput));

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 

}

int LCAIO_ioctl_set_aux_output (int fd, unsigned long auxoutput)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_AUX_OUTPUT, &auxoutput, sizeof(auxoutput), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_get_device_error (int fd, unsigned long *device_error)
{
    int	status;
    unsigned long device_errcode;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_GET_DEVICE_ERROR, device_errcode, sizeof(device_error), device_errcode, sizeof(device_error));
    *device_error = device_errcode;

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_pci_config_regs (int fd, 
                                      struct lcaio_pci_config_regs *pci_config_regs)
{
    int	status;

//	printf("ioctl read pci config reg parms [%p], [%d], [%p], [%d]\n", pci_config_regs, sizeof(struct lcaio_pci_config_regs), pci_config_regs, sizeof(struct lcaio_pci_config_regs));
    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_PCI_CONFIG, pci_config_regs, sizeof(struct lcaio_pci_config_regs), pci_config_regs, sizeof(struct lcaio_pci_config_regs));
        
    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_local_config_regs (int fd, 
                                        struct lcaio_local_config_regs *local_config_regs)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_LOCAL_CONFIG, local_config_regs, sizeof(struct lcaio_local_config_regs), local_config_regs, sizeof(struct lcaio_local_config_regs) );

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_write_pci_config_regs (int fd, unsigned long register_num, unsigned long register_value) 
						
{
    int	status;
    struct lcaio_registers registers;

    registers.lcaio_register = register_num;
    registers.register_value = register_value;
 
    status	= ioctl(fd, IOCTL_LCAIO_WRITE_PCI_CONFIG_REG, (unsigned long)&registers, sizeof(registers), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_write_local_config_regs (int fd, unsigned long register_num, unsigned long register_value) 
{
    int	status;

    registers.lcaio_register = register_num;
    registers.register_value = register_value;
 
    status	= qnx_ioctl(fd, IOCTL_LCAIO_WRITE_LOCAL_CONFIG_REG, (unsigned long)&registers, sizeof(registers), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_timeout(int fd, unsigned long timeout)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_TIMEOUT, &timeout, sizeof(timeout), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_set_dma_enable(int fd, unsigned int dmaenable)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_SET_DMA_ENABLE, &dmaenable, sizeof(dmaenable), NULL, 0);

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

int LCAIO_ioctl_read_debug(int fd, unsigned long *debug)
{
    int	status;

    status	= qnx_ioctl(fd, IOCTL_LCAIO_READ_DEBUG, &debug, sizeof(debug), &debug, sizeof(debug));

    if (status == -1)
    {
        printf("ioctl() failure, errno = %d\n", errno);
        return (GetLastError());
    }
    else return (LCAIO_SUCCESS); 
}

static void ReadFIFO()
{
    unsigned long	    WordsToRead;
    long	    BytesRead;
    unsigned long       ErrorCode;

    unsigned long       no_of_loops;
    int         iCh;
    int         i;
    int         status;
    int         word; 
    long        j;   

    do
    {     
        printf ("\nEnter the number of words to read ( in hex [1-0x9FFF] ) : ");
        scanf ("%x", &WordsToRead);

        if ( WordsToRead  <= 0 ||
             WordsToRead > (READ_BUFFER_SIZE-1) ) 
            printf ("Error : Invalid no. of words to write\n"); 

    } 
    while ( WordsToRead > READ_BUFFER_SIZE-1 
            || WordsToRead  <= 0 );

    printf ("\nEnter the number of loops for read (dec) : ");
    scanf ("%d", &no_of_loops);

    for ( i = 0; i < no_of_loops; i++)
    {
        BytesRead = read(fd, ReadBuffer,  WordsToRead*4);
        if (BytesRead < 0)
        {
	    printf("Read operation failed ...\n");
	    return;
        }
	else
        {
#ifdef DEBUG
            printf ("Data read from Input Buffer : \n");
            j = 0;
            for ( word = 0; word < BytesRead/4; word++)
            {
                j++;
 
                printf ("(%4.4d) %8.8X ", word, ReadBuffer[word]);
                if( j%32 == 0) {
                    printf ("\n");
                }
            }
#endif
            printf ("\n");

	    if (BytesRead != (4*WordsToRead))
            {
	 	printf("Loop %d : Only Read %d Words (%X)\n",
                       i, BytesRead/4, BytesRead/4);
            }
            else  printf ("Loop %d : Read %d words (%X) successfully\n",
                          i, BytesRead/4, BytesRead/4);
        }
    }

}  /* readfifo */


static void WriteFIFO()
{
    int         iCh;
    int         status;
    int         iPattType;
    int         iFirstChannel;
    int         iLastChannel;
    int         iChannel;
    int         no_of_loops;
    unsigned long       Loop, i;
    unsigned long       Value;
    unsigned long	    WordsToWrite;
    long	    BytesWritten = 0;
    unsigned long       ErrorCode;
    unsigned long       NumLoops;
    unsigned long       bGroup;
    unsigned long       bBurst;

  
    do
    {     
        printf ("\nEnter Number of Words to Write ( hex : 1 - 0x9FFF ): ");
        scanf ("%x", &WordsToWrite);

        if ( WordsToWrite <= 0 ||
             WordsToWrite > (WRITE_BUFFER_SIZE-1) ) 
            printf ("Error : Invalid no. of words to write\n"); 

    } 
    while ( WordsToWrite > WRITE_BUFFER_SIZE-1
            || WordsToWrite  <= 0 );

    do
    {
        printf("\n1 = Fixed Pattern\n");
        printf("2 = Incrementing Pattern\n");
        printf("3 = Inverting Pattern\n");
        printf("4 = Walking Ones\n");
        printf("5 = Walking Zeros\n");
        printf("6 = Specific Pattern\n");

        printf("\nEnter Pattern Type : ");
        scanf ("%d", &iPattType);

    } while ((iPattType < FIXED_PATTERN) || (iPattType > SPECIFIC_PATTERN));

    if ( (iPattType == FIXED_PATTERN) ||
         (iPattType == INCREMENTING_PATTERN) ||
         (iPattType == INVERTING_PATTERN) ) 
    {
        printf ("\nEnter Value to Write (in Hex) : ");
        scanf ("%x", &Value);

        Value = Value & ulDataMask;
        if (iPattType == FIXED_PATTERN)
        {
            for (i=0;i<(int) WordsToWrite;i++)
            {
                WriteBuffer[i] = Value;
            }
        }
    }
    else
    {
        if (iPattType == SPECIFIC_PATTERN)
        {
            for (i=0;i<(int) WordsToWrite;i++)
            {
                WriteBuffer[i] = i & ulDataMask;
            }
        }
        else
        {
            Value = 1;
        }
    }

    printf ("\nEnter First Channel ( 0-3 ): ");
    scanf ("%d", &iFirstChannel); 

    printf ("\nEnter Last Channel ( 0-3 ) : ");
    scanf ("%d", &iLastChannel); 

    bGroup = YorN("\nEnter Group Output (Y/N): ");

    if (bGroup == 0)
    {
        bBurst = YorN("\nEnter Burst Output (Y/N): ");
    }
    else
    {
        bBurst = 0;
    }
    for ( i = 0; i < (int) WordsToWrite; i++ )
    {  
	switch (iPattType)
        {
        case INCREMENTING_PATTERN:
            WriteBuffer[i] = Value;
            if (Value == ulDataMask)
            {
                Value = 0;
            }
            else
            {
                Value++;
            }
            break;
 
        case INVERTING_PATTERN:
            WriteBuffer[i] = Value;
            Value = (~ Value) & ulDataMask;
            break;

        case WALKING_ZEROS:
            WriteBuffer[i] = ~Value & ulDataMask;
            Value = Value << 1;
            if (! (Value & ulDataMask))
            {
                Value = 1;
            }
            break;

        case WALKING_ONES:
            WriteBuffer[i] = Value;
            Value = Value << 1;
            if (! (Value & ulDataMask))
            {
                Value = 1;
            }
            break;

        default:
            break;
        }
	/* Determine which channel */
	iChannel = iFirstChannel + (i % (iLastChannel - iFirstChannel + 1));

	/* Store Channel Tag */
	WriteBuffer[i] = WriteBuffer[i] | (iChannel << ODB_CHAN_TAG_SHIFT);

	/* Decide whether to store end burst or group bits */

	if (iChannel == iLastChannel)
        {
            if (bBurst)
            {
		WriteBuffer[i] = WriteBuffer[i] |
                    ODB_BURST_END_FLAG_MASK;
            }
            else
            {
		if (bGroup)
	  	{
                    WriteBuffer[i] = WriteBuffer[i] | ODB_GROUP_END_FLAG_MASK;
	        }
            }
        }
    }

#ifdef DEBUG
    printf ("testappl.c : write buffer :: \n");

    for ( i = 0; i < WordsToWrite; i++)
    {  
        printf(" %08lx ", WriteBuffer[i]);
    }
    printf("\n");
#endif

    printf ("Enter the number of loops for write (dec) : ");
    scanf ("%d", &no_of_loops);

    for ( Loop = 0; Loop < no_of_loops; Loop++)
    {
        BytesWritten = write(fd, WriteBuffer,  WordsToWrite*4);
        if (BytesWritten < 0)
        {
	    printf("\nWrite failed\n");
	    return;
        }
        else
        {
	    if (BytesWritten != (4*WordsToWrite))
            {
		printf("\nLoop %d : Only Wrote %d Words ( %X )\n", Loop, 
                       BytesWritten/4, BytesWritten/4);
            }
            else  printf ("\nLoop %d : Wrote %d words (%X) successfully\n", 
                          Loop, BytesWritten/4, BytesWritten/4);
        }
    }

}

void 
signal_initialize()
{
    int flags;

#ifdef DEBUG
    printf ("testappl.c :: <signal_initialize> function entry ...\n"); 
#endif


    flags = fcntl (fd, F_GETFL);
    if ( flags == -1)
    {
        printf ("Error : flags\n");
    }

}



int GetLastError()
{
    unsigned int ErrorCode;
    unsigned long DeviceError;
    ErrorCode = LCAIO_ioctl_get_device_error(fd, &DeviceError);
    printf ("........error = %x\n", DeviceError);

    if ( ErrorCode > 0 ) return DeviceError;
    else return ErrorCode;
}

/*
 * End of testappl.c 
 */
